# GAUL Level 1 admin lookup dictionary for Whisp
lookup_dict = {
    1000: {
        "gaul0_name": "Abyei",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xAB",
    },
    1001: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Adrar",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1002: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Ain-Defla",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1003: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Ain-Temouchent",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1004: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Alger",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1005: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Annaba",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1006: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Batna",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1007: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Bechar",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1008: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Bejaia",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1009: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Biskra",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1010: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Blida",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1011: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Bordj Bou Arrer",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1012: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Bouira",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1013: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Boumerdes",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1014: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Chlef",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1015: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Constantine",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1016: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Djelfa",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1017: {
        "gaul0_name": "Algeria",
        "gaul1_name": "El Bayadh",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1018: {
        "gaul0_name": "Algeria",
        "gaul1_name": "El Oued",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1019: {
        "gaul0_name": "Algeria",
        "gaul1_name": "El-Tarf",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1020: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Ghardaia",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1021: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Guelma",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1022: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Illizi",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1023: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Jijel",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1024: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Khenchela",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1025: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Laghouat",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1026: {
        "gaul0_name": "Algeria",
        "gaul1_name": "M'Sila",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1027: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Mascara",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1028: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Medea",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1029: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Mila",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1030: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Mostaganem",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1031: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Naama",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1032: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Oran",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1033: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Ouargla",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1034: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Oum El Bouaghi",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1035: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Relizane",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1036: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Saida",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1037: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Setif",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1038: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Sidi Bel Abbes",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1039: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Skikda",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1040: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Souk-Ahras",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1041: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Tamanrasset",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1042: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Tebessa",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1043: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Tiaret",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1044: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Tindouf",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1045: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Tipaza",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1046: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Tissemsilt",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1047: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Tizi Ouzou",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1048: {
        "gaul0_name": "Algeria",
        "gaul1_name": "Tlemcen",
        "iso2_code": "DZ",
        "iso3_code": "DZA",
    },
    1049: {
        "gaul0_name": "Angola",
        "gaul1_name": "Bengo",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1050: {
        "gaul0_name": "Angola",
        "gaul1_name": "Benguela",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1051: {
        "gaul0_name": "Angola",
        "gaul1_name": "Bié",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1052: {
        "gaul0_name": "Angola",
        "gaul1_name": "Cabinda",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1053: {
        "gaul0_name": "Angola",
        "gaul1_name": "Cuando Cubango",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1054: {
        "gaul0_name": "Angola",
        "gaul1_name": "Cuanza Norte",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1055: {
        "gaul0_name": "Angola",
        "gaul1_name": "Cuanza Sul",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1056: {
        "gaul0_name": "Angola",
        "gaul1_name": "Cunene",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1057: {
        "gaul0_name": "Angola",
        "gaul1_name": "Huambo",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1058: {
        "gaul0_name": "Angola",
        "gaul1_name": "Huíla",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1059: {
        "gaul0_name": "Angola",
        "gaul1_name": "Luanda",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1060: {
        "gaul0_name": "Angola",
        "gaul1_name": "Lunda Norte",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1061: {
        "gaul0_name": "Angola",
        "gaul1_name": "Lunda Sul",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1062: {
        "gaul0_name": "Angola",
        "gaul1_name": "Malanje",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1063: {
        "gaul0_name": "Angola",
        "gaul1_name": "Moxico",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1064: {
        "gaul0_name": "Angola",
        "gaul1_name": "Namibe",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1065: {
        "gaul0_name": "Angola",
        "gaul1_name": "Uíge",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1066: {
        "gaul0_name": "Angola",
        "gaul1_name": "Zaire",
        "iso2_code": "AO",
        "iso3_code": "AGO",
    },
    1067: {
        "gaul0_name": "Ascension, Saint Helena and Tristan da Cunha",
        "gaul1_name": "Ascension",
        "iso2_code": "SH",
        "iso3_code": "SHN",
    },
    1068: {
        "gaul0_name": "Ascension, Saint Helena and Tristan da Cunha",
        "gaul1_name": "Saint Helena",
        "iso2_code": "SH",
        "iso3_code": "SHN",
    },
    1069: {
        "gaul0_name": "Ascension, Saint Helena and Tristan da Cunha",
        "gaul1_name": "Tristan Da Cunha",
        "iso2_code": "SH",
        "iso3_code": "SHN",
    },
    1070: {
        "gaul0_name": "Bassas Da India",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xFR",
    },
    1071: {
        "gaul0_name": "Benin",
        "gaul1_name": "Alibori",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1072: {
        "gaul0_name": "Benin",
        "gaul1_name": "Atacora",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1073: {
        "gaul0_name": "Benin",
        "gaul1_name": "Atlantique",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1074: {
        "gaul0_name": "Benin",
        "gaul1_name": "Borgou",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1075: {
        "gaul0_name": "Benin",
        "gaul1_name": "Collines",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1076: {
        "gaul0_name": "Benin",
        "gaul1_name": "Couffo",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1077: {
        "gaul0_name": "Benin",
        "gaul1_name": "Donga",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1078: {
        "gaul0_name": "Benin",
        "gaul1_name": "Littoral",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1079: {
        "gaul0_name": "Benin",
        "gaul1_name": "Mono",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1080: {
        "gaul0_name": "Benin",
        "gaul1_name": "Oueme",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1081: {
        "gaul0_name": "Benin",
        "gaul1_name": "Plateau",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1082: {
        "gaul0_name": "Benin",
        "gaul1_name": "Zou",
        "iso2_code": "BJ",
        "iso3_code": "BEN",
    },
    1083: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Central",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1084: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Chobe",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1085: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Francistown",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1086: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Gaborone",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1087: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Ghanzi",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1088: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Jwaneng",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1089: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Kgalagadi",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1090: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Kgatleng",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1091: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Kweneng",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1092: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Lobatse",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1093: {
        "gaul0_name": "Botswana",
        "gaul1_name": "North East",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1094: {
        "gaul0_name": "Botswana",
        "gaul1_name": "North West",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1095: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Orapa",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1096: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Selibe Phikwe",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1097: {
        "gaul0_name": "Botswana",
        "gaul1_name": "South East",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1098: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Southern",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1099: {
        "gaul0_name": "Botswana",
        "gaul1_name": "Sowa Town",
        "iso2_code": "BW",
        "iso3_code": "BWA",
    },
    1100: {
        "gaul0_name": "Bouvet Island",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "BV",
        "iso3_code": "BVT",
    },
    1101: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Boucle Du Mouhoun",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1102: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Cascades",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1103: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Centre",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1104: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Centre-Est",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1105: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Centre-Nord",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1106: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Centre-Ouest",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1107: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Centre-Sud",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1108: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Est",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1109: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Hauts-Bassins",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1110: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Nord",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1111: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Plateau-Central",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1112: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Sahel",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1113: {
        "gaul0_name": "Burkina Faso",
        "gaul1_name": "Sud-Ouest",
        "iso2_code": "BF",
        "iso3_code": "BFA",
    },
    1114: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Bubanza",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1115: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Bujumbura",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1116: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Bururi",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1117: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Cankuzo",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1118: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Cibitoke",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1119: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Gitega",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1120: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Karuzi",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1121: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Kayanza",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1122: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Kirundo",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1123: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Mairie De Bujumbura",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1124: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Makamba",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1125: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Muramvya",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1126: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Muyinga",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1127: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Mwaro",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1128: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Ngozi",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1129: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Rumonge",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1130: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Rutana",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1131: {
        "gaul0_name": "Burundi",
        "gaul1_name": "Ruyigi",
        "iso2_code": "BI",
        "iso3_code": "BDI",
    },
    1132: {
        "gaul0_name": "Bīr Ṭawīl",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xMS",
    },
    1133: {
        "gaul0_name": "Cabo Verde",
        "gaul1_name": "Boavista",
        "iso2_code": "CV",
        "iso3_code": "CPV",
    },
    1134: {
        "gaul0_name": "Cabo Verde",
        "gaul1_name": "Brava",
        "iso2_code": "CV",
        "iso3_code": "CPV",
    },
    1135: {
        "gaul0_name": "Cabo Verde",
        "gaul1_name": "Fogo",
        "iso2_code": "CV",
        "iso3_code": "CPV",
    },
    1136: {
        "gaul0_name": "Cabo Verde",
        "gaul1_name": "Maio",
        "iso2_code": "CV",
        "iso3_code": "CPV",
    },
    1137: {
        "gaul0_name": "Cabo Verde",
        "gaul1_name": "Sal",
        "iso2_code": "CV",
        "iso3_code": "CPV",
    },
    1138: {
        "gaul0_name": "Cabo Verde",
        "gaul1_name": "Santiago",
        "iso2_code": "CV",
        "iso3_code": "CPV",
    },
    1139: {
        "gaul0_name": "Cabo Verde",
        "gaul1_name": "Santo Antao",
        "iso2_code": "CV",
        "iso3_code": "CPV",
    },
    1140: {
        "gaul0_name": "Cabo Verde",
        "gaul1_name": "Sao Nicolau",
        "iso2_code": "CV",
        "iso3_code": "CPV",
    },
    1141: {
        "gaul0_name": "Cabo Verde",
        "gaul1_name": "Sao Vicente",
        "iso2_code": "CV",
        "iso3_code": "CPV",
    },
    1142: {
        "gaul0_name": "Cameroon",
        "gaul1_name": "Adamaoua",
        "iso2_code": "CM",
        "iso3_code": "CMR",
    },
    1143: {
        "gaul0_name": "Cameroon",
        "gaul1_name": "Centre",
        "iso2_code": "CM",
        "iso3_code": "CMR",
    },
    1144: {
        "gaul0_name": "Cameroon",
        "gaul1_name": "Est",
        "iso2_code": "CM",
        "iso3_code": "CMR",
    },
    1145: {
        "gaul0_name": "Cameroon",
        "gaul1_name": "Extrême-Nord",
        "iso2_code": "CM",
        "iso3_code": "CMR",
    },
    1146: {
        "gaul0_name": "Cameroon",
        "gaul1_name": "Littoral",
        "iso2_code": "CM",
        "iso3_code": "CMR",
    },
    1147: {
        "gaul0_name": "Cameroon",
        "gaul1_name": "Nord",
        "iso2_code": "CM",
        "iso3_code": "CMR",
    },
    1148: {
        "gaul0_name": "Cameroon",
        "gaul1_name": "Nord-Ouest",
        "iso2_code": "CM",
        "iso3_code": "CMR",
    },
    1149: {
        "gaul0_name": "Cameroon",
        "gaul1_name": "Ouest",
        "iso2_code": "CM",
        "iso3_code": "CMR",
    },
    1150: {
        "gaul0_name": "Cameroon",
        "gaul1_name": "Sud",
        "iso2_code": "CM",
        "iso3_code": "CMR",
    },
    1151: {
        "gaul0_name": "Cameroon",
        "gaul1_name": "Sud-Ouest",
        "iso2_code": "CM",
        "iso3_code": "CMR",
    },
    1152: {
        "gaul0_name": "Central African Republic",
        "gaul1_name": "Bas-Oubangui",
        "iso2_code": "CF",
        "iso3_code": "CAF",
    },
    1153: {
        "gaul0_name": "Central African Republic",
        "gaul1_name": "Equateur",
        "iso2_code": "CF",
        "iso3_code": "CAF",
    },
    1154: {
        "gaul0_name": "Central African Republic",
        "gaul1_name": "Fertit",
        "iso2_code": "CF",
        "iso3_code": "CAF",
    },
    1155: {
        "gaul0_name": "Central African Republic",
        "gaul1_name": "Haut-Oubangui",
        "iso2_code": "CF",
        "iso3_code": "CAF",
    },
    1156: {
        "gaul0_name": "Central African Republic",
        "gaul1_name": "Kagas",
        "iso2_code": "CF",
        "iso3_code": "CAF",
    },
    1157: {
        "gaul0_name": "Central African Republic",
        "gaul1_name": "Plateaux",
        "iso2_code": "CF",
        "iso3_code": "CAF",
    },
    1158: {
        "gaul0_name": "Central African Republic",
        "gaul1_name": "Yadé",
        "iso2_code": "CF",
        "iso3_code": "CAF",
    },
    1159: {
        "gaul0_name": "Chad",
        "gaul1_name": "Barh-El-Gazel",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1160: {
        "gaul0_name": "Chad",
        "gaul1_name": "Batha",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1161: {
        "gaul0_name": "Chad",
        "gaul1_name": "Borkou",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1162: {
        "gaul0_name": "Chad",
        "gaul1_name": "Chari-Baguirmi",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1163: {
        "gaul0_name": "Chad",
        "gaul1_name": "Ennedi Est",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1164: {
        "gaul0_name": "Chad",
        "gaul1_name": "Ennedi Ouest",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1165: {
        "gaul0_name": "Chad",
        "gaul1_name": "Guéra",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1166: {
        "gaul0_name": "Chad",
        "gaul1_name": "Hadjer Lamis",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1167: {
        "gaul0_name": "Chad",
        "gaul1_name": "Hadjer-Lamis",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1168: {
        "gaul0_name": "Chad",
        "gaul1_name": "Kanem",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1169: {
        "gaul0_name": "Chad",
        "gaul1_name": "Lac",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1170: {
        "gaul0_name": "Chad",
        "gaul1_name": "Logone Occidental",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1171: {
        "gaul0_name": "Chad",
        "gaul1_name": "Logone Oriental",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1172: {
        "gaul0_name": "Chad",
        "gaul1_name": "Mandoul",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1173: {
        "gaul0_name": "Chad",
        "gaul1_name": "Mayo-Kebbi Est",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1174: {
        "gaul0_name": "Chad",
        "gaul1_name": "Mayo-Kebbi Ouest",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1175: {
        "gaul0_name": "Chad",
        "gaul1_name": "Moyen-Chari",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1176: {
        "gaul0_name": "Chad",
        "gaul1_name": "N'Djamena",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1177: {
        "gaul0_name": "Chad",
        "gaul1_name": "Ouaddaï",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1178: {
        "gaul0_name": "Chad",
        "gaul1_name": "Salamat",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1179: {
        "gaul0_name": "Chad",
        "gaul1_name": "Sila",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1180: {
        "gaul0_name": "Chad",
        "gaul1_name": "Tandjilé",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1181: {
        "gaul0_name": "Chad",
        "gaul1_name": "Tibesti",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1182: {
        "gaul0_name": "Chad",
        "gaul1_name": "Wadi Fira",
        "iso2_code": "TD",
        "iso3_code": "TCD",
    },
    1183: {
        "gaul0_name": "Comoros",
        "gaul1_name": "Anjouan (Ndzouani)",
        "iso2_code": "KM",
        "iso3_code": "COM",
    },
    1184: {
        "gaul0_name": "Comoros",
        "gaul1_name": "Grande Comore (Ngazidja)",
        "iso2_code": "KM",
        "iso3_code": "COM",
    },
    1185: {
        "gaul0_name": "Comoros",
        "gaul1_name": "Mohéli (Mwali)",
        "iso2_code": "KM",
        "iso3_code": "COM",
    },
    1186: {
        "gaul0_name": "Congo",
        "gaul1_name": "Bouenza",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1187: {
        "gaul0_name": "Congo",
        "gaul1_name": "Brazzaville",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1188: {
        "gaul0_name": "Congo",
        "gaul1_name": "Cuvette",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1189: {
        "gaul0_name": "Congo",
        "gaul1_name": "Cuvette-Ouest",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1190: {
        "gaul0_name": "Congo",
        "gaul1_name": "Kouilou",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1191: {
        "gaul0_name": "Congo",
        "gaul1_name": "Lekoumou",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1192: {
        "gaul0_name": "Congo",
        "gaul1_name": "Likouala",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1193: {
        "gaul0_name": "Congo",
        "gaul1_name": "Niari",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1194: {
        "gaul0_name": "Congo",
        "gaul1_name": "Plateaux",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1195: {
        "gaul0_name": "Congo",
        "gaul1_name": "Point-Noire",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1196: {
        "gaul0_name": "Congo",
        "gaul1_name": "Pool",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1197: {
        "gaul0_name": "Congo",
        "gaul1_name": "Sangha",
        "iso2_code": "CG",
        "iso3_code": "COG",
    },
    1198: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Abidjan",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1199: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Bas-Sassandra",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1200: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Comoé",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1201: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Denguélé",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1202: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Gôh-Djiboua",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1203: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Lacs",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1204: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Lagunes",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1205: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Montagnes",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1206: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Sassandra-Marahoué",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1207: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Savanes",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1208: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Vallée Du Bandama",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1209: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Woroba",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1210: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Yamoussoukro",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1211: {
        "gaul0_name": "Côte D'Ivoire",
        "gaul1_name": "Zanzan",
        "iso2_code": "CI",
        "iso3_code": "CIV",
    },
    1212: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Bas-Uélé",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1213: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Haut-Katanga",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1214: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Haut-Lomami",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1215: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Haut-Uélé",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1216: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Ituri",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1217: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Kasaï",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1218: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Kasaï-Central",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1219: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Kasaï-Oriental",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1220: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Kinshasa",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1221: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Kongo Central",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1222: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Kwango",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1223: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Kwilu",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1224: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Lomami",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1225: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Lualaba",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1226: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Mai-Ndombe",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1227: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Maniema",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1228: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Mongala",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1229: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Nord-Kivu",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1230: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Nord-Ubangi",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1231: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Sankuru",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1232: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Sud-Kivu",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1233: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Sud-Ubangi",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1234: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Tanganyika",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1235: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Tshopo",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1236: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Tshuapa",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1237: {
        "gaul0_name": "Democratic Republic of the Congo",
        "gaul1_name": "Équateur",
        "iso2_code": "CD",
        "iso3_code": "COD",
    },
    1238: {
        "gaul0_name": "Djibouti",
        "gaul1_name": "Ali Sabieh",
        "iso2_code": "DJ",
        "iso3_code": "DJI",
    },
    1239: {
        "gaul0_name": "Djibouti",
        "gaul1_name": "Arta",
        "iso2_code": "DJ",
        "iso3_code": "DJI",
    },
    1240: {
        "gaul0_name": "Djibouti",
        "gaul1_name": "Dikhil",
        "iso2_code": "DJ",
        "iso3_code": "DJI",
    },
    1241: {
        "gaul0_name": "Djibouti",
        "gaul1_name": "Djiboutii",
        "iso2_code": "DJ",
        "iso3_code": "DJI",
    },
    1242: {
        "gaul0_name": "Djibouti",
        "gaul1_name": "Obock",
        "iso2_code": "DJ",
        "iso3_code": "DJI",
    },
    1243: {
        "gaul0_name": "Djibouti",
        "gaul1_name": "Tadjoura",
        "iso2_code": "DJ",
        "iso3_code": "DJI",
    },
    1244: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Al-, Ismailia",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1245: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Alexandria",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1246: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Assiut",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1247: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Aswan",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1248: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Behera",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1249: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Beni Suef",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1250: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Cairo",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1251: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Dakahlia",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1252: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Damietta",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1253: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Fayoum",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1254: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Gharbia",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1255: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Giza",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1256: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Ismailia",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1257: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Kafr El-Shikh",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1258: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Kalyoubia",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1259: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Luxor",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1260: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Matrouh",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1261: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Menia",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1262: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Menoufia",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1263: {
        "gaul0_name": "Egypt",
        "gaul1_name": "New Valley",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1264: {
        "gaul0_name": "Egypt",
        "gaul1_name": "North Sinai",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1265: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Port Said",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1266: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Qena",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1267: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Red Sea",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1268: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Sharkia",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1269: {
        "gaul0_name": "Egypt",
        "gaul1_name": "South Sinai",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1270: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Suez",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1271: {
        "gaul0_name": "Egypt",
        "gaul1_name": "Suhag",
        "iso2_code": "EG",
        "iso3_code": "EGY",
    },
    1272: {
        "gaul0_name": "Equatorial Guinea",
        "gaul1_name": "Annobón",
        "iso2_code": "GQ",
        "iso3_code": "GNQ",
    },
    1273: {
        "gaul0_name": "Equatorial Guinea",
        "gaul1_name": "Bioko Norte",
        "iso2_code": "GQ",
        "iso3_code": "GNQ",
    },
    1274: {
        "gaul0_name": "Equatorial Guinea",
        "gaul1_name": "Bioko Sur",
        "iso2_code": "GQ",
        "iso3_code": "GNQ",
    },
    1275: {
        "gaul0_name": "Equatorial Guinea",
        "gaul1_name": "Centro Sur",
        "iso2_code": "GQ",
        "iso3_code": "GNQ",
    },
    1276: {
        "gaul0_name": "Equatorial Guinea",
        "gaul1_name": "Kié-Ntem",
        "iso2_code": "GQ",
        "iso3_code": "GNQ",
    },
    1277: {
        "gaul0_name": "Equatorial Guinea",
        "gaul1_name": "Litoral",
        "iso2_code": "GQ",
        "iso3_code": "GNQ",
    },
    1278: {
        "gaul0_name": "Equatorial Guinea",
        "gaul1_name": "Wele-Nzas",
        "iso2_code": "GQ",
        "iso3_code": "GNQ",
    },
    1279: {
        "gaul0_name": "Eritrea",
        "gaul1_name": "Anseba",
        "iso2_code": "ER",
        "iso3_code": "ERI",
    },
    1280: {
        "gaul0_name": "Eritrea",
        "gaul1_name": "Debub",
        "iso2_code": "ER",
        "iso3_code": "ERI",
    },
    1281: {
        "gaul0_name": "Eritrea",
        "gaul1_name": "Debubawi Keih Bahri",
        "iso2_code": "ER",
        "iso3_code": "ERI",
    },
    1282: {
        "gaul0_name": "Eritrea",
        "gaul1_name": "Gash Barka",
        "iso2_code": "ER",
        "iso3_code": "ERI",
    },
    1283: {
        "gaul0_name": "Eritrea",
        "gaul1_name": "Maekel",
        "iso2_code": "ER",
        "iso3_code": "ERI",
    },
    1284: {
        "gaul0_name": "Eritrea",
        "gaul1_name": "Semienawi Keih Bahri",
        "iso2_code": "ER",
        "iso3_code": "ERI",
    },
    1285: {
        "gaul0_name": "Eswatini",
        "gaul1_name": "Hhohho",
        "iso2_code": "SZ",
        "iso3_code": "SWZ",
    },
    1286: {
        "gaul0_name": "Eswatini",
        "gaul1_name": "Lubombo",
        "iso2_code": "SZ",
        "iso3_code": "SWZ",
    },
    1287: {
        "gaul0_name": "Eswatini",
        "gaul1_name": "Manzini",
        "iso2_code": "SZ",
        "iso3_code": "SWZ",
    },
    1288: {
        "gaul0_name": "Eswatini",
        "gaul1_name": "Shiselweni",
        "iso2_code": "SZ",
        "iso3_code": "SWZ",
    },
    1289: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Addis Ababa",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1290: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Afar",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1291: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Amhara",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1292: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Benishangul Gumz",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1293: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Dire Dawa",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1294: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Gambela",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1295: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Harari",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1296: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Oromia",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1297: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Sidama",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1298: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Snnp",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1299: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Somali",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1300: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "South West Ethiopia",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1301: {
        "gaul0_name": "Ethiopia",
        "gaul1_name": "Tigray",
        "iso2_code": "ET",
        "iso3_code": "ETH",
    },
    1302: {
        "gaul0_name": "Europa Island",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xFR",
    },
    1303: {
        "gaul0_name": "French Southern Territories",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "TF",
        "iso3_code": "ATF",
    },
    1304: {
        "gaul0_name": "Gabon",
        "gaul1_name": "Estuaire",
        "iso2_code": "GA",
        "iso3_code": "GAB",
    },
    1305: {
        "gaul0_name": "Gabon",
        "gaul1_name": "Haut-Ogooue",
        "iso2_code": "GA",
        "iso3_code": "GAB",
    },
    1306: {
        "gaul0_name": "Gabon",
        "gaul1_name": "Moyen-Ogooue",
        "iso2_code": "GA",
        "iso3_code": "GAB",
    },
    1307: {
        "gaul0_name": "Gabon",
        "gaul1_name": "Ngounie",
        "iso2_code": "GA",
        "iso3_code": "GAB",
    },
    1308: {
        "gaul0_name": "Gabon",
        "gaul1_name": "Nyanga",
        "iso2_code": "GA",
        "iso3_code": "GAB",
    },
    1309: {
        "gaul0_name": "Gabon",
        "gaul1_name": "Ogooue-Ivindo",
        "iso2_code": "GA",
        "iso3_code": "GAB",
    },
    1310: {
        "gaul0_name": "Gabon",
        "gaul1_name": "Ogooue-Lolo",
        "iso2_code": "GA",
        "iso3_code": "GAB",
    },
    1311: {
        "gaul0_name": "Gabon",
        "gaul1_name": "Ogooue-Maritime",
        "iso2_code": "GA",
        "iso3_code": "GAB",
    },
    1312: {
        "gaul0_name": "Gabon",
        "gaul1_name": "Woleu-Ntem",
        "iso2_code": "GA",
        "iso3_code": "GAB",
    },
    1313: {
        "gaul0_name": "Gambia",
        "gaul1_name": "Banjul City Council",
        "iso2_code": "GM",
        "iso3_code": "GMB",
    },
    1314: {
        "gaul0_name": "Gambia",
        "gaul1_name": "Central River North",
        "iso2_code": "GM",
        "iso3_code": "GMB",
    },
    1315: {
        "gaul0_name": "Gambia",
        "gaul1_name": "Central River South",
        "iso2_code": "GM",
        "iso3_code": "GMB",
    },
    1316: {
        "gaul0_name": "Gambia",
        "gaul1_name": "Kanifing Municipal Council",
        "iso2_code": "GM",
        "iso3_code": "GMB",
    },
    1317: {
        "gaul0_name": "Gambia",
        "gaul1_name": "Lower River",
        "iso2_code": "GM",
        "iso3_code": "GMB",
    },
    1318: {
        "gaul0_name": "Gambia",
        "gaul1_name": "North Bank",
        "iso2_code": "GM",
        "iso3_code": "GMB",
    },
    1319: {
        "gaul0_name": "Gambia",
        "gaul1_name": "Upper River",
        "iso2_code": "GM",
        "iso3_code": "GMB",
    },
    1320: {
        "gaul0_name": "Gambia",
        "gaul1_name": "West Coast",
        "iso2_code": "GM",
        "iso3_code": "GMB",
    },
    1321: {
        "gaul0_name": "Ghana",
        "gaul1_name": "Ashanti",
        "iso2_code": "GH",
        "iso3_code": "GHA",
    },
    1322: {
        "gaul0_name": "Ghana",
        "gaul1_name": "Brong Ahafo",
        "iso2_code": "GH",
        "iso3_code": "GHA",
    },
    1323: {
        "gaul0_name": "Ghana",
        "gaul1_name": "Central",
        "iso2_code": "GH",
        "iso3_code": "GHA",
    },
    1324: {
        "gaul0_name": "Ghana",
        "gaul1_name": "Eastern",
        "iso2_code": "GH",
        "iso3_code": "GHA",
    },
    1325: {
        "gaul0_name": "Ghana",
        "gaul1_name": "Greater Accra",
        "iso2_code": "GH",
        "iso3_code": "GHA",
    },
    1326: {
        "gaul0_name": "Ghana",
        "gaul1_name": "Northern",
        "iso2_code": "GH",
        "iso3_code": "GHA",
    },
    1327: {
        "gaul0_name": "Ghana",
        "gaul1_name": "Upper East",
        "iso2_code": "GH",
        "iso3_code": "GHA",
    },
    1328: {
        "gaul0_name": "Ghana",
        "gaul1_name": "Upper West",
        "iso2_code": "GH",
        "iso3_code": "GHA",
    },
    1329: {
        "gaul0_name": "Ghana",
        "gaul1_name": "Volta",
        "iso2_code": "GH",
        "iso3_code": "GHA",
    },
    1330: {
        "gaul0_name": "Ghana",
        "gaul1_name": "Western",
        "iso2_code": "GH",
        "iso3_code": "GHA",
    },
    1331: {
        "gaul0_name": "Glorioso Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xFR",
    },
    1332: {
        "gaul0_name": "Guinea",
        "gaul1_name": "Boke",
        "iso2_code": "GN",
        "iso3_code": "GIN",
    },
    1333: {
        "gaul0_name": "Guinea",
        "gaul1_name": "Conakry",
        "iso2_code": "GN",
        "iso3_code": "GIN",
    },
    1334: {
        "gaul0_name": "Guinea",
        "gaul1_name": "Faranah",
        "iso2_code": "GN",
        "iso3_code": "GIN",
    },
    1335: {
        "gaul0_name": "Guinea",
        "gaul1_name": "Kankan",
        "iso2_code": "GN",
        "iso3_code": "GIN",
    },
    1336: {
        "gaul0_name": "Guinea",
        "gaul1_name": "Kindia",
        "iso2_code": "GN",
        "iso3_code": "GIN",
    },
    1337: {
        "gaul0_name": "Guinea",
        "gaul1_name": "Labe",
        "iso2_code": "GN",
        "iso3_code": "GIN",
    },
    1338: {
        "gaul0_name": "Guinea",
        "gaul1_name": "Mamou",
        "iso2_code": "GN",
        "iso3_code": "GIN",
    },
    1339: {
        "gaul0_name": "Guinea",
        "gaul1_name": "Nzerekore",
        "iso2_code": "GN",
        "iso3_code": "GIN",
    },
    1340: {
        "gaul0_name": "Guinea-Bissau",
        "gaul1_name": "Bafatá",
        "iso2_code": "GW",
        "iso3_code": "GNB",
    },
    1341: {
        "gaul0_name": "Guinea-Bissau",
        "gaul1_name": "Biombo",
        "iso2_code": "GW",
        "iso3_code": "GNB",
    },
    1342: {
        "gaul0_name": "Guinea-Bissau",
        "gaul1_name": "Bolama/Bijagós",
        "iso2_code": "GW",
        "iso3_code": "GNB",
    },
    1343: {
        "gaul0_name": "Guinea-Bissau",
        "gaul1_name": "Cacheu",
        "iso2_code": "GW",
        "iso3_code": "GNB",
    },
    1344: {
        "gaul0_name": "Guinea-Bissau",
        "gaul1_name": "Gabú",
        "iso2_code": "GW",
        "iso3_code": "GNB",
    },
    1345: {
        "gaul0_name": "Guinea-Bissau",
        "gaul1_name": "Oio",
        "iso2_code": "GW",
        "iso3_code": "GNB",
    },
    1346: {
        "gaul0_name": "Guinea-Bissau",
        "gaul1_name": "Quinara",
        "iso2_code": "GW",
        "iso3_code": "GNB",
    },
    1347: {
        "gaul0_name": "Guinea-Bissau",
        "gaul1_name": "Sector Autonomo De Bissau",
        "iso2_code": "GW",
        "iso3_code": "GNB",
    },
    1348: {
        "gaul0_name": "Guinea-Bissau",
        "gaul1_name": "Tombali",
        "iso2_code": "GW",
        "iso3_code": "GNB",
    },
    1349: {
        "gaul0_name": "Hala'Ib Triangle",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xHT",
    },
    1350: {
        "gaul0_name": "Heard Isl. & McDonald Is. (Aust.)",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "HM",
        "iso3_code": "HMD",
    },
    1351: {
        "gaul0_name": "Ilemi Triangle",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xIT",
    },
    1352: {
        "gaul0_name": "Juan De Nova Island",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xFR",
    },
    1353: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Baringo",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1354: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Bomet",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1355: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Bungoma",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1356: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Busia",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1357: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Elgeyo-Marakwet",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1358: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Embu",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1359: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Garissa",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1360: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Homa Bay",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1361: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Isiolo",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1362: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Kajiado",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1363: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Kakamega",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1364: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Kericho",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1365: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Kiambu",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1366: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Kilifi",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1367: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Kirinyaga",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1368: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Kisii",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1369: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Kisumu",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1370: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Kitui",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1371: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Kwale",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1372: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Laikipia",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1373: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Lamu",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1374: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Machakos",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1375: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Makueni",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1376: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Mandera",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1377: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Marsabit",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1378: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Meru",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1379: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Migori",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1380: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Mombasa",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1381: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Murang'A",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1382: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Nairobi",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1383: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Nakuru",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1384: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Nandi",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1385: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Narok",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1386: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Nyamira",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1387: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Nyandarua",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1388: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Nyeri",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1389: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Samburu",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1390: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Siaya",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1391: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Taita Taveta",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1392: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Tana River",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1393: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Tharaka-Nithi",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1394: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Trans Nzoia",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1395: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Turkana",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1396: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Uasin Gishu",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1397: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Vihiga",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1398: {
        "gaul0_name": "Kenya",
        "gaul1_name": "Wajir",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1399: {
        "gaul0_name": "Kenya",
        "gaul1_name": "West Pokot",
        "iso2_code": "KE",
        "iso3_code": "KEN",
    },
    1400: {
        "gaul0_name": "Lesotho",
        "gaul1_name": "Berea",
        "iso2_code": "LS",
        "iso3_code": "LSO",
    },
    1401: {
        "gaul0_name": "Lesotho",
        "gaul1_name": "Butha-Buthe",
        "iso2_code": "LS",
        "iso3_code": "LSO",
    },
    1402: {
        "gaul0_name": "Lesotho",
        "gaul1_name": "Leribe",
        "iso2_code": "LS",
        "iso3_code": "LSO",
    },
    1403: {
        "gaul0_name": "Lesotho",
        "gaul1_name": "Mafeteng",
        "iso2_code": "LS",
        "iso3_code": "LSO",
    },
    1404: {
        "gaul0_name": "Lesotho",
        "gaul1_name": "Maseru",
        "iso2_code": "LS",
        "iso3_code": "LSO",
    },
    1405: {
        "gaul0_name": "Lesotho",
        "gaul1_name": "Mohale'S Hoek",
        "iso2_code": "LS",
        "iso3_code": "LSO",
    },
    1406: {
        "gaul0_name": "Lesotho",
        "gaul1_name": "Mokhotlong",
        "iso2_code": "LS",
        "iso3_code": "LSO",
    },
    1407: {
        "gaul0_name": "Lesotho",
        "gaul1_name": "Qacha'S Nek",
        "iso2_code": "LS",
        "iso3_code": "LSO",
    },
    1408: {
        "gaul0_name": "Lesotho",
        "gaul1_name": "Quthing",
        "iso2_code": "LS",
        "iso3_code": "LSO",
    },
    1409: {
        "gaul0_name": "Lesotho",
        "gaul1_name": "Thaba-Tseka",
        "iso2_code": "LS",
        "iso3_code": "LSO",
    },
    1410: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Bomi",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1411: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Bong",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1412: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Gbarpolu",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1413: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Grand Bassa",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1414: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Grand Cape Mount",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1415: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Grand Gedeh",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1416: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Grand Kru",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1417: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Lofa",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1418: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Margibi",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1419: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Maryland",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1420: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Montserrado",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1421: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Nimba",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1422: {
        "gaul0_name": "Liberia",
        "gaul1_name": "River Gee",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1423: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Rivercess",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1424: {
        "gaul0_name": "Liberia",
        "gaul1_name": "Sinoe",
        "iso2_code": "LR",
        "iso3_code": "LBR",
    },
    1425: {
        "gaul0_name": "Libya",
        "gaul1_name": "East",
        "iso2_code": "LY",
        "iso3_code": "LBY",
    },
    1426: {
        "gaul0_name": "Libya",
        "gaul1_name": "South",
        "iso2_code": "LY",
        "iso3_code": "LBY",
    },
    1427: {
        "gaul0_name": "Libya",
        "gaul1_name": "West",
        "iso2_code": "LY",
        "iso3_code": "LBY",
    },
    1428: {
        "gaul0_name": "Madagascar",
        "gaul1_name": "Antananarivo",
        "iso2_code": "MG",
        "iso3_code": "MDG",
    },
    1429: {
        "gaul0_name": "Madagascar",
        "gaul1_name": "Antsiranana",
        "iso2_code": "MG",
        "iso3_code": "MDG",
    },
    1430: {
        "gaul0_name": "Madagascar",
        "gaul1_name": "Fianarantsoa",
        "iso2_code": "MG",
        "iso3_code": "MDG",
    },
    1431: {
        "gaul0_name": "Madagascar",
        "gaul1_name": "Mahajanga",
        "iso2_code": "MG",
        "iso3_code": "MDG",
    },
    1432: {
        "gaul0_name": "Madagascar",
        "gaul1_name": "Toamasina",
        "iso2_code": "MG",
        "iso3_code": "MDG",
    },
    1433: {
        "gaul0_name": "Madagascar",
        "gaul1_name": "Toliara",
        "iso2_code": "MG",
        "iso3_code": "MDG",
    },
    1434: {
        "gaul0_name": "Malawi",
        "gaul1_name": "Central Region",
        "iso2_code": "MW",
        "iso3_code": "MWI",
    },
    1435: {
        "gaul0_name": "Malawi",
        "gaul1_name": "Northern Region",
        "iso2_code": "MW",
        "iso3_code": "MWI",
    },
    1436: {
        "gaul0_name": "Malawi",
        "gaul1_name": "Southern Region",
        "iso2_code": "MW",
        "iso3_code": "MWI",
    },
    1437: {
        "gaul0_name": "Mali",
        "gaul1_name": "Bamako",
        "iso2_code": "ML",
        "iso3_code": "MLI",
    },
    1438: {
        "gaul0_name": "Mali",
        "gaul1_name": "Gao",
        "iso2_code": "ML",
        "iso3_code": "MLI",
    },
    1439: {
        "gaul0_name": "Mali",
        "gaul1_name": "Kayes",
        "iso2_code": "ML",
        "iso3_code": "MLI",
    },
    1440: {
        "gaul0_name": "Mali",
        "gaul1_name": "Kidal",
        "iso2_code": "ML",
        "iso3_code": "MLI",
    },
    1441: {
        "gaul0_name": "Mali",
        "gaul1_name": "Koulikoro",
        "iso2_code": "ML",
        "iso3_code": "MLI",
    },
    1442: {
        "gaul0_name": "Mali",
        "gaul1_name": "Mopti",
        "iso2_code": "ML",
        "iso3_code": "MLI",
    },
    1443: {
        "gaul0_name": "Mali",
        "gaul1_name": "Sikasso",
        "iso2_code": "ML",
        "iso3_code": "MLI",
    },
    1444: {
        "gaul0_name": "Mali",
        "gaul1_name": "Ségou",
        "iso2_code": "ML",
        "iso3_code": "MLI",
    },
    1445: {
        "gaul0_name": "Mali",
        "gaul1_name": "Tombouctou",
        "iso2_code": "ML",
        "iso3_code": "MLI",
    },
    1446: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Adrar",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1447: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Assaba",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1448: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Brakna",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1449: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Dakhlet-Nouadhibou",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1450: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Gorgol",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1451: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Guidimakha",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1452: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Hodh Ech Chargi",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1453: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Hodh El Gharbi",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1454: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Inchiri",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1455: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Nouakchott-Nord",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1456: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Nouakchott-Ouest",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1457: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Nouakchott-Sud",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1458: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Tagant",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1459: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Tiris-Zemmou",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1460: {
        "gaul0_name": "Mauritania",
        "gaul1_name": "Trarza",
        "iso2_code": "MR",
        "iso3_code": "MRT",
    },
    1461: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Agalnga Islands",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1462: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Cargados Carajos Shoals",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1463: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Flacq",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1464: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Grand Port",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1465: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Moka",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1466: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Ole Rodrigues",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1467: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Pamplemousses",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1468: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Plaines Wilhems",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1469: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Port-Louis",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1470: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Rivinre Du Rempart",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1471: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Rivinre Noire",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1472: {
        "gaul0_name": "Mauritius",
        "gaul1_name": "Savannah",
        "iso2_code": "MU",
        "iso3_code": "MUS",
    },
    1473: {
        "gaul0_name": "Mayotte",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "YT",
        "iso3_code": "MYT",
    },
    1474: {
        "gaul0_name": "Morocco",
        "gaul1_name": "Béni Mellal-Khénifra",
        "iso2_code": "MA",
        "iso3_code": "MAR",
    },
    1475: {
        "gaul0_name": "Morocco",
        "gaul1_name": "Drâa-Tafilalet",
        "iso2_code": "MA",
        "iso3_code": "MAR",
    },
    1476: {
        "gaul0_name": "Morocco",
        "gaul1_name": "Fès-Meknès",
        "iso2_code": "MA",
        "iso3_code": "MAR",
    },
    1477: {
        "gaul0_name": "Morocco",
        "gaul1_name": "Grand Casablanca-Settat",
        "iso2_code": "MA",
        "iso3_code": "MAR",
    },
    1478: {
        "gaul0_name": "Morocco",
        "gaul1_name": "Guelmim-Oued Noun",
        "iso2_code": "MA",
        "iso3_code": "MAR",
    },
    1479: {
        "gaul0_name": "Morocco",
        "gaul1_name": "Marrakech-Safi",
        "iso2_code": "MA",
        "iso3_code": "MAR",
    },
    1480: {
        "gaul0_name": "Morocco",
        "gaul1_name": "Oriental",
        "iso2_code": "MA",
        "iso3_code": "MAR",
    },
    1481: {
        "gaul0_name": "Morocco",
        "gaul1_name": "Rabat-Salé-Kénitra",
        "iso2_code": "MA",
        "iso3_code": "MAR",
    },
    1482: {
        "gaul0_name": "Morocco",
        "gaul1_name": "Souss-Massa",
        "iso2_code": "MA",
        "iso3_code": "MAR",
    },
    1483: {
        "gaul0_name": "Morocco",
        "gaul1_name": "Tanger-Tétouan-Al Hoceima",
        "iso2_code": "MA",
        "iso3_code": "MAR",
    },
    1484: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Cabo Delgado",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1485: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Cidade De Maputo",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1486: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Gaza",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1487: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Inhambane",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1488: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Manica",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1489: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Maputo",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1490: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Nampula",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1491: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Niassa",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1492: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Sofala",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1493: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Tete",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1494: {
        "gaul0_name": "Mozambique",
        "gaul1_name": "Zambézia",
        "iso2_code": "MZ",
        "iso3_code": "MOZ",
    },
    1495: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Erongo",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1496: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Hardap",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1497: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Karas",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1498: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Kavango East",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1499: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Kavango West",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1500: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Khomas",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1501: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Kunene",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1502: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Ohangwena",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1503: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Omaheke",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1504: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Omusati",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1505: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Oshana",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1506: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Oshikoto",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1507: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Otjozondjupa",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1508: {
        "gaul0_name": "Namibia",
        "gaul1_name": "Zambezi",
        "iso2_code": "NA",
        "iso3_code": "NAM",
    },
    1509: {
        "gaul0_name": "Niger",
        "gaul1_name": "Agadez",
        "iso2_code": "NE",
        "iso3_code": "NER",
    },
    1510: {
        "gaul0_name": "Niger",
        "gaul1_name": "Communauté Urbaine De Niamey",
        "iso2_code": "NE",
        "iso3_code": "NER",
    },
    1511: {
        "gaul0_name": "Niger",
        "gaul1_name": "Diffa",
        "iso2_code": "NE",
        "iso3_code": "NER",
    },
    1512: {
        "gaul0_name": "Niger",
        "gaul1_name": "Dosso",
        "iso2_code": "NE",
        "iso3_code": "NER",
    },
    1513: {
        "gaul0_name": "Niger",
        "gaul1_name": "Maradi",
        "iso2_code": "NE",
        "iso3_code": "NER",
    },
    1514: {
        "gaul0_name": "Niger",
        "gaul1_name": "Tahoua",
        "iso2_code": "NE",
        "iso3_code": "NER",
    },
    1515: {
        "gaul0_name": "Niger",
        "gaul1_name": "Tillabéri",
        "iso2_code": "NE",
        "iso3_code": "NER",
    },
    1516: {
        "gaul0_name": "Niger",
        "gaul1_name": "Zinder",
        "iso2_code": "NE",
        "iso3_code": "NER",
    },
    1517: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Abia",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1518: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Adamawa",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1519: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Akwa Lbom",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1520: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Anambra",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1521: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Bauchi",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1522: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Bayelsa",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1523: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Benue",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1524: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Borno",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1525: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Cross River",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1526: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Delta",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1527: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Ebonyi",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1528: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Edo",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1529: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Ekiti",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1530: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Enugu",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1531: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Federal Capital Territory",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1532: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Gombe",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1533: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Imo",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1534: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Jigawa",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1535: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Kaduna",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1536: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Kano",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1537: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Katsina",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1538: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Kebbi",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1539: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Kogi",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1540: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Kwara",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1541: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Lagos",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1542: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Nassarawa",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1543: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Niger",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1544: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Ogun",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1545: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Ondo",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1546: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Osun",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1547: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Oyo",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1548: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Plateau",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1549: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Rivers",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1550: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Sokoto",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1551: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Taraba",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1552: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Yobe",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1553: {
        "gaul0_name": "Nigeria",
        "gaul1_name": "Zamfara",
        "iso2_code": "NG",
        "iso3_code": "NGA",
    },
    1554: {
        "gaul0_name": "Rwanda",
        "gaul1_name": "Amajyaruguru",
        "iso2_code": "RW",
        "iso3_code": "RWA",
    },
    1555: {
        "gaul0_name": "Rwanda",
        "gaul1_name": "Amajyepfo",
        "iso2_code": "RW",
        "iso3_code": "RWA",
    },
    1556: {
        "gaul0_name": "Rwanda",
        "gaul1_name": "Iburasirazuba",
        "iso2_code": "RW",
        "iso3_code": "RWA",
    },
    1557: {
        "gaul0_name": "Rwanda",
        "gaul1_name": "Iburengerazuba",
        "iso2_code": "RW",
        "iso3_code": "RWA",
    },
    1558: {
        "gaul0_name": "Rwanda",
        "gaul1_name": "Umujyi Wa Kigali",
        "iso2_code": "RW",
        "iso3_code": "RWA",
    },
    1559: {
        "gaul0_name": "Réunion",
        "gaul1_name": "Saint Benoit",
        "iso2_code": "RE",
        "iso3_code": "REU",
    },
    1560: {
        "gaul0_name": "Réunion",
        "gaul1_name": "Saint Denis",
        "iso2_code": "RE",
        "iso3_code": "REU",
    },
    1561: {
        "gaul0_name": "Réunion",
        "gaul1_name": "Saint Paul",
        "iso2_code": "RE",
        "iso3_code": "REU",
    },
    1562: {
        "gaul0_name": "Réunion",
        "gaul1_name": "Saint Pierre",
        "iso2_code": "RE",
        "iso3_code": "REU",
    },
    1563: {
        "gaul0_name": "Sao Tome And Principe",
        "gaul1_name": "Príncipe",
        "iso2_code": "ST",
        "iso3_code": "STP",
    },
    1564: {
        "gaul0_name": "Sao Tome And Principe",
        "gaul1_name": "São Tomé",
        "iso2_code": "ST",
        "iso3_code": "STP",
    },
    1565: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Dakar",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1566: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Diourbel",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1567: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Fatick",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1568: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Kaffrine",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1569: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Kaolack",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1570: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Kolda",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1571: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Kédougou",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1572: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Louga",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1573: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Matam",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1574: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Saint Louis",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1575: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Saint-Louis",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1576: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Sédhiou",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1577: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Tambacounda",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1578: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Thiès",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1579: {
        "gaul0_name": "Senegal",
        "gaul1_name": "Ziguinchor",
        "iso2_code": "SN",
        "iso3_code": "SEN",
    },
    1580: {
        "gaul0_name": "Seychelles",
        "gaul1_name": "La Digue",
        "iso2_code": "SC",
        "iso3_code": "SYC",
    },
    1581: {
        "gaul0_name": "Seychelles",
        "gaul1_name": "Mahe",
        "iso2_code": "SC",
        "iso3_code": "SYC",
    },
    1582: {
        "gaul0_name": "Seychelles",
        "gaul1_name": "Other Islands",
        "iso2_code": "SC",
        "iso3_code": "SYC",
    },
    1583: {
        "gaul0_name": "Seychelles",
        "gaul1_name": "Praslin",
        "iso2_code": "SC",
        "iso3_code": "SYC",
    },
    1584: {
        "gaul0_name": "Sierra Leone",
        "gaul1_name": "Eastern",
        "iso2_code": "SL",
        "iso3_code": "SLE",
    },
    1585: {
        "gaul0_name": "Sierra Leone",
        "gaul1_name": "North Western",
        "iso2_code": "SL",
        "iso3_code": "SLE",
    },
    1586: {
        "gaul0_name": "Sierra Leone",
        "gaul1_name": "Northern",
        "iso2_code": "SL",
        "iso3_code": "SLE",
    },
    1587: {
        "gaul0_name": "Sierra Leone",
        "gaul1_name": "Southern",
        "iso2_code": "SL",
        "iso3_code": "SLE",
    },
    1588: {
        "gaul0_name": "Sierra Leone",
        "gaul1_name": "Western",
        "iso2_code": "SL",
        "iso3_code": "SLE",
    },
    1589: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Awdal",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1590: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Bakool",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1591: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Banadir",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1592: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Bari",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1593: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Bay",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1594: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Galgaduud",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1595: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Gedo",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1596: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Hiraan",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1597: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Lower Juba",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1598: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Lower Shabelle",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1599: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Middle Juba",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1600: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Middle Shabelle",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1601: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Mudug",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1602: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Nugaal",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1603: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Sanaag",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1604: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Sool",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1605: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Togdheer",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1606: {
        "gaul0_name": "Somalia",
        "gaul1_name": "Woqooyi Galbeed",
        "iso2_code": "SO",
        "iso3_code": "SOM",
    },
    1607: {
        "gaul0_name": "South Africa",
        "gaul1_name": "Eastern Cape",
        "iso2_code": "ZA",
        "iso3_code": "ZAF",
    },
    1608: {
        "gaul0_name": "South Africa",
        "gaul1_name": "Free State",
        "iso2_code": "ZA",
        "iso3_code": "ZAF",
    },
    1609: {
        "gaul0_name": "South Africa",
        "gaul1_name": "Gauteng",
        "iso2_code": "ZA",
        "iso3_code": "ZAF",
    },
    1610: {
        "gaul0_name": "South Africa",
        "gaul1_name": "Kwa-Zulu Natal",
        "iso2_code": "ZA",
        "iso3_code": "ZAF",
    },
    1611: {
        "gaul0_name": "South Africa",
        "gaul1_name": "Limpopo",
        "iso2_code": "ZA",
        "iso3_code": "ZAF",
    },
    1612: {
        "gaul0_name": "South Africa",
        "gaul1_name": "Mpumalanga",
        "iso2_code": "ZA",
        "iso3_code": "ZAF",
    },
    1613: {
        "gaul0_name": "South Africa",
        "gaul1_name": "North West",
        "iso2_code": "ZA",
        "iso3_code": "ZAF",
    },
    1614: {
        "gaul0_name": "South Africa",
        "gaul1_name": "Northern Cape",
        "iso2_code": "ZA",
        "iso3_code": "ZAF",
    },
    1615: {
        "gaul0_name": "South Africa",
        "gaul1_name": "Wc",
        "iso2_code": "ZA",
        "iso3_code": "ZAF",
    },
    1616: {
        "gaul0_name": "South Africa",
        "gaul1_name": "Western Cape",
        "iso2_code": "ZA",
        "iso3_code": "ZAF",
    },
    1617: {
        "gaul0_name": "South Sudan",
        "gaul1_name": "Central Equatoria",
        "iso2_code": "SS",
        "iso3_code": "SSD",
    },
    1618: {
        "gaul0_name": "South Sudan",
        "gaul1_name": "Eastern Equatoria",
        "iso2_code": "SS",
        "iso3_code": "SSD",
    },
    1619: {
        "gaul0_name": "South Sudan",
        "gaul1_name": "Jonglei",
        "iso2_code": "SS",
        "iso3_code": "SSD",
    },
    1620: {
        "gaul0_name": "South Sudan",
        "gaul1_name": "Lakes",
        "iso2_code": "SS",
        "iso3_code": "SSD",
    },
    1621: {
        "gaul0_name": "South Sudan",
        "gaul1_name": "Northern Bahr El Ghazal",
        "iso2_code": "SS",
        "iso3_code": "SSD",
    },
    1622: {
        "gaul0_name": "South Sudan",
        "gaul1_name": "Unity",
        "iso2_code": "SS",
        "iso3_code": "SSD",
    },
    1623: {
        "gaul0_name": "South Sudan",
        "gaul1_name": "Upper Nile",
        "iso2_code": "SS",
        "iso3_code": "SSD",
    },
    1624: {
        "gaul0_name": "South Sudan",
        "gaul1_name": "Warrap",
        "iso2_code": "SS",
        "iso3_code": "SSD",
    },
    1625: {
        "gaul0_name": "South Sudan",
        "gaul1_name": "Western Bahr El Ghazal",
        "iso2_code": "SS",
        "iso3_code": "SSD",
    },
    1626: {
        "gaul0_name": "South Sudan",
        "gaul1_name": "Western Equatoria",
        "iso2_code": "SS",
        "iso3_code": "SSD",
    },
    1627: {
        "gaul0_name": "Sudan",
        "gaul1_name": "Aj Jazirah",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1628: {
        "gaul0_name": "Sudan",
        "gaul1_name": "Blue Nile",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1629: {
        "gaul0_name": "Sudan",
        "gaul1_name": "Central Darfur",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1630: {
        "gaul0_name": "Sudan",
        "gaul1_name": "East Darfur",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1631: {
        "gaul0_name": "Sudan",
        "gaul1_name": "Gedaref",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1632: {
        "gaul0_name": "Sudan",
        "gaul1_name": "Kassala",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1633: {
        "gaul0_name": "Sudan",
        "gaul1_name": "Khartoum",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1634: {
        "gaul0_name": "Sudan",
        "gaul1_name": "North Darfur",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1635: {
        "gaul0_name": "Sudan",
        "gaul1_name": "North Kordofan",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1636: {
        "gaul0_name": "Sudan",
        "gaul1_name": "Northern",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1637: {
        "gaul0_name": "Sudan",
        "gaul1_name": "Red Sea",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1638: {
        "gaul0_name": "Sudan",
        "gaul1_name": "River Nile",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1639: {
        "gaul0_name": "Sudan",
        "gaul1_name": "Sennar",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1640: {
        "gaul0_name": "Sudan",
        "gaul1_name": "South Darfur",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1641: {
        "gaul0_name": "Sudan",
        "gaul1_name": "South Kordofan",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1642: {
        "gaul0_name": "Sudan",
        "gaul1_name": "West Darfur",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1643: {
        "gaul0_name": "Sudan",
        "gaul1_name": "West Kordofan",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1644: {
        "gaul0_name": "Sudan",
        "gaul1_name": "White Nile",
        "iso2_code": "SD",
        "iso3_code": "SDN",
    },
    1645: {
        "gaul0_name": "Togo",
        "gaul1_name": "Centrale",
        "iso2_code": "TG",
        "iso3_code": "TGO",
    },
    1646: {
        "gaul0_name": "Togo",
        "gaul1_name": "Kara",
        "iso2_code": "TG",
        "iso3_code": "TGO",
    },
    1647: {
        "gaul0_name": "Togo",
        "gaul1_name": "Maritime",
        "iso2_code": "TG",
        "iso3_code": "TGO",
    },
    1648: {
        "gaul0_name": "Togo",
        "gaul1_name": "Plateaux",
        "iso2_code": "TG",
        "iso3_code": "TGO",
    },
    1649: {
        "gaul0_name": "Togo",
        "gaul1_name": "Savanes",
        "iso2_code": "TG",
        "iso3_code": "TGO",
    },
    1650: {
        "gaul0_name": "Tromelin Island",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xFR",
    },
    1651: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Ariana",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1652: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Beja",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1653: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Ben Arous",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1654: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Bizerte",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1655: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Gabes",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1656: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Gafsa",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1657: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Jendouba",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1658: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Kairouan",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1659: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Kasserine",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1660: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Kebili",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1661: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Le Kef",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1662: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Mahdia",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1663: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Manouba",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1664: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Mednine",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1665: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Monastir",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1666: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Nabeul",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1667: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Sfax",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1668: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Sidi Bouzid",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1669: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Siliana",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1670: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Sousse",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1671: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Tataouine",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1672: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Tozeur",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1673: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Tunis",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1674: {
        "gaul0_name": "Tunisia",
        "gaul1_name": "Zaghouan",
        "iso2_code": "TN",
        "iso3_code": "TUN",
    },
    1675: {
        "gaul0_name": "Uganda",
        "gaul1_name": "Central",
        "iso2_code": "UG",
        "iso3_code": "UGA",
    },
    1676: {
        "gaul0_name": "Uganda",
        "gaul1_name": "Eastern",
        "iso2_code": "UG",
        "iso3_code": "UGA",
    },
    1677: {
        "gaul0_name": "Uganda",
        "gaul1_name": "Northern",
        "iso2_code": "UG",
        "iso3_code": "UGA",
    },
    1678: {
        "gaul0_name": "Uganda",
        "gaul1_name": "Western",
        "iso2_code": "UG",
        "iso3_code": "UGA",
    },
    1679: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Arusha",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1680: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Dar Es Salaam",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1681: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Dodoma",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1682: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Geita",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1683: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Iringa",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1684: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Kagera",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1685: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Kaskazini Pemba",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1686: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Kaskazini Unguja",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1687: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Katavi",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1688: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Kigoma",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1689: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Kilimanjaro",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1690: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Kusini Pemba",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1691: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Kusini Unguja",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1692: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Lindi",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1693: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Manyara",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1694: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Mara",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1695: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Mbeya",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1696: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Mjini Magharibi",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1697: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Morogoro",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1698: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Mtwara",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1699: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Mwanza",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1700: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Njombe",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1701: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Pwani",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1702: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Rukwa",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1703: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Ruvuma",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1704: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Shinyanga",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1705: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Simiyu",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1706: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Singida",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1707: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Songwe",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1708: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Tabora",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1709: {
        "gaul0_name": "United Republic of Tanzania",
        "gaul1_name": "Tanga",
        "iso2_code": "TZ",
        "iso3_code": "TZA",
    },
    1710: {
        "gaul0_name": "Western Sahara",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "EH",
        "iso3_code": "ESH",
    },
    1711: {
        "gaul0_name": "Zambia",
        "gaul1_name": "Central",
        "iso2_code": "ZM",
        "iso3_code": "ZMB",
    },
    1712: {
        "gaul0_name": "Zambia",
        "gaul1_name": "Copperbelt",
        "iso2_code": "ZM",
        "iso3_code": "ZMB",
    },
    1713: {
        "gaul0_name": "Zambia",
        "gaul1_name": "Eastern",
        "iso2_code": "ZM",
        "iso3_code": "ZMB",
    },
    1714: {
        "gaul0_name": "Zambia",
        "gaul1_name": "Luapula",
        "iso2_code": "ZM",
        "iso3_code": "ZMB",
    },
    1715: {
        "gaul0_name": "Zambia",
        "gaul1_name": "Lusaka",
        "iso2_code": "ZM",
        "iso3_code": "ZMB",
    },
    1716: {
        "gaul0_name": "Zambia",
        "gaul1_name": "Muchiga",
        "iso2_code": "ZM",
        "iso3_code": "ZMB",
    },
    1717: {
        "gaul0_name": "Zambia",
        "gaul1_name": "North-Western",
        "iso2_code": "ZM",
        "iso3_code": "ZMB",
    },
    1718: {
        "gaul0_name": "Zambia",
        "gaul1_name": "Northern",
        "iso2_code": "ZM",
        "iso3_code": "ZMB",
    },
    1719: {
        "gaul0_name": "Zambia",
        "gaul1_name": "Southern",
        "iso2_code": "ZM",
        "iso3_code": "ZMB",
    },
    1720: {
        "gaul0_name": "Zambia",
        "gaul1_name": "Western",
        "iso2_code": "ZM",
        "iso3_code": "ZMB",
    },
    1721: {
        "gaul0_name": "Zimbabwe",
        "gaul1_name": "Bulawayo",
        "iso2_code": "ZW",
        "iso3_code": "ZWE",
    },
    1722: {
        "gaul0_name": "Zimbabwe",
        "gaul1_name": "Harare",
        "iso2_code": "ZW",
        "iso3_code": "ZWE",
    },
    1723: {
        "gaul0_name": "Zimbabwe",
        "gaul1_name": "Manicaland",
        "iso2_code": "ZW",
        "iso3_code": "ZWE",
    },
    1724: {
        "gaul0_name": "Zimbabwe",
        "gaul1_name": "Mashonaland Central",
        "iso2_code": "ZW",
        "iso3_code": "ZWE",
    },
    1725: {
        "gaul0_name": "Zimbabwe",
        "gaul1_name": "Mashonaland East",
        "iso2_code": "ZW",
        "iso3_code": "ZWE",
    },
    1726: {
        "gaul0_name": "Zimbabwe",
        "gaul1_name": "Mashonaland West",
        "iso2_code": "ZW",
        "iso3_code": "ZWE",
    },
    1727: {
        "gaul0_name": "Zimbabwe",
        "gaul1_name": "Masvingo",
        "iso2_code": "ZW",
        "iso3_code": "ZWE",
    },
    1728: {
        "gaul0_name": "Zimbabwe",
        "gaul1_name": "Matebeleland North",
        "iso2_code": "ZW",
        "iso3_code": "ZWE",
    },
    1729: {
        "gaul0_name": "Zimbabwe",
        "gaul1_name": "Matebeleland South",
        "iso2_code": "ZW",
        "iso3_code": "ZWE",
    },
    1730: {
        "gaul0_name": "Zimbabwe",
        "gaul1_name": "Midlands",
        "iso2_code": "ZW",
        "iso3_code": "ZWE",
    },
    1731: {
        "gaul0_name": "Anguilla",
        "gaul1_name": "Blowing Point",
        "iso2_code": "AI",
        "iso3_code": "AIA",
    },
    1732: {
        "gaul0_name": "Anguilla",
        "gaul1_name": "George Hill",
        "iso2_code": "AI",
        "iso3_code": "AIA",
    },
    1733: {
        "gaul0_name": "Anguilla",
        "gaul1_name": "Island Harbour",
        "iso2_code": "AI",
        "iso3_code": "AIA",
    },
    1734: {
        "gaul0_name": "Anguilla",
        "gaul1_name": "North Side",
        "iso2_code": "AI",
        "iso3_code": "AIA",
    },
    1735: {
        "gaul0_name": "Anguilla",
        "gaul1_name": "Sandy Hill",
        "iso2_code": "AI",
        "iso3_code": "AIA",
    },
    1736: {
        "gaul0_name": "Anguilla",
        "gaul1_name": "Sombrero Island",
        "iso2_code": "AI",
        "iso3_code": "AIA",
    },
    1737: {
        "gaul0_name": "Anguilla",
        "gaul1_name": "South Hill",
        "iso2_code": "AI",
        "iso3_code": "AIA",
    },
    1738: {
        "gaul0_name": "Anguilla",
        "gaul1_name": "The Farrington",
        "iso2_code": "AI",
        "iso3_code": "AIA",
    },
    1739: {
        "gaul0_name": "Anguilla",
        "gaul1_name": "West End Including Long Bay",
        "iso2_code": "AI",
        "iso3_code": "AIA",
    },
    1740: {
        "gaul0_name": "Antigua And Barbuda",
        "gaul1_name": "Barbuda",
        "iso2_code": "AG",
        "iso3_code": "ATG",
    },
    1741: {
        "gaul0_name": "Antigua And Barbuda",
        "gaul1_name": "Redonda",
        "iso2_code": "AG",
        "iso3_code": "ATG",
    },
    1742: {
        "gaul0_name": "Antigua And Barbuda",
        "gaul1_name": "Saint George",
        "iso2_code": "AG",
        "iso3_code": "ATG",
    },
    1743: {
        "gaul0_name": "Antigua And Barbuda",
        "gaul1_name": "Saint John'S",
        "iso2_code": "AG",
        "iso3_code": "ATG",
    },
    1744: {
        "gaul0_name": "Antigua And Barbuda",
        "gaul1_name": "Saint Mary",
        "iso2_code": "AG",
        "iso3_code": "ATG",
    },
    1745: {
        "gaul0_name": "Antigua And Barbuda",
        "gaul1_name": "Saint Paul",
        "iso2_code": "AG",
        "iso3_code": "ATG",
    },
    1746: {
        "gaul0_name": "Antigua And Barbuda",
        "gaul1_name": "Saint Peter",
        "iso2_code": "AG",
        "iso3_code": "ATG",
    },
    1747: {
        "gaul0_name": "Antigua And Barbuda",
        "gaul1_name": "Saint Philip",
        "iso2_code": "AG",
        "iso3_code": "ATG",
    },
    1748: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Buenos Aires",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1749: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Catamarca",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1750: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Chaco",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1751: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Chubut",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1752: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Ciudad De Buenos Aires",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1753: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Corrientes",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1754: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Córdoba",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1755: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Entre Ríos",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1756: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Formosa",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1757: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Jujuy",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1758: {
        "gaul0_name": "Argentina",
        "gaul1_name": "La Pampa",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1759: {
        "gaul0_name": "Argentina",
        "gaul1_name": "La Rioja",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1760: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Mendoza",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1761: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Misiones",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1762: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Neuquén",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1763: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Río Negro",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1764: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Salta",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1765: {
        "gaul0_name": "Argentina",
        "gaul1_name": "San Juan",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1766: {
        "gaul0_name": "Argentina",
        "gaul1_name": "San Luis",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1767: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Santa Cruz",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1768: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Santa Fe",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1769: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Santiago Del Estero",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1770: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Tierra Del Fuego",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1771: {
        "gaul0_name": "Argentina",
        "gaul1_name": "Tucumán",
        "iso2_code": "AR",
        "iso3_code": "ARG",
    },
    1772: {
        "gaul0_name": "Aruba",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "AW",
        "iso3_code": "ABW",
    },
    1773: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Acklins",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1774: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Berry Islands",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1775: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Biminis",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1776: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Black Point",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1777: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Cat Island",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1778: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Central Abaco",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1779: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Central Andros",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1780: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Central Eleuthera",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1781: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "City Of Freeport",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1782: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Crooked Island",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1783: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "East Grand Bahama",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1784: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Exuma",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1785: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Grand Cay",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1786: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Harbour Island",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1787: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Hope Town",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1788: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Inagua",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1789: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Long Island",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1790: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Mangrove Cay",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1791: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Mayaguana",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1792: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Moore'S Island",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1793: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "New Providence",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1794: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "North Abaco",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1795: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "North Andros",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1796: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "North Eleuthera",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1797: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Ragged Island",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1798: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Rum Cay",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1799: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "San Salvador",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1800: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "South Abaco",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1801: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "South Andros",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1802: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "South Eleuthera",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1803: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "Spanish Wells",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1804: {
        "gaul0_name": "Bahamas",
        "gaul1_name": "West Grand Bahama",
        "iso2_code": "BS",
        "iso3_code": "BHS",
    },
    1805: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Christ Church",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1806: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Saint Andrew",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1807: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Saint George",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1808: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Saint James",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1809: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Saint John",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1810: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Saint Joseph",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1811: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Saint Lucy",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1812: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Saint Michael",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1813: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Saint Peter",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1814: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Saint Philip",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1815: {
        "gaul0_name": "Barbados",
        "gaul1_name": "Saint Thomas",
        "iso2_code": "BB",
        "iso3_code": "BRB",
    },
    1816: {
        "gaul0_name": "Belize",
        "gaul1_name": "Belize",
        "iso2_code": "BZ",
        "iso3_code": "BLZ",
    },
    1817: {
        "gaul0_name": "Belize",
        "gaul1_name": "Cayo",
        "iso2_code": "BZ",
        "iso3_code": "BLZ",
    },
    1818: {
        "gaul0_name": "Belize",
        "gaul1_name": "Corozal",
        "iso2_code": "BZ",
        "iso3_code": "BLZ",
    },
    1819: {
        "gaul0_name": "Belize",
        "gaul1_name": "Orange Walk",
        "iso2_code": "BZ",
        "iso3_code": "BLZ",
    },
    1820: {
        "gaul0_name": "Belize",
        "gaul1_name": "Stann Creek",
        "iso2_code": "BZ",
        "iso3_code": "BLZ",
    },
    1821: {
        "gaul0_name": "Belize",
        "gaul1_name": "Toledo",
        "iso2_code": "BZ",
        "iso3_code": "BLZ",
    },
    1822: {
        "gaul0_name": "Bermuda",
        "gaul1_name": "Devonshire",
        "iso2_code": "BM",
        "iso3_code": "BMU",
    },
    1823: {
        "gaul0_name": "Bermuda",
        "gaul1_name": "Hamilton",
        "iso2_code": "BM",
        "iso3_code": "BMU",
    },
    1824: {
        "gaul0_name": "Bermuda",
        "gaul1_name": "Paget",
        "iso2_code": "BM",
        "iso3_code": "BMU",
    },
    1825: {
        "gaul0_name": "Bermuda",
        "gaul1_name": "Pembroke",
        "iso2_code": "BM",
        "iso3_code": "BMU",
    },
    1826: {
        "gaul0_name": "Bermuda",
        "gaul1_name": "Saint George'S",
        "iso2_code": "BM",
        "iso3_code": "BMU",
    },
    1827: {
        "gaul0_name": "Bermuda",
        "gaul1_name": "Sandys",
        "iso2_code": "BM",
        "iso3_code": "BMU",
    },
    1828: {
        "gaul0_name": "Bermuda",
        "gaul1_name": "Smiths",
        "iso2_code": "BM",
        "iso3_code": "BMU",
    },
    1829: {
        "gaul0_name": "Bermuda",
        "gaul1_name": "Southampton",
        "iso2_code": "BM",
        "iso3_code": "BMU",
    },
    1830: {
        "gaul0_name": "Bermuda",
        "gaul1_name": "Warwick",
        "iso2_code": "BM",
        "iso3_code": "BMU",
    },
    1831: {
        "gaul0_name": "Bolivia (Plurinational State of)",
        "gaul1_name": "Beni",
        "iso2_code": "BO",
        "iso3_code": "BOL",
    },
    1832: {
        "gaul0_name": "Bolivia (Plurinational State of)",
        "gaul1_name": "Chuquisaca",
        "iso2_code": "BO",
        "iso3_code": "BOL",
    },
    1833: {
        "gaul0_name": "Bolivia (Plurinational State of)",
        "gaul1_name": "Cochabamba",
        "iso2_code": "BO",
        "iso3_code": "BOL",
    },
    1834: {
        "gaul0_name": "Bolivia (Plurinational State of)",
        "gaul1_name": "La Paz",
        "iso2_code": "BO",
        "iso3_code": "BOL",
    },
    1835: {
        "gaul0_name": "Bolivia (Plurinational State of)",
        "gaul1_name": "Oruro",
        "iso2_code": "BO",
        "iso3_code": "BOL",
    },
    1836: {
        "gaul0_name": "Bolivia (Plurinational State of)",
        "gaul1_name": "Pando",
        "iso2_code": "BO",
        "iso3_code": "BOL",
    },
    1837: {
        "gaul0_name": "Bolivia (Plurinational State of)",
        "gaul1_name": "Potosí",
        "iso2_code": "BO",
        "iso3_code": "BOL",
    },
    1838: {
        "gaul0_name": "Bolivia (Plurinational State of)",
        "gaul1_name": "Santa Cruz",
        "iso2_code": "BO",
        "iso3_code": "BOL",
    },
    1839: {
        "gaul0_name": "Bolivia (Plurinational State of)",
        "gaul1_name": "Tarija",
        "iso2_code": "BO",
        "iso3_code": "BOL",
    },
    1840: {
        "gaul0_name": "Bonaire, Sint Eustatius And Saba",
        "gaul1_name": "Bonaire",
        "iso2_code": "BQ",
        "iso3_code": "BES",
    },
    1841: {
        "gaul0_name": "Bonaire, Sint Eustatius And Saba",
        "gaul1_name": "Saba",
        "iso2_code": "BQ",
        "iso3_code": "BES",
    },
    1842: {
        "gaul0_name": "Bonaire, Sint Eustatius And Saba",
        "gaul1_name": "Sint Eustatius",
        "iso2_code": "BQ",
        "iso3_code": "BES",
    },
    1843: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Acre",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1844: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Alagoas",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1845: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Amapá",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1846: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Amazonas",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1847: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Bahia",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1848: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Ceará",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1849: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Distrito Federal",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1850: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Espírito Santo",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1851: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Goiás",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1852: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Maranhão",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1853: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Mato Grosso",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1854: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Mato Grosso Do Sul",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1855: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Minas Gerais",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1856: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Paraná",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1857: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Paraíba",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1858: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Pará",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1859: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Pernambuco",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1860: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Piauí",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1861: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Rio De Janeiro",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1862: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Rio Grande Do Norte",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1863: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Rio Grande Do Sul",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1864: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Rondônia",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1865: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Roraima",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1866: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Santa Catarina",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1867: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Sergipe",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1868: {
        "gaul0_name": "Brazil",
        "gaul1_name": "São Paulo",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1869: {
        "gaul0_name": "Brazil",
        "gaul1_name": "Tocantins",
        "iso2_code": "BR",
        "iso3_code": "BRA",
    },
    1870: {
        "gaul0_name": "British Virgin Islands",
        "gaul1_name": "Anegada",
        "iso2_code": "VG",
        "iso3_code": "VGB",
    },
    1871: {
        "gaul0_name": "British Virgin Islands",
        "gaul1_name": "Cooper Island",
        "iso2_code": "VG",
        "iso3_code": "VGB",
    },
    1872: {
        "gaul0_name": "British Virgin Islands",
        "gaul1_name": "Great Camanoe",
        "iso2_code": "VG",
        "iso3_code": "VGB",
    },
    1873: {
        "gaul0_name": "British Virgin Islands",
        "gaul1_name": "Jost Van Dyke",
        "iso2_code": "VG",
        "iso3_code": "VGB",
    },
    1874: {
        "gaul0_name": "British Virgin Islands",
        "gaul1_name": "Other Islands",
        "iso2_code": "VG",
        "iso3_code": "VGB",
    },
    1875: {
        "gaul0_name": "British Virgin Islands",
        "gaul1_name": "Tortola",
        "iso2_code": "VG",
        "iso3_code": "VGB",
    },
    1876: {
        "gaul0_name": "British Virgin Islands",
        "gaul1_name": "Virgin Gorda",
        "iso2_code": "VG",
        "iso3_code": "VGB",
    },
    1877: {
        "gaul0_name": "Canada",
        "gaul1_name": "Alberta",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1878: {
        "gaul0_name": "Canada",
        "gaul1_name": "British Columbia / Colombie-Britannique",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1879: {
        "gaul0_name": "Canada",
        "gaul1_name": "Manitoba",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1880: {
        "gaul0_name": "Canada",
        "gaul1_name": "New Brunswick / Nouveau-Brunswick",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1881: {
        "gaul0_name": "Canada",
        "gaul1_name": "Newfoundland And Labrador / Terre-Neuve-Et-Labrador",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1882: {
        "gaul0_name": "Canada",
        "gaul1_name": "Northwest Territories / Territoires Du Nord-Ouest",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1883: {
        "gaul0_name": "Canada",
        "gaul1_name": "Nova Scotia / Nouvelle-Écosse",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1884: {
        "gaul0_name": "Canada",
        "gaul1_name": "Nunavut",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1885: {
        "gaul0_name": "Canada",
        "gaul1_name": "Ontario",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1886: {
        "gaul0_name": "Canada",
        "gaul1_name": "Prince Edward Island / Île-Du-Prince-Édouard",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1887: {
        "gaul0_name": "Canada",
        "gaul1_name": "Quebec / Québec",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1888: {
        "gaul0_name": "Canada",
        "gaul1_name": "Saskatchewan",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1889: {
        "gaul0_name": "Canada",
        "gaul1_name": "Waterbody",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1890: {
        "gaul0_name": "Canada",
        "gaul1_name": "Yukon",
        "iso2_code": "CA",
        "iso3_code": "CAN",
    },
    1891: {
        "gaul0_name": "Cayman Islands",
        "gaul1_name": "Bodden Town",
        "iso2_code": "KY",
        "iso3_code": "CYM",
    },
    1892: {
        "gaul0_name": "Cayman Islands",
        "gaul1_name": "East End",
        "iso2_code": "KY",
        "iso3_code": "CYM",
    },
    1893: {
        "gaul0_name": "Cayman Islands",
        "gaul1_name": "George Town",
        "iso2_code": "KY",
        "iso3_code": "CYM",
    },
    1894: {
        "gaul0_name": "Cayman Islands",
        "gaul1_name": "North Side",
        "iso2_code": "KY",
        "iso3_code": "CYM",
    },
    1895: {
        "gaul0_name": "Cayman Islands",
        "gaul1_name": "Sister Islands",
        "iso2_code": "KY",
        "iso3_code": "CYM",
    },
    1896: {
        "gaul0_name": "Cayman Islands",
        "gaul1_name": "West Bay",
        "iso2_code": "KY",
        "iso3_code": "CYM",
    },
    1897: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Antofagasta",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1898: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Arica Y Parinacota",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1899: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Atacama",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1900: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Aysén Del Gral. Ibañez Del Campo",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1901: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Coquimbo",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1902: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De La Araucanía",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1903: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Los Lagos",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1904: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Los Ríos",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1905: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Magallanes Y Antártica Chilena",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1906: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Tarapacá",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1907: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Valparaíso",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1908: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región De Ñuble",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1909: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región Del Bío-Bío",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1910: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región Del Libertador Bernardo O'Higgins",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1911: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región Del Maule",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1912: {
        "gaul0_name": "Chile",
        "gaul1_name": "Región Metropolitana De Santiago",
        "iso2_code": "CL",
        "iso3_code": "CHL",
    },
    1913: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Amazonas",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1914: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Antioquia",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1915: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Arauca",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1916: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Atlantico",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1917: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Bolivar",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1918: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Boyaca",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1919: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Caldas",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1920: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Caqueta",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1921: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Casanare",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1922: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Cauca",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1923: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Cesar",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1924: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Choco",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1925: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Cordoba",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1926: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Cundinamarca",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1927: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Guainia",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1928: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Guaviare",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1929: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Huila",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1930: {
        "gaul0_name": "Colombia",
        "gaul1_name": "La Guajira",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1931: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Magdalena",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1932: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Meta",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1933: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Narino",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1934: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Norte De Santander",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1935: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Not Available",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1936: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Putumayo",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1937: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Quindio",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1938: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Risaralda",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1939: {
        "gaul0_name": "Colombia",
        "gaul1_name": "San Andres Y Providencia",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1940: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Santander",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1941: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Sucre",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1942: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Tolima",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1943: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Valle Del Cauca",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1944: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Vaupes",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1945: {
        "gaul0_name": "Colombia",
        "gaul1_name": "Vichada",
        "iso2_code": "CO",
        "iso3_code": "COL",
    },
    1946: {
        "gaul0_name": "Costa Rica",
        "gaul1_name": "Alajuela",
        "iso2_code": "CR",
        "iso3_code": "CRI",
    },
    1947: {
        "gaul0_name": "Costa Rica",
        "gaul1_name": "Cartago",
        "iso2_code": "CR",
        "iso3_code": "CRI",
    },
    1948: {
        "gaul0_name": "Costa Rica",
        "gaul1_name": "Guanacaste",
        "iso2_code": "CR",
        "iso3_code": "CRI",
    },
    1949: {
        "gaul0_name": "Costa Rica",
        "gaul1_name": "Heredia",
        "iso2_code": "CR",
        "iso3_code": "CRI",
    },
    1950: {
        "gaul0_name": "Costa Rica",
        "gaul1_name": "Limón",
        "iso2_code": "CR",
        "iso3_code": "CRI",
    },
    1951: {
        "gaul0_name": "Costa Rica",
        "gaul1_name": "Puntarenas",
        "iso2_code": "CR",
        "iso3_code": "CRI",
    },
    1952: {
        "gaul0_name": "Costa Rica",
        "gaul1_name": "San José",
        "iso2_code": "CR",
        "iso3_code": "CRI",
    },
    1953: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Artemisa",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1954: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Camagüey",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1955: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Ciego De Ávila",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1956: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Cienfuegos",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1957: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Granma",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1958: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Guantánamo",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1959: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Holguín",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1960: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Isla De La Juventud",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1961: {
        "gaul0_name": "Cuba",
        "gaul1_name": "La Habana",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1962: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Las Tunas",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1963: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Matanzas",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1964: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Mayabeque",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1965: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Pinar Del Río",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1966: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Sancti Spíritus",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1967: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Santiago De Cuba",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1968: {
        "gaul0_name": "Cuba",
        "gaul1_name": "Villa Clara",
        "iso2_code": "CU",
        "iso3_code": "CUB",
    },
    1969: {
        "gaul0_name": "Dominica",
        "gaul1_name": "St. Andrew",
        "iso2_code": "DM",
        "iso3_code": "DMA",
    },
    1970: {
        "gaul0_name": "Dominica",
        "gaul1_name": "St. David",
        "iso2_code": "DM",
        "iso3_code": "DMA",
    },
    1971: {
        "gaul0_name": "Dominica",
        "gaul1_name": "St. George",
        "iso2_code": "DM",
        "iso3_code": "DMA",
    },
    1972: {
        "gaul0_name": "Dominica",
        "gaul1_name": "St. John",
        "iso2_code": "DM",
        "iso3_code": "DMA",
    },
    1973: {
        "gaul0_name": "Dominica",
        "gaul1_name": "St. Joseph",
        "iso2_code": "DM",
        "iso3_code": "DMA",
    },
    1974: {
        "gaul0_name": "Dominica",
        "gaul1_name": "St. Luke",
        "iso2_code": "DM",
        "iso3_code": "DMA",
    },
    1975: {
        "gaul0_name": "Dominica",
        "gaul1_name": "St. Mark",
        "iso2_code": "DM",
        "iso3_code": "DMA",
    },
    1976: {
        "gaul0_name": "Dominica",
        "gaul1_name": "St. Patrick",
        "iso2_code": "DM",
        "iso3_code": "DMA",
    },
    1977: {
        "gaul0_name": "Dominica",
        "gaul1_name": "St. Paul",
        "iso2_code": "DM",
        "iso3_code": "DMA",
    },
    1978: {
        "gaul0_name": "Dominica",
        "gaul1_name": "St. Peter",
        "iso2_code": "DM",
        "iso3_code": "DMA",
    },
    1979: {
        "gaul0_name": "Dominican Republic",
        "gaul1_name": "Cibao Nordeste",
        "iso2_code": "DO",
        "iso3_code": "DOM",
    },
    1980: {
        "gaul0_name": "Dominican Republic",
        "gaul1_name": "Cibao Noroeste",
        "iso2_code": "DO",
        "iso3_code": "DOM",
    },
    1981: {
        "gaul0_name": "Dominican Republic",
        "gaul1_name": "Cibao Norte",
        "iso2_code": "DO",
        "iso3_code": "DOM",
    },
    1982: {
        "gaul0_name": "Dominican Republic",
        "gaul1_name": "Cibao Sur",
        "iso2_code": "DO",
        "iso3_code": "DOM",
    },
    1983: {
        "gaul0_name": "Dominican Republic",
        "gaul1_name": "El Valle",
        "iso2_code": "DO",
        "iso3_code": "DOM",
    },
    1984: {
        "gaul0_name": "Dominican Republic",
        "gaul1_name": "Enriquillo",
        "iso2_code": "DO",
        "iso3_code": "DOM",
    },
    1985: {
        "gaul0_name": "Dominican Republic",
        "gaul1_name": "Higuamo",
        "iso2_code": "DO",
        "iso3_code": "DOM",
    },
    1986: {
        "gaul0_name": "Dominican Republic",
        "gaul1_name": "Ozama",
        "iso2_code": "DO",
        "iso3_code": "DOM",
    },
    1987: {
        "gaul0_name": "Dominican Republic",
        "gaul1_name": "Valdesia",
        "iso2_code": "DO",
        "iso3_code": "DOM",
    },
    1988: {
        "gaul0_name": "Dominican Republic",
        "gaul1_name": "Yuma",
        "iso2_code": "DO",
        "iso3_code": "DOM",
    },
    1989: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Azuay",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    1990: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Bolivar",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    1991: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Canar",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    1992: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Carchi",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    1993: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Chimborazo",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    1994: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Cotopaxi",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    1995: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "El Oro",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    1996: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Emeralds",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    1997: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Galapagos",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    1998: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Guayas",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    1999: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Imbabura",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2000: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Loja",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2001: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Los Rios",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2002: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Manabi",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2003: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Morona Santiago",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2004: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Napo",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2005: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Orellana",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2006: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Pastaza",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2007: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Pichincha",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2008: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Santa Elena",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2009: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Santo Domingo De Los Tsachilas",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2010: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Sucumbios",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2011: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Tungurahua",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2012: {
        "gaul0_name": "Ecuador",
        "gaul1_name": "Zamora Chinchipe",
        "iso2_code": "EC",
        "iso3_code": "ECU",
    },
    2013: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "Ahuachapán",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2014: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "Cabañas",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2015: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "Chalatenango",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2016: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "Cuscatlán",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2017: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "La Libertad",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2018: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "La Paz",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2019: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "La Unión",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2020: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "Morazán",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2021: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "San Miguel",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2022: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "San Salvador",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2023: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "San Vicente",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2024: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "Santa Ana",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2025: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "Sonsonate",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2026: {
        "gaul0_name": "El Salvador",
        "gaul1_name": "Usulután",
        "iso2_code": "SV",
        "iso3_code": "SLV",
    },
    2027: {
        "gaul0_name": "Falkland Islands (Malvinas)",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "FK",
        "iso3_code": "FLK",
    },
    2028: {
        "gaul0_name": "French Guiana",
        "gaul1_name": "Cayenne",
        "iso2_code": "GF",
        "iso3_code": "GUF",
    },
    2029: {
        "gaul0_name": "French Guiana",
        "gaul1_name": "Saint-Laurent-Du-Maroni",
        "iso2_code": "GF",
        "iso3_code": "GUF",
    },
    2030: {
        "gaul0_name": "Grenada",
        "gaul1_name": "Carriacou",
        "iso2_code": "GD",
        "iso3_code": "GRD",
    },
    2031: {
        "gaul0_name": "Grenada",
        "gaul1_name": "Saint Andrew",
        "iso2_code": "GD",
        "iso3_code": "GRD",
    },
    2032: {
        "gaul0_name": "Grenada",
        "gaul1_name": "Saint David",
        "iso2_code": "GD",
        "iso3_code": "GRD",
    },
    2033: {
        "gaul0_name": "Grenada",
        "gaul1_name": "Saint George",
        "iso2_code": "GD",
        "iso3_code": "GRD",
    },
    2034: {
        "gaul0_name": "Grenada",
        "gaul1_name": "Saint John",
        "iso2_code": "GD",
        "iso3_code": "GRD",
    },
    2035: {
        "gaul0_name": "Grenada",
        "gaul1_name": "Saint Mark",
        "iso2_code": "GD",
        "iso3_code": "GRD",
    },
    2036: {
        "gaul0_name": "Grenada",
        "gaul1_name": "Saint Patrick",
        "iso2_code": "GD",
        "iso3_code": "GRD",
    },
    2037: {
        "gaul0_name": "Guadeloupe",
        "gaul1_name": "Basse-Terre",
        "iso2_code": "GP",
        "iso3_code": "GLP",
    },
    2038: {
        "gaul0_name": "Guadeloupe",
        "gaul1_name": "Pointe-À-Pitre",
        "iso2_code": "GP",
        "iso3_code": "GLP",
    },
    2039: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Alta Verapaz",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2040: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Baja Verapaz",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2041: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Chimaltenango",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2042: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Chiquimula",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2043: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "El Progreso",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2044: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Escuintla",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2045: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Guatemala",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2046: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Huehuetenango",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2047: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Izabal",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2048: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Jalapa",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2049: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Jutiapa",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2050: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Petén",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2051: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Quetzaltenango",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2052: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Quiché",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2053: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Retalhuleu",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2054: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Sacatepéquez",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2055: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "San Marcos",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2056: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Santa Rosa",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2057: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Sololá",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2058: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Suchitepéquez",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2059: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Totonicapán",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2060: {
        "gaul0_name": "Guatemala",
        "gaul1_name": "Zacapa",
        "iso2_code": "GT",
        "iso3_code": "GTM",
    },
    2061: {
        "gaul0_name": "Guyana",
        "gaul1_name": "Barima-Waini",
        "iso2_code": "GY",
        "iso3_code": "GUY",
    },
    2062: {
        "gaul0_name": "Guyana",
        "gaul1_name": "Cuyuni-Mazaruni",
        "iso2_code": "GY",
        "iso3_code": "GUY",
    },
    2063: {
        "gaul0_name": "Guyana",
        "gaul1_name": "Demerara-Mahaica",
        "iso2_code": "GY",
        "iso3_code": "GUY",
    },
    2064: {
        "gaul0_name": "Guyana",
        "gaul1_name": "East Berbice-Corentyne",
        "iso2_code": "GY",
        "iso3_code": "GUY",
    },
    2065: {
        "gaul0_name": "Guyana",
        "gaul1_name": "Essequibo Islands-West Demerara",
        "iso2_code": "GY",
        "iso3_code": "GUY",
    },
    2066: {
        "gaul0_name": "Guyana",
        "gaul1_name": "Mahaica Berbice",
        "iso2_code": "GY",
        "iso3_code": "GUY",
    },
    2067: {
        "gaul0_name": "Guyana",
        "gaul1_name": "Pomeroon-Supenaam",
        "iso2_code": "GY",
        "iso3_code": "GUY",
    },
    2068: {
        "gaul0_name": "Guyana",
        "gaul1_name": "Potaro-Siparuni",
        "iso2_code": "GY",
        "iso3_code": "GUY",
    },
    2069: {
        "gaul0_name": "Guyana",
        "gaul1_name": "Upper Demerara-Berbice",
        "iso2_code": "GY",
        "iso3_code": "GUY",
    },
    2070: {
        "gaul0_name": "Guyana",
        "gaul1_name": "Upper Takutu-Upper Essequibo",
        "iso2_code": "GY",
        "iso3_code": "GUY",
    },
    2071: {
        "gaul0_name": "Haiti",
        "gaul1_name": "Artibonite",
        "iso2_code": "HT",
        "iso3_code": "HTI",
    },
    2072: {
        "gaul0_name": "Haiti",
        "gaul1_name": "Centre",
        "iso2_code": "HT",
        "iso3_code": "HTI",
    },
    2073: {
        "gaul0_name": "Haiti",
        "gaul1_name": "Grand-Anse",
        "iso2_code": "HT",
        "iso3_code": "HTI",
    },
    2074: {
        "gaul0_name": "Haiti",
        "gaul1_name": "Nippes",
        "iso2_code": "HT",
        "iso3_code": "HTI",
    },
    2075: {
        "gaul0_name": "Haiti",
        "gaul1_name": "Nord",
        "iso2_code": "HT",
        "iso3_code": "HTI",
    },
    2076: {
        "gaul0_name": "Haiti",
        "gaul1_name": "Nord-Est",
        "iso2_code": "HT",
        "iso3_code": "HTI",
    },
    2077: {
        "gaul0_name": "Haiti",
        "gaul1_name": "Nord-Ouest",
        "iso2_code": "HT",
        "iso3_code": "HTI",
    },
    2078: {
        "gaul0_name": "Haiti",
        "gaul1_name": "Ouest",
        "iso2_code": "HT",
        "iso3_code": "HTI",
    },
    2079: {
        "gaul0_name": "Haiti",
        "gaul1_name": "Sud",
        "iso2_code": "HT",
        "iso3_code": "HTI",
    },
    2080: {
        "gaul0_name": "Haiti",
        "gaul1_name": "Sud-Est",
        "iso2_code": "HT",
        "iso3_code": "HTI",
    },
    2081: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Atlantida",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2082: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Choluteca",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2083: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Colon",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2084: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Comayagua",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2085: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Copan",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2086: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Cortes",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2087: {
        "gaul0_name": "Honduras",
        "gaul1_name": "El Paraiso",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2088: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Francisco Morazan",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2089: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Gracias A Dios",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2090: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Intibuca",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2091: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Islas De La Bahia",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2092: {
        "gaul0_name": "Honduras",
        "gaul1_name": "La Paz",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2093: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Lempira",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2094: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Ocotepeque",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2095: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Olancho",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2096: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Santa Barbara",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2097: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Valle",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2098: {
        "gaul0_name": "Honduras",
        "gaul1_name": "Yoro",
        "iso2_code": "HN",
        "iso3_code": "HND",
    },
    2099: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Clarendon",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2100: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Hanover",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2101: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Kingston",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2102: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Manchester",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2103: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Portland",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2104: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Saint Andrew",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2105: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Saint Ann",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2106: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Saint Catherine",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2107: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Saint Elizabeth",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2108: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Saint James",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2109: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Saint Mary",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2110: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Saint Thomas",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2111: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Trelawny",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2112: {
        "gaul0_name": "Jamaica",
        "gaul1_name": "Westmoreland",
        "iso2_code": "JM",
        "iso3_code": "JAM",
    },
    2113: {
        "gaul0_name": "Martinique",
        "gaul1_name": "Fort-De-France",
        "iso2_code": "MQ",
        "iso3_code": "MTQ",
    },
    2114: {
        "gaul0_name": "Martinique",
        "gaul1_name": "La Trinité",
        "iso2_code": "MQ",
        "iso3_code": "MTQ",
    },
    2115: {
        "gaul0_name": "Martinique",
        "gaul1_name": "Le Marin",
        "iso2_code": "MQ",
        "iso3_code": "MTQ",
    },
    2116: {
        "gaul0_name": "Martinique",
        "gaul1_name": "Saint-Pierre",
        "iso2_code": "MQ",
        "iso3_code": "MTQ",
    },
    2117: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Aguascalientes",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2118: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Baja California",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2119: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Baja California Sur",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2120: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Campeche",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2121: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Chiapas",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2122: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Chihuahua",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2123: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Ciudad De México",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2124: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Coahuila De Zaragoza",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2125: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Colima",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2126: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Durango",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2127: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Guanajuato",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2128: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Guerrero",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2129: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Hidalgo",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2130: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Jalisco",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2131: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Michoacán De Ocampo",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2132: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Morelos",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2133: {
        "gaul0_name": "Mexico",
        "gaul1_name": "México",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2134: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Nayarit",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2135: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Nuevo León",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2136: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Oaxaca",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2137: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Puebla",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2138: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Querétaro",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2139: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Quintana Roo",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2140: {
        "gaul0_name": "Mexico",
        "gaul1_name": "San Luis Potosí",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2141: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Sinaloa",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2142: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Sonora",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2143: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Tabasco",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2144: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Tamaulipas",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2145: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Tlaxcala",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2146: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Veracruz De Ignacio De La Llave",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2147: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Yucatán",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2148: {
        "gaul0_name": "Mexico",
        "gaul1_name": "Zacatecas",
        "iso2_code": "MX",
        "iso3_code": "MEX",
    },
    2149: {
        "gaul0_name": "Montserrat",
        "gaul1_name": "Saint Anthony",
        "iso2_code": "MS",
        "iso3_code": "MSR",
    },
    2150: {
        "gaul0_name": "Montserrat",
        "gaul1_name": "Saint Georges",
        "iso2_code": "MS",
        "iso3_code": "MSR",
    },
    2151: {
        "gaul0_name": "Montserrat",
        "gaul1_name": "Saint Peter",
        "iso2_code": "MS",
        "iso3_code": "MSR",
    },
    2152: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Boaco",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2153: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Carazo",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2154: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Chinandega",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2155: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Chontales",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2156: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Estelí",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2157: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Granada",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2158: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Jinotega",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2159: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "León",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2160: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Madriz",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2161: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Managua",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2162: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Masaya",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2163: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Matagalpa",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2164: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Nueva Segovia",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2165: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Región Autónoma Atlántico Norte",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2166: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Región Autónoma Atlántico Sur",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2167: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Rivas",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2168: {
        "gaul0_name": "Nicaragua",
        "gaul1_name": "Río San Juan",
        "iso2_code": "NI",
        "iso3_code": "NIC",
    },
    2169: {
        "gaul0_name": "Panama",
        "gaul1_name": "Bocas Del Toro",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2170: {
        "gaul0_name": "Panama",
        "gaul1_name": "Chiriquí",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2171: {
        "gaul0_name": "Panama",
        "gaul1_name": "Coclé",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2172: {
        "gaul0_name": "Panama",
        "gaul1_name": "Colón",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2173: {
        "gaul0_name": "Panama",
        "gaul1_name": "Darién",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2174: {
        "gaul0_name": "Panama",
        "gaul1_name": "Emberá",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2175: {
        "gaul0_name": "Panama",
        "gaul1_name": "Herrera",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2176: {
        "gaul0_name": "Panama",
        "gaul1_name": "Kuna Yala",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2177: {
        "gaul0_name": "Panama",
        "gaul1_name": "Los Santos",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2178: {
        "gaul0_name": "Panama",
        "gaul1_name": "Ngöbe Buglé",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2179: {
        "gaul0_name": "Panama",
        "gaul1_name": "Panamá",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2180: {
        "gaul0_name": "Panama",
        "gaul1_name": "Panamá Oeste",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2181: {
        "gaul0_name": "Panama",
        "gaul1_name": "Veraguas",
        "iso2_code": "PA",
        "iso3_code": "PAN",
    },
    2182: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Alto Paraguay",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2183: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Alto Paraná",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2184: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Amambay",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2185: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Asunción",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2186: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Boquerón",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2187: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Caaguazú",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2188: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Caazapá",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2189: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Canindeyú",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2190: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Central",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2191: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Concepción",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2192: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Cordillera",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2193: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Guairá",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2194: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Itapúa",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2195: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Misiones",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2196: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Paraguarí",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2197: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Presidente Hayes",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2198: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "San Pedro",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2199: {
        "gaul0_name": "Paraguay",
        "gaul1_name": "Ñeembucú",
        "iso2_code": "PY",
        "iso3_code": "PRY",
    },
    2200: {
        "gaul0_name": "Peru",
        "gaul1_name": "Amazonas",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2201: {
        "gaul0_name": "Peru",
        "gaul1_name": "Ancash",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2202: {
        "gaul0_name": "Peru",
        "gaul1_name": "Apurimac",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2203: {
        "gaul0_name": "Peru",
        "gaul1_name": "Arequipa",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2204: {
        "gaul0_name": "Peru",
        "gaul1_name": "Ayacucho",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2205: {
        "gaul0_name": "Peru",
        "gaul1_name": "Cajamarca",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2206: {
        "gaul0_name": "Peru",
        "gaul1_name": "Callao",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2207: {
        "gaul0_name": "Peru",
        "gaul1_name": "Cusco",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2208: {
        "gaul0_name": "Peru",
        "gaul1_name": "Huancavelica",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2209: {
        "gaul0_name": "Peru",
        "gaul1_name": "Huanuco",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2210: {
        "gaul0_name": "Peru",
        "gaul1_name": "Ica",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2211: {
        "gaul0_name": "Peru",
        "gaul1_name": "Junin",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2212: {
        "gaul0_name": "Peru",
        "gaul1_name": "La Libertad",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2213: {
        "gaul0_name": "Peru",
        "gaul1_name": "Lambayeque",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2214: {
        "gaul0_name": "Peru",
        "gaul1_name": "Lima",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2215: {
        "gaul0_name": "Peru",
        "gaul1_name": "Loreto",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2216: {
        "gaul0_name": "Peru",
        "gaul1_name": "Madre De Dios",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2217: {
        "gaul0_name": "Peru",
        "gaul1_name": "Moquegua",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2218: {
        "gaul0_name": "Peru",
        "gaul1_name": "Pasco",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2219: {
        "gaul0_name": "Peru",
        "gaul1_name": "Piura",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2220: {
        "gaul0_name": "Peru",
        "gaul1_name": "Puno",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2221: {
        "gaul0_name": "Peru",
        "gaul1_name": "San Martin",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2222: {
        "gaul0_name": "Peru",
        "gaul1_name": "Tacna",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2223: {
        "gaul0_name": "Peru",
        "gaul1_name": "Tumbes",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2224: {
        "gaul0_name": "Peru",
        "gaul1_name": "Ucayali",
        "iso2_code": "PE",
        "iso3_code": "PER",
    },
    2225: {
        "gaul0_name": "Puerto Rico",
        "gaul1_name": "Aguadilla",
        "iso2_code": "PR",
        "iso3_code": "PRI",
    },
    2226: {
        "gaul0_name": "Puerto Rico",
        "gaul1_name": "Arecibo",
        "iso2_code": "PR",
        "iso3_code": "PRI",
    },
    2227: {
        "gaul0_name": "Puerto Rico",
        "gaul1_name": "Bayamon",
        "iso2_code": "PR",
        "iso3_code": "PRI",
    },
    2228: {
        "gaul0_name": "Puerto Rico",
        "gaul1_name": "Guayama",
        "iso2_code": "PR",
        "iso3_code": "PRI",
    },
    2229: {
        "gaul0_name": "Puerto Rico",
        "gaul1_name": "Humacao",
        "iso2_code": "PR",
        "iso3_code": "PRI",
    },
    2230: {
        "gaul0_name": "Puerto Rico",
        "gaul1_name": "Mayaguez",
        "iso2_code": "PR",
        "iso3_code": "PRI",
    },
    2231: {
        "gaul0_name": "Puerto Rico",
        "gaul1_name": "Ponce",
        "iso2_code": "PR",
        "iso3_code": "PRI",
    },
    2232: {
        "gaul0_name": "Puerto Rico",
        "gaul1_name": "San Juan",
        "iso2_code": "PR",
        "iso3_code": "PRI",
    },
    2233: {
        "gaul0_name": "Saint Barthélemy",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "BL",
        "iso3_code": "BLM",
    },
    2234: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Christ Church Nichola Town",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2235: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint Anne Sandy Point",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2236: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint George Basseterre",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2237: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint George Gingerland",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2238: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint James Windward",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2239: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint John Capisterre",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2240: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint John Figtree",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2241: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint Mary Cayon",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2242: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint Paul Capisterre",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2243: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint Paul Charlestown",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2244: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint Peter Basseterre",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2245: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint Thomas Lowland",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2246: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Saint Thomas Middle Island",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2247: {
        "gaul0_name": "Saint Kitts And Nevis",
        "gaul1_name": "Trinity Palmetto Point",
        "iso2_code": "KN",
        "iso3_code": "KNA",
    },
    2248: {
        "gaul0_name": "Saint Lucia",
        "gaul1_name": "Anse La Raye",
        "iso2_code": "LC",
        "iso3_code": "LCA",
    },
    2249: {
        "gaul0_name": "Saint Lucia",
        "gaul1_name": "Canaries",
        "iso2_code": "LC",
        "iso3_code": "LCA",
    },
    2250: {
        "gaul0_name": "Saint Lucia",
        "gaul1_name": "Castries",
        "iso2_code": "LC",
        "iso3_code": "LCA",
    },
    2251: {
        "gaul0_name": "Saint Lucia",
        "gaul1_name": "Choiseul",
        "iso2_code": "LC",
        "iso3_code": "LCA",
    },
    2252: {
        "gaul0_name": "Saint Lucia",
        "gaul1_name": "Dennery",
        "iso2_code": "LC",
        "iso3_code": "LCA",
    },
    2253: {
        "gaul0_name": "Saint Lucia",
        "gaul1_name": "Gros Islet",
        "iso2_code": "LC",
        "iso3_code": "LCA",
    },
    2254: {
        "gaul0_name": "Saint Lucia",
        "gaul1_name": "Laborie",
        "iso2_code": "LC",
        "iso3_code": "LCA",
    },
    2255: {
        "gaul0_name": "Saint Lucia",
        "gaul1_name": "Micoud",
        "iso2_code": "LC",
        "iso3_code": "LCA",
    },
    2256: {
        "gaul0_name": "Saint Lucia",
        "gaul1_name": "Soufriere",
        "iso2_code": "LC",
        "iso3_code": "LCA",
    },
    2257: {
        "gaul0_name": "Saint Lucia",
        "gaul1_name": "Vieux-Fort",
        "iso2_code": "LC",
        "iso3_code": "LCA",
    },
    2258: {
        "gaul0_name": "Saint Martin (French part)",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "MF",
        "iso3_code": "MAF",
    },
    2259: {
        "gaul0_name": "Saint Pierre and Miquelon",
        "gaul1_name": "Miquelon-Langlade",
        "iso2_code": "PM",
        "iso3_code": "SPM",
    },
    2260: {
        "gaul0_name": "Saint Pierre and Miquelon",
        "gaul1_name": "Saint-Pierre",
        "iso2_code": "PM",
        "iso3_code": "SPM",
    },
    2261: {
        "gaul0_name": "Saint Vincent and the Grenadines",
        "gaul1_name": "Charlotte",
        "iso2_code": "VC",
        "iso3_code": "VCT",
    },
    2262: {
        "gaul0_name": "Saint Vincent and the Grenadines",
        "gaul1_name": "Grenadines",
        "iso2_code": "VC",
        "iso3_code": "VCT",
    },
    2263: {
        "gaul0_name": "Saint Vincent and the Grenadines",
        "gaul1_name": "Saint Andrew",
        "iso2_code": "VC",
        "iso3_code": "VCT",
    },
    2264: {
        "gaul0_name": "Saint Vincent and the Grenadines",
        "gaul1_name": "Saint David",
        "iso2_code": "VC",
        "iso3_code": "VCT",
    },
    2265: {
        "gaul0_name": "Saint Vincent and the Grenadines",
        "gaul1_name": "Saint George",
        "iso2_code": "VC",
        "iso3_code": "VCT",
    },
    2266: {
        "gaul0_name": "Saint Vincent and the Grenadines",
        "gaul1_name": "Saint Patrick",
        "iso2_code": "VC",
        "iso3_code": "VCT",
    },
    2267: {
        "gaul0_name": "South Georgia and the South Sandwich Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "GS",
        "iso3_code": "SGS",
    },
    2268: {
        "gaul0_name": "Suriname",
        "gaul1_name": "Brokopondo",
        "iso2_code": "SR",
        "iso3_code": "SUR",
    },
    2269: {
        "gaul0_name": "Suriname",
        "gaul1_name": "Commewijne",
        "iso2_code": "SR",
        "iso3_code": "SUR",
    },
    2270: {
        "gaul0_name": "Suriname",
        "gaul1_name": "Coronie",
        "iso2_code": "SR",
        "iso3_code": "SUR",
    },
    2271: {
        "gaul0_name": "Suriname",
        "gaul1_name": "Marowijne",
        "iso2_code": "SR",
        "iso3_code": "SUR",
    },
    2272: {
        "gaul0_name": "Suriname",
        "gaul1_name": "Nickerie",
        "iso2_code": "SR",
        "iso3_code": "SUR",
    },
    2273: {
        "gaul0_name": "Suriname",
        "gaul1_name": "Para",
        "iso2_code": "SR",
        "iso3_code": "SUR",
    },
    2274: {
        "gaul0_name": "Suriname",
        "gaul1_name": "Paramaribo",
        "iso2_code": "SR",
        "iso3_code": "SUR",
    },
    2275: {
        "gaul0_name": "Suriname",
        "gaul1_name": "Saramacca",
        "iso2_code": "SR",
        "iso3_code": "SUR",
    },
    2276: {
        "gaul0_name": "Suriname",
        "gaul1_name": "Sipaliwini",
        "iso2_code": "SR",
        "iso3_code": "SUR",
    },
    2277: {
        "gaul0_name": "Suriname",
        "gaul1_name": "Wanica",
        "iso2_code": "SR",
        "iso3_code": "SUR",
    },
    2278: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Arima",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2279: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Chaguanas",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2280: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Couva-Tabaquite-Talparo",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2281: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Diego Martin",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2282: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Mayaro/Rio Claro",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2283: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Penal-Debe",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2284: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Point Fortin",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2285: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Port Of Spain",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2286: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Princes Town",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2287: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "San Fernando",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2288: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "San Juan-Laventille",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2289: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Sangre Grande",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2290: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Siparia",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2291: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Tobago",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2292: {
        "gaul0_name": "Trinidad And Tobago",
        "gaul1_name": "Tunapuna/Piarco",
        "iso2_code": "TT",
        "iso3_code": "TTO",
    },
    2293: {
        "gaul0_name": "Turks And Caicos Islands",
        "gaul1_name": "Grand Turk",
        "iso2_code": "TC",
        "iso3_code": "TCA",
    },
    2294: {
        "gaul0_name": "Turks And Caicos Islands",
        "gaul1_name": "Middle Caicos",
        "iso2_code": "TC",
        "iso3_code": "TCA",
    },
    2295: {
        "gaul0_name": "Turks And Caicos Islands",
        "gaul1_name": "North Caicos",
        "iso2_code": "TC",
        "iso3_code": "TCA",
    },
    2296: {
        "gaul0_name": "Turks And Caicos Islands",
        "gaul1_name": "Providenciales And West Caicos",
        "iso2_code": "TC",
        "iso3_code": "TCA",
    },
    2297: {
        "gaul0_name": "Turks And Caicos Islands",
        "gaul1_name": "Salt Cay",
        "iso2_code": "TC",
        "iso3_code": "TCA",
    },
    2298: {
        "gaul0_name": "Turks And Caicos Islands",
        "gaul1_name": "South Caicos And East Caicos",
        "iso2_code": "TC",
        "iso3_code": "TCA",
    },
    2299: {
        "gaul0_name": "United States Virgin Islands",
        "gaul1_name": "St. Croix",
        "iso2_code": "VI",
        "iso3_code": "VIR",
    },
    2300: {
        "gaul0_name": "United States Virgin Islands",
        "gaul1_name": "St. John",
        "iso2_code": "VI",
        "iso3_code": "VIR",
    },
    2301: {
        "gaul0_name": "United States Virgin Islands",
        "gaul1_name": "St. Thomas",
        "iso2_code": "VI",
        "iso3_code": "VIR",
    },
    2302: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Alabama",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2303: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Alaska",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2304: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Arizona",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2305: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Arkansas",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2306: {
        "gaul0_name": "United States of America",
        "gaul1_name": "California",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2307: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Colorado",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2308: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Connecticut",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2309: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Delaware",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2310: {
        "gaul0_name": "United States of America",
        "gaul1_name": "District Of Columbia",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2311: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Florida",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2312: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Georgia",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2313: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Hawaii",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2314: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Idaho",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2315: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Illinois",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2316: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Indiana",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2317: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Iowa",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2318: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Kansas",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2319: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Kentucky",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2320: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Louisiana",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2321: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Maine",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2322: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Maryland",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2323: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Massachusetts",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2324: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Michigan",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2325: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Minnesota",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2326: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Mississippi",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2327: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Missouri",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2328: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Montana",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2329: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Nebraska",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2330: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Nevada",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2331: {
        "gaul0_name": "United States of America",
        "gaul1_name": "New Hampshire",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2332: {
        "gaul0_name": "United States of America",
        "gaul1_name": "New Jersey",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2333: {
        "gaul0_name": "United States of America",
        "gaul1_name": "New Mexico",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2334: {
        "gaul0_name": "United States of America",
        "gaul1_name": "New York",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2335: {
        "gaul0_name": "United States of America",
        "gaul1_name": "North Carolina",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2336: {
        "gaul0_name": "United States of America",
        "gaul1_name": "North Dakota",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2337: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Ohio",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2338: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Oklahoma",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2339: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Oregon",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2340: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Pennsylvania",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2341: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Rhode Island",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2342: {
        "gaul0_name": "United States of America",
        "gaul1_name": "South Carolina",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2343: {
        "gaul0_name": "United States of America",
        "gaul1_name": "South Dakota",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2344: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Tennessee",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2345: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Texas",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2346: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Utah",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2347: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Vermont",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2348: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Virginia",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2349: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Washington",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2350: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Waterbody",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2351: {
        "gaul0_name": "United States of America",
        "gaul1_name": "West Virginia",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2352: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Wisconsin",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2353: {
        "gaul0_name": "United States of America",
        "gaul1_name": "Wyoming",
        "iso2_code": "US",
        "iso3_code": "USA",
    },
    2354: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Artigas",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2355: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Canelones",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2356: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Cerro Largo",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2357: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Colonia",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2358: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Durazno",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2359: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Flores",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2360: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Florida",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2361: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Lavalleja",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2362: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Maldonado",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2363: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Montevideo",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2364: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Paysandú",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2365: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Rivera",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2366: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Rocha",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2367: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Río Negro",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2368: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Salto",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2369: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "San José",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2370: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Soriano",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2371: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Tacuarembó",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2372: {
        "gaul0_name": "Uruguay",
        "gaul1_name": "Treinta Y Tres",
        "iso2_code": "UY",
        "iso3_code": "URY",
    },
    2373: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Amazonas",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2374: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Anzoategui",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2375: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Apure",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2376: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Aragua",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2377: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Barinas",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2378: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Bolivar",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2379: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Carabobo",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2380: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Cojedes",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2381: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Delta Amacuro",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2382: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Distrito Capital",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2383: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Falcon",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2384: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Federal Dependencies",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2385: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Guarico",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2386: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "La Guaira",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2387: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Lara",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2388: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Merida",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2389: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Miranda",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2390: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Monagas",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2391: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Nueva Esparta",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2392: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Portuguesa",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2393: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Sucre",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2394: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Tachira",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2395: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Trujillo",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2396: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Yaracuy",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2397: {
        "gaul0_name": "Venezuela (Bolivarian Republic Of)",
        "gaul1_name": "Zulia",
        "iso2_code": "VE",
        "iso3_code": "VEN",
    },
    2398: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Badakhshan",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2399: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Badghis",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2400: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Baghlan",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2401: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Balkh",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2402: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Bamyan",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2403: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Daykundi",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2404: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Farah",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2405: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Faryab",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2406: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Ghazni",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2407: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Ghor",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2408: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Hilmand",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2409: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Hirat",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2410: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Jawzjan",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2411: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Kabul",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2412: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Kandahar",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2413: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Kapisa",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2414: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Khost",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2415: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Kunar",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2416: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Kunduz",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2417: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Laghman",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2418: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Logar",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2419: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Maidan Wardak",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2420: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Nangarhar",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2421: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Nimroz",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2422: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Nuristan",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2423: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Paktika",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2424: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Paktya",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2425: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Panjsher",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2426: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Parwan",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2427: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Samangan",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2428: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Sar-E-Pul",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2429: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Takhar",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2430: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Uruzgan",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2431: {
        "gaul0_name": "Afghanistan",
        "gaul1_name": "Zabul",
        "iso2_code": "AF",
        "iso3_code": "AFG",
    },
    2432: {
        "gaul0_name": "Aksai Chin",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xAC",
    },
    2433: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Aragatsotn",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2434: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Ararat",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2435: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Armavir",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2436: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Gegharkunik",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2437: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Kotayk",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2438: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Lori",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2439: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Shirak",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2440: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Syunik",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2441: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Tavush",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2442: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Vayots Dzor",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2443: {
        "gaul0_name": "Armenia",
        "gaul1_name": "Yerevan",
        "iso2_code": "AM",
        "iso3_code": "ARM",
    },
    2444: {
        "gaul0_name": "Arunachal Pradesh",
        "gaul1_name": "Arunachal Pradesh",
        "iso2_code": "not found",
        "iso3_code": "xAP",
    },
    2445: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Absheron-Khizi Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2446: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Baku Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2447: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Central Aran Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2448: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "East Zangezur Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2449: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Ganja-Dashkasan Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2450: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Gazakh-Tovuz Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2451: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Guba-Khachmaz Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2452: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Karabakh Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2453: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Lankaran-Astara Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2454: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Mil-Mughan Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2455: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Mountainous Shirvan Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2456: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Nakhchivan Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2457: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Shaki-Zagatala Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2458: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Shirvan-Salyan Economic Region",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2459: {
        "gaul0_name": "Azerbaijan",
        "gaul1_name": "Waterbody",
        "iso2_code": "AZ",
        "iso3_code": "AZE",
    },
    2460: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Central Region",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2461: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Hamad Town",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2462: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Hawar Island",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2463: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Hidd",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2464: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Isa Town",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2465: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Jidhafs",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2466: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Manama",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2467: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Muharraq",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2468: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Northern Region",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2469: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Riffa",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2470: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Sitra",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2471: {
        "gaul0_name": "Bahrain",
        "gaul1_name": "Western Region",
        "iso2_code": "BH",
        "iso3_code": "BHR",
    },
    2472: {
        "gaul0_name": "Bangladesh",
        "gaul1_name": "Barisal",
        "iso2_code": "BD",
        "iso3_code": "BGD",
    },
    2473: {
        "gaul0_name": "Bangladesh",
        "gaul1_name": "Chittagong",
        "iso2_code": "BD",
        "iso3_code": "BGD",
    },
    2474: {
        "gaul0_name": "Bangladesh",
        "gaul1_name": "Dhaka",
        "iso2_code": "BD",
        "iso3_code": "BGD",
    },
    2475: {
        "gaul0_name": "Bangladesh",
        "gaul1_name": "Khulna",
        "iso2_code": "BD",
        "iso3_code": "BGD",
    },
    2476: {
        "gaul0_name": "Bangladesh",
        "gaul1_name": "Mymensingh",
        "iso2_code": "BD",
        "iso3_code": "BGD",
    },
    2477: {
        "gaul0_name": "Bangladesh",
        "gaul1_name": "Rajshahi",
        "iso2_code": "BD",
        "iso3_code": "BGD",
    },
    2478: {
        "gaul0_name": "Bangladesh",
        "gaul1_name": "Rangpur",
        "iso2_code": "BD",
        "iso3_code": "BGD",
    },
    2479: {
        "gaul0_name": "Bangladesh",
        "gaul1_name": "Sylhet",
        "iso2_code": "BD",
        "iso3_code": "BGD",
    },
    2480: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Bumthang",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2481: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Chhukha",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2482: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Dagana",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2483: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Gasa",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2484: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Haa",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2485: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Lhuentse",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2486: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Monggar",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2487: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Paro",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2488: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Pemagatshel",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2489: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Punakha",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2490: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Samdrupjongkhar",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2491: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Samtse",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2492: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Sarpang",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2493: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Thimphu",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2494: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Trashigang",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2495: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Trashiyangtse",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2496: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Trongsa",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2497: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Tsirang",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2498: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Wangduephodrang",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2499: {
        "gaul0_name": "Bhutan",
        "gaul1_name": "Zhemgang",
        "iso2_code": "BT",
        "iso3_code": "BTN",
    },
    2500: {
        "gaul0_name": "Brunei Darussalam",
        "gaul1_name": "Belait",
        "iso2_code": "BN",
        "iso3_code": "BRN",
    },
    2501: {
        "gaul0_name": "Brunei Darussalam",
        "gaul1_name": "Brunei And Muara",
        "iso2_code": "BN",
        "iso3_code": "BRN",
    },
    2502: {
        "gaul0_name": "Brunei Darussalam",
        "gaul1_name": "Temburong",
        "iso2_code": "BN",
        "iso3_code": "BRN",
    },
    2503: {
        "gaul0_name": "Brunei Darussalam",
        "gaul1_name": "Tutong",
        "iso2_code": "BN",
        "iso3_code": "BRN",
    },
    2504: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Banteay Meanchey",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2505: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Battambang",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2506: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Kampong Cham",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2507: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Kampong Chhnang",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2508: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Kampong Speu",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2509: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Kampong Thom",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2510: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Kampot",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2511: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Kandal",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2512: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Kep",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2513: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Koh Kong",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2514: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Kratie",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2515: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Mondul Kiri",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2516: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Oddar Meanchey",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2517: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Pailin",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2518: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Phnom Penh",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2519: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Preah Sihanouk",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2520: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Preah Vihear",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2521: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Prey Veng",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2522: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Pursat",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2523: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Ratanak Kiri",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2524: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Siemreap",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2525: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Stung Treng",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2526: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Svay Rieng",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2527: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Takeo",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2528: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Tboung Khmum",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2529: {
        "gaul0_name": "Cambodia",
        "gaul1_name": "Under National Administration",
        "iso2_code": "KH",
        "iso3_code": "KHM",
    },
    2530: {
        "gaul0_name": "Chagos Archipelagio",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "IO",
        "iso3_code": "IOT",
    },
    2531: {
        "gaul0_name": "China",
        "gaul1_name": "Anhui Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2532: {
        "gaul0_name": "China",
        "gaul1_name": "Beijing Municipality",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2533: {
        "gaul0_name": "China",
        "gaul1_name": "Chongqing Municipality",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2534: {
        "gaul0_name": "China",
        "gaul1_name": "Fujian Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2535: {
        "gaul0_name": "China",
        "gaul1_name": "Gansu Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2536: {
        "gaul0_name": "China",
        "gaul1_name": "Guangdong Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2537: {
        "gaul0_name": "China",
        "gaul1_name": "Guangxi Zhuang Autonomous Region",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2538: {
        "gaul0_name": "China",
        "gaul1_name": "Guizhou Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2539: {
        "gaul0_name": "China",
        "gaul1_name": "Hainan Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2540: {
        "gaul0_name": "China",
        "gaul1_name": "Hebei Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2541: {
        "gaul0_name": "China",
        "gaul1_name": "Heilongjiang Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2542: {
        "gaul0_name": "China",
        "gaul1_name": "Henan Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2543: {
        "gaul0_name": "China",
        "gaul1_name": "Hubei Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2544: {
        "gaul0_name": "China",
        "gaul1_name": "Hunan Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2545: {
        "gaul0_name": "China",
        "gaul1_name": "Inner Mongolia Autonomous Region",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2546: {
        "gaul0_name": "China",
        "gaul1_name": "Jiangsu Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2547: {
        "gaul0_name": "China",
        "gaul1_name": "Jiangxi Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2548: {
        "gaul0_name": "China",
        "gaul1_name": "Jilin Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2549: {
        "gaul0_name": "China",
        "gaul1_name": "Liaoning Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2550: {
        "gaul0_name": "China",
        "gaul1_name": "Ningxia Hui Autonomous Region",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2551: {
        "gaul0_name": "China",
        "gaul1_name": "Qinghai Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2552: {
        "gaul0_name": "China",
        "gaul1_name": "Shaanxi Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2553: {
        "gaul0_name": "China",
        "gaul1_name": "Shandong Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2554: {
        "gaul0_name": "China",
        "gaul1_name": "Shanghai Municipality",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2555: {
        "gaul0_name": "China",
        "gaul1_name": "Shanxi Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2556: {
        "gaul0_name": "China",
        "gaul1_name": "Sichuan Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2557: {
        "gaul0_name": "China",
        "gaul1_name": "Tianjin Municipality",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2558: {
        "gaul0_name": "China",
        "gaul1_name": "Tibet Autonomous Region",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2559: {
        "gaul0_name": "China",
        "gaul1_name": "Xinjiang Uygur Autonomous Region",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2560: {
        "gaul0_name": "China",
        "gaul1_name": "Yunnan Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2561: {
        "gaul0_name": "China",
        "gaul1_name": "Zhejiang Province",
        "iso2_code": "CN",
        "iso3_code": "CHN",
    },
    2562: {
        "gaul0_name": "China, Hong Kong SAR",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "HK",
        "iso3_code": "HKG",
    },
    2563: {
        "gaul0_name": "China, Macao SAR",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "MO",
        "iso3_code": "MAC",
    },
    2564: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "Jagang",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2565: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "Kangwon",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2566: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "Nampo",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2567: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "North Hamgyong",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2568: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "North Hwanghae",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2569: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "North Pyongan",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2570: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "Pyongyang",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2571: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "Ryanggang",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2572: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "South Hamgyong",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2573: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "South Hwanghae",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2574: {
        "gaul0_name": "Democratic People's Republic of Korea",
        "gaul1_name": "South Pyongan",
        "iso2_code": "KP",
        "iso3_code": "PRK",
    },
    2575: {
        "gaul0_name": "Disputed Area(xxx)",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xxx",
    },
    2576: {
        "gaul0_name": "Disputed Area(xJL)",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xJL",
    },
    2577: {
        "gaul0_name": "India",
        "gaul1_name": "Andaman & Nicobar",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2578: {
        "gaul0_name": "India",
        "gaul1_name": "Andhra Pradesh",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2579: {
        "gaul0_name": "India",
        "gaul1_name": "Arunchal Pradesh",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2580: {
        "gaul0_name": "India",
        "gaul1_name": "Assam",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2581: {
        "gaul0_name": "India",
        "gaul1_name": "Bihar",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2582: {
        "gaul0_name": "India",
        "gaul1_name": "Chandigarh",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2583: {
        "gaul0_name": "India",
        "gaul1_name": "Chhattisgarh",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2584: {
        "gaul0_name": "India",
        "gaul1_name": "Dadra & Nagar Haveli & Daman & Diu",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2585: {
        "gaul0_name": "India",
        "gaul1_name": "Delhi",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2586: {
        "gaul0_name": "India",
        "gaul1_name": "Disputed (Madhya Pradesh & Gujarat)",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2587: {
        "gaul0_name": "India",
        "gaul1_name": "Disputed (Madhya Pradesh & Rajasthan)",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2588: {
        "gaul0_name": "India",
        "gaul1_name": "Disputed (Rajasthan & Gujarat)",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2589: {
        "gaul0_name": "India",
        "gaul1_name": "Disputed (West Bengal, Bihar & Jharkhand)",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2590: {
        "gaul0_name": "India",
        "gaul1_name": "Goa",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2591: {
        "gaul0_name": "India",
        "gaul1_name": "Gujarat",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2592: {
        "gaul0_name": "India",
        "gaul1_name": "Haryana",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2593: {
        "gaul0_name": "India",
        "gaul1_name": "Himachal Pradesh",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2594: {
        "gaul0_name": "India",
        "gaul1_name": "Jharkhand",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2595: {
        "gaul0_name": "India",
        "gaul1_name": "Karnataka",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2596: {
        "gaul0_name": "India",
        "gaul1_name": "Kerala",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2597: {
        "gaul0_name": "India",
        "gaul1_name": "Lakshadweep",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2598: {
        "gaul0_name": "India",
        "gaul1_name": "Madhya Pradesh",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2599: {
        "gaul0_name": "India",
        "gaul1_name": "Maharashtra",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2600: {
        "gaul0_name": "India",
        "gaul1_name": "Manipur",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2601: {
        "gaul0_name": "India",
        "gaul1_name": "Meghalaya",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2602: {
        "gaul0_name": "India",
        "gaul1_name": "Mizoram",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2603: {
        "gaul0_name": "India",
        "gaul1_name": "Nagaland",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2604: {
        "gaul0_name": "India",
        "gaul1_name": "Odisha",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2605: {
        "gaul0_name": "India",
        "gaul1_name": "Puducherry",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2606: {
        "gaul0_name": "India",
        "gaul1_name": "Punjab",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2607: {
        "gaul0_name": "India",
        "gaul1_name": "Rajasthan",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2608: {
        "gaul0_name": "India",
        "gaul1_name": "Sikkim",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2609: {
        "gaul0_name": "India",
        "gaul1_name": "Tamil Nadu",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2610: {
        "gaul0_name": "India",
        "gaul1_name": "Telangana",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2611: {
        "gaul0_name": "India",
        "gaul1_name": "Tripura",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2612: {
        "gaul0_name": "India",
        "gaul1_name": "Uttar Pradesh",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2613: {
        "gaul0_name": "India",
        "gaul1_name": "Uttarakhand",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2614: {
        "gaul0_name": "India",
        "gaul1_name": "West Bengal",
        "iso2_code": "IN",
        "iso3_code": "IND",
    },
    2615: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Aceh",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2616: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Bali",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2617: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Banten",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2618: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Bengkulu",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2619: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Daerah Istimewa Yogyakarta",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2620: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Dki Jakarta",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2621: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Gorontalo",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2622: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Jambi",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2623: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Jawa Barat",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2624: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Jawa Tengah",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2625: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Jawa Timur",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2626: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Kalimantan Barat",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2627: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Kalimantan Selatan",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2628: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Kalimantan Tengah",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2629: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Kalimantan Timur",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2630: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Kalimantan Utara",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2631: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Kepulauan Bangka Belitung",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2632: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Kepulauan Riau",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2633: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Lampung",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2634: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Maluku",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2635: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Maluku Utara",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2636: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Nusa Tenggara Barat",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2637: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Nusa Tenggara Timur",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2638: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Papua",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2639: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Papua Barat",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2640: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Riau",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2641: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Sulawesi Barat",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2642: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Sulawesi Selatan",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2643: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Sulawesi Tengah",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2644: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Sulawesi Tenggara",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2645: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Sulawesi Utara",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2646: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Sumatera Barat",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2647: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Sumatera Selatan",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2648: {
        "gaul0_name": "Indonesia",
        "gaul1_name": "Sumatera Utara",
        "iso2_code": "ID",
        "iso3_code": "IDN",
    },
    2649: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Alborz",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2650: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Ardabil",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2651: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Bushehr",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2652: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Chaharmahal And Bakhtiari",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2653: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "East Azerbaijan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2654: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Fars",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2655: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Gilan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2656: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Golestan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2657: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Hamadan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2658: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Hormozgan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2659: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Ilam",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2660: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Isfahan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2661: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Kerman",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2662: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Kermanshah",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2663: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Khuzestan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2664: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Kohgiluyeh And Boyer-Ahmad",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2665: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Kurdistan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2666: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Lorestan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2667: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Markazi",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2668: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Mazandaran",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2669: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "North Khorasan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2670: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Qazvin",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2671: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Qom",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2672: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Razavi Khorasan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2673: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Semnan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2674: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Sistan And Baluchestan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2675: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "South Khorasan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2676: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Tehran",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2677: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Waterbody",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2678: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "West Azerbaijan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2679: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Yazd",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2680: {
        "gaul0_name": "Iran (Islamic Republic Of)",
        "gaul1_name": "Zanjan",
        "iso2_code": "IR",
        "iso3_code": "IRN",
    },
    2681: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Al-Anbar",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2682: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Al-Basrah",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2683: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Al-Muthanna",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2684: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Al-Najaf",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2685: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Al-Qadissiya",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2686: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Al-Sulaymaniyah",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2687: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Babil",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2688: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Baghdad",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2689: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Diyala",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2690: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Duhok",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2691: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Erbil",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2692: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Kerbala",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2693: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Kirkuk",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2694: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Maysan",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2695: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Ninewa",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2696: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Salah Al-Din",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2697: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Thi Qar",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2698: {
        "gaul0_name": "Iraq",
        "gaul1_name": "Wassit",
        "iso2_code": "IQ",
        "iso3_code": "IRQ",
    },
    2699: {
        "gaul0_name": "Israel",
        "gaul1_name": "Central District",
        "iso2_code": "IL",
        "iso3_code": "ISR",
    },
    2700: {
        "gaul0_name": "Israel",
        "gaul1_name": "Haifa",
        "iso2_code": "IL",
        "iso3_code": "ISR",
    },
    2701: {
        "gaul0_name": "Israel",
        "gaul1_name": "Jerusalem",
        "iso2_code": "IL",
        "iso3_code": "ISR",
    },
    2702: {
        "gaul0_name": "Israel",
        "gaul1_name": "Northern",
        "iso2_code": "IL",
        "iso3_code": "ISR",
    },
    2703: {
        "gaul0_name": "Israel",
        "gaul1_name": "Southern District",
        "iso2_code": "IL",
        "iso3_code": "ISR",
    },
    2704: {
        "gaul0_name": "Israel",
        "gaul1_name": "Tel Aviv",
        "iso2_code": "IL",
        "iso3_code": "ISR",
    },
    2705: {
        "gaul0_name": "Jammu And Kashmir",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xJK",
    },
    2706: {
        "gaul0_name": "Japan",
        "gaul1_name": "Aichi",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2707: {
        "gaul0_name": "Japan",
        "gaul1_name": "Akita",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2708: {
        "gaul0_name": "Japan",
        "gaul1_name": "Aomori",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2709: {
        "gaul0_name": "Japan",
        "gaul1_name": "Chiba",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2710: {
        "gaul0_name": "Japan",
        "gaul1_name": "Ehime",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2711: {
        "gaul0_name": "Japan",
        "gaul1_name": "Fukui",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2712: {
        "gaul0_name": "Japan",
        "gaul1_name": "Fukuoka",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2713: {
        "gaul0_name": "Japan",
        "gaul1_name": "Fukushima",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2714: {
        "gaul0_name": "Japan",
        "gaul1_name": "Gifu",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2715: {
        "gaul0_name": "Japan",
        "gaul1_name": "Gunma",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2716: {
        "gaul0_name": "Japan",
        "gaul1_name": "Hiroshima",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2717: {
        "gaul0_name": "Japan",
        "gaul1_name": "Hokkaido",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2718: {
        "gaul0_name": "Japan",
        "gaul1_name": "Hyōgo",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2719: {
        "gaul0_name": "Japan",
        "gaul1_name": "Ibaraki",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2720: {
        "gaul0_name": "Japan",
        "gaul1_name": "Ishikawa",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2721: {
        "gaul0_name": "Japan",
        "gaul1_name": "Iwate",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2722: {
        "gaul0_name": "Japan",
        "gaul1_name": "Kagawa",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2723: {
        "gaul0_name": "Japan",
        "gaul1_name": "Kagoshima",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2724: {
        "gaul0_name": "Japan",
        "gaul1_name": "Kanagawa",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2725: {
        "gaul0_name": "Japan",
        "gaul1_name": "Kumamoto",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2726: {
        "gaul0_name": "Japan",
        "gaul1_name": "Kyoto",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2727: {
        "gaul0_name": "Japan",
        "gaul1_name": "Kōchi",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2728: {
        "gaul0_name": "Japan",
        "gaul1_name": "Mie",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2729: {
        "gaul0_name": "Japan",
        "gaul1_name": "Miyagi",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2730: {
        "gaul0_name": "Japan",
        "gaul1_name": "Miyazaki",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2731: {
        "gaul0_name": "Japan",
        "gaul1_name": "Nagano",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2732: {
        "gaul0_name": "Japan",
        "gaul1_name": "Nagasaki",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2733: {
        "gaul0_name": "Japan",
        "gaul1_name": "Nara",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2734: {
        "gaul0_name": "Japan",
        "gaul1_name": "Niigata",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2735: {
        "gaul0_name": "Japan",
        "gaul1_name": "Okayama",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2736: {
        "gaul0_name": "Japan",
        "gaul1_name": "Okinawa",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2737: {
        "gaul0_name": "Japan",
        "gaul1_name": "Osaka",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2738: {
        "gaul0_name": "Japan",
        "gaul1_name": "Saga",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2739: {
        "gaul0_name": "Japan",
        "gaul1_name": "Saitama",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2740: {
        "gaul0_name": "Japan",
        "gaul1_name": "Shiga",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2741: {
        "gaul0_name": "Japan",
        "gaul1_name": "Shimane",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2742: {
        "gaul0_name": "Japan",
        "gaul1_name": "Shizuoka",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2743: {
        "gaul0_name": "Japan",
        "gaul1_name": "Tochigi",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2744: {
        "gaul0_name": "Japan",
        "gaul1_name": "Tokushima",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2745: {
        "gaul0_name": "Japan",
        "gaul1_name": "Tokyo",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2746: {
        "gaul0_name": "Japan",
        "gaul1_name": "Tottori",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2747: {
        "gaul0_name": "Japan",
        "gaul1_name": "Toyama",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2748: {
        "gaul0_name": "Japan",
        "gaul1_name": "Wakayama",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2749: {
        "gaul0_name": "Japan",
        "gaul1_name": "Yamagata",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2750: {
        "gaul0_name": "Japan",
        "gaul1_name": "Yamaguchi",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2751: {
        "gaul0_name": "Japan",
        "gaul1_name": "Yamanashi",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2752: {
        "gaul0_name": "Japan",
        "gaul1_name": "Ōita",
        "iso2_code": "JP",
        "iso3_code": "JPN",
    },
    2753: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Ajloon",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2754: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Amman",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2755: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Aqaba",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2756: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Balqa",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2757: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Irbid",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2758: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Jarash",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2759: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Karak",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2760: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Ma'An",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2761: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Madaba",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2762: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Mafraq",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2763: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Tafiela",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2764: {
        "gaul0_name": "Jordan",
        "gaul1_name": "Zarqa",
        "iso2_code": "JO",
        "iso3_code": "JOR",
    },
    2765: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Abay Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2766: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Akmola Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2767: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Aktobe Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2768: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Almaty",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2769: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Almaty Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2770: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Astana",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2771: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Atyrau Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2772: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "East Kazakhstan Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2773: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Jambyl Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2774: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Jetisu Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2775: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Karaganda Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2776: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Kostanay Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2777: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Kyzylorda Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2778: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Mangystau Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2779: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "North Kazakhstan Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2780: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Pavlodar Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2781: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Shymkent",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2782: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Turkistan Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2783: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Ulytau Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2784: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "Waterbody",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2785: {
        "gaul0_name": "Kazakhstan",
        "gaul1_name": "West Kazakhstan Region",
        "iso2_code": "KZ",
        "iso3_code": "KAZ",
    },
    2786: {
        "gaul0_name": "Kuril Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xKI",
    },
    2787: {
        "gaul0_name": "Kuwait",
        "gaul1_name": "Al Ahmadi",
        "iso2_code": "KW",
        "iso3_code": "KWT",
    },
    2788: {
        "gaul0_name": "Kuwait",
        "gaul1_name": "Al Farwaniyah",
        "iso2_code": "KW",
        "iso3_code": "KWT",
    },
    2789: {
        "gaul0_name": "Kuwait",
        "gaul1_name": "Al Jahrah",
        "iso2_code": "KW",
        "iso3_code": "KWT",
    },
    2790: {
        "gaul0_name": "Kuwait",
        "gaul1_name": "Al Kuwayt",
        "iso2_code": "KW",
        "iso3_code": "KWT",
    },
    2791: {
        "gaul0_name": "Kuwait",
        "gaul1_name": "Hawalli",
        "iso2_code": "KW",
        "iso3_code": "KWT",
    },
    2792: {
        "gaul0_name": "Kuwait",
        "gaul1_name": "Mubarak Al-Kabeer",
        "iso2_code": "KW",
        "iso3_code": "KWT",
    },
    2793: {
        "gaul0_name": "Kyrgyzstan",
        "gaul1_name": "Batken",
        "iso2_code": "KG",
        "iso3_code": "KGZ",
    },
    2794: {
        "gaul0_name": "Kyrgyzstan",
        "gaul1_name": "Chui",
        "iso2_code": "KG",
        "iso3_code": "KGZ",
    },
    2795: {
        "gaul0_name": "Kyrgyzstan",
        "gaul1_name": "Issyk-Kul",
        "iso2_code": "KG",
        "iso3_code": "KGZ",
    },
    2796: {
        "gaul0_name": "Kyrgyzstan",
        "gaul1_name": "Jalal-Abad",
        "iso2_code": "KG",
        "iso3_code": "KGZ",
    },
    2797: {
        "gaul0_name": "Kyrgyzstan",
        "gaul1_name": "Naryn",
        "iso2_code": "KG",
        "iso3_code": "KGZ",
    },
    2798: {
        "gaul0_name": "Kyrgyzstan",
        "gaul1_name": "Osh",
        "iso2_code": "KG",
        "iso3_code": "KGZ",
    },
    2799: {
        "gaul0_name": "Kyrgyzstan",
        "gaul1_name": "Talas",
        "iso2_code": "KG",
        "iso3_code": "KGZ",
    },
    2800: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Attapeu",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2801: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Bokeo",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2802: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Bolikhamxai",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2803: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Champasak",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2804: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Houaphan",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2805: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Khammouan",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2806: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Louangnamtha",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2807: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Louangphabang",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2808: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Oudomxai",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2809: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Phongsaly",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2810: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Salavan",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2811: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Savannakhet",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2812: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Vientiane",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2813: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Vientiane Capital",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2814: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Xaignabouly",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2815: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Xaisomboun",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2816: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Xekong",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2817: {
        "gaul0_name": "Lao People's Democratic Republic",
        "gaul1_name": "Xiengkhouang",
        "iso2_code": "LA",
        "iso3_code": "LAO",
    },
    2818: {
        "gaul0_name": "Lebanon",
        "gaul1_name": "Akkar",
        "iso2_code": "LB",
        "iso3_code": "LBN",
    },
    2819: {
        "gaul0_name": "Lebanon",
        "gaul1_name": "Baalbek-El Hermel",
        "iso2_code": "LB",
        "iso3_code": "LBN",
    },
    2820: {
        "gaul0_name": "Lebanon",
        "gaul1_name": "Beirut",
        "iso2_code": "LB",
        "iso3_code": "LBN",
    },
    2821: {
        "gaul0_name": "Lebanon",
        "gaul1_name": "Bekaa",
        "iso2_code": "LB",
        "iso3_code": "LBN",
    },
    2822: {
        "gaul0_name": "Lebanon",
        "gaul1_name": "El Nabatieh",
        "iso2_code": "LB",
        "iso3_code": "LBN",
    },
    2823: {
        "gaul0_name": "Lebanon",
        "gaul1_name": "Mount Lebanon",
        "iso2_code": "LB",
        "iso3_code": "LBN",
    },
    2824: {
        "gaul0_name": "Lebanon",
        "gaul1_name": "North",
        "iso2_code": "LB",
        "iso3_code": "LBN",
    },
    2825: {
        "gaul0_name": "Lebanon",
        "gaul1_name": "South",
        "iso2_code": "LB",
        "iso3_code": "LBN",
    },
    2826: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Johor",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2827: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Kedah",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2828: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Kelantan",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2829: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Melaka",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2830: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Negeri Sembilan",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2831: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Pahang",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2832: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Perak",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2833: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Perlis",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2834: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Pulau Pinang",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2835: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Sabah",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2836: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Sarawak",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2837: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Selangor",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2838: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "Terengganu",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2839: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "W.P. Kuala Lumpur",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2840: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "W.P. Labuan",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2841: {
        "gaul0_name": "Malaysia",
        "gaul1_name": "W.P. Putrajaya",
        "iso2_code": "MY",
        "iso3_code": "MYS",
    },
    2842: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Alifu Alifu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2843: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Alifu Dhaalu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2844: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Baa",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2845: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Dhaalu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2846: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Faafu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2847: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Gaafu Alifu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2848: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Gaafu Dhaalu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2849: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Haa Alifu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2850: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Haa Dhaalu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2851: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Kaafu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2852: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Laamu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2853: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Lhaviyani",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2854: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Male'",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2855: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Meemu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2856: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Noonu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2857: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Raa",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2858: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Seenu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2859: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Shaviyani",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2860: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Thaa",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2861: {
        "gaul0_name": "Maldives",
        "gaul1_name": "Vaavu",
        "iso2_code": "MV",
        "iso3_code": "MDV",
    },
    2862: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Arkhangai",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2863: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Bayan-Ölgii",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2864: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Bayankhongor",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2865: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Bulgan",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2866: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Darkhan-Uul",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2867: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Dornod",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2868: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Dornogovi",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2869: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Dundgovi",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2870: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Govi-Altai",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2871: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Govisümber",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2872: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Khentii",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2873: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Khovd",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2874: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Khövsgöl",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2875: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Orkhon",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2876: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Selenge",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2877: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Sükhbaatar",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2878: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Töv",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2879: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Ulaanbaatar",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2880: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Uvs",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2881: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Zavkhan",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2882: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Ömnögovi",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2883: {
        "gaul0_name": "Mongolia",
        "gaul1_name": "Övörkhangai",
        "iso2_code": "MN",
        "iso3_code": "MNG",
    },
    2884: {
        "gaul0_name": "Myanmar",
        "gaul1_name": "Region",
        "iso2_code": "MM",
        "iso3_code": "MMR",
    },
    2885: {
        "gaul0_name": "Myanmar",
        "gaul1_name": "State",
        "iso2_code": "MM",
        "iso3_code": "MMR",
    },
    2886: {
        "gaul0_name": "Myanmar",
        "gaul1_name": "Union Territory",
        "iso2_code": "MM",
        "iso3_code": "MMR",
    },
    2887: {
        "gaul0_name": "Nepal",
        "gaul1_name": "Bagmati",
        "iso2_code": "NP",
        "iso3_code": "NPL",
    },
    2888: {
        "gaul0_name": "Nepal",
        "gaul1_name": "Gandaki",
        "iso2_code": "NP",
        "iso3_code": "NPL",
    },
    2889: {
        "gaul0_name": "Nepal",
        "gaul1_name": "Karnali",
        "iso2_code": "NP",
        "iso3_code": "NPL",
    },
    2890: {
        "gaul0_name": "Nepal",
        "gaul1_name": "Koshi",
        "iso2_code": "NP",
        "iso3_code": "NPL",
    },
    2891: {
        "gaul0_name": "Nepal",
        "gaul1_name": "Lumbini",
        "iso2_code": "NP",
        "iso3_code": "NPL",
    },
    2892: {
        "gaul0_name": "Nepal",
        "gaul1_name": "Madhesh",
        "iso2_code": "NP",
        "iso3_code": "NPL",
    },
    2893: {
        "gaul0_name": "Nepal",
        "gaul1_name": "Sudurpaschim",
        "iso2_code": "NP",
        "iso3_code": "NPL",
    },
    2894: {
        "gaul0_name": "Oman",
        "gaul1_name": "Ad Dakhliyah",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2895: {
        "gaul0_name": "Oman",
        "gaul1_name": "Al Batinah North",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2896: {
        "gaul0_name": "Oman",
        "gaul1_name": "Al Batinah South",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2897: {
        "gaul0_name": "Oman",
        "gaul1_name": "Al Buraymi",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2898: {
        "gaul0_name": "Oman",
        "gaul1_name": "Al Dhahira",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2899: {
        "gaul0_name": "Oman",
        "gaul1_name": "Al Wusta",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2900: {
        "gaul0_name": "Oman",
        "gaul1_name": "Ash Sharqiyah North",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2901: {
        "gaul0_name": "Oman",
        "gaul1_name": "Ash Sharqiyah South",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2902: {
        "gaul0_name": "Oman",
        "gaul1_name": "Dhofar",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2903: {
        "gaul0_name": "Oman",
        "gaul1_name": "Musandam",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2904: {
        "gaul0_name": "Oman",
        "gaul1_name": "Muscat",
        "iso2_code": "OM",
        "iso3_code": "OMN",
    },
    2905: {
        "gaul0_name": "Pakistan",
        "gaul1_name": "Balochistan",
        "iso2_code": "PK",
        "iso3_code": "PAK",
    },
    2906: {
        "gaul0_name": "Pakistan",
        "gaul1_name": "Islamabad",
        "iso2_code": "PK",
        "iso3_code": "PAK",
    },
    2907: {
        "gaul0_name": "Pakistan",
        "gaul1_name": "Khyber Pakhtunkhwa",
        "iso2_code": "PK",
        "iso3_code": "PAK",
    },
    2908: {
        "gaul0_name": "Pakistan",
        "gaul1_name": "Punjab",
        "iso2_code": "PK",
        "iso3_code": "PAK",
    },
    2909: {
        "gaul0_name": "Pakistan",
        "gaul1_name": "Sindh",
        "iso2_code": "PK",
        "iso3_code": "PAK",
    },
    2910: {
        "gaul0_name": "Palestine",
        "gaul1_name": "Gaza Strip",
        "iso2_code": "PS",
        "iso3_code": "PSE",
    },
    2911: {
        "gaul0_name": "Palestine",
        "gaul1_name": "West Bank",
        "iso2_code": "PS",
        "iso3_code": "PSE",
    },
    2912: {
        "gaul0_name": "Paracel Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xPI",
    },
    2913: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Autonomous Region In Muslim Mindanao (Armm)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2914: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Cordillera Administrative Region (Car)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2915: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Mimaropa Region",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2916: {
        "gaul0_name": "Philippines",
        "gaul1_name": "National Capital Region (Ncr)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2917: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region I (Ilocos Region)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2918: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region Ii (Cagayan Valley)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2919: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region Iii (Central Luzon)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2920: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region Iv-A (Calabarzon)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2921: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region Ix (Zamboanga Peninsula)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2922: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region V (Bicol Region)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2923: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region Vi (Western Visayas)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2924: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region Vii (Central Visayas)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2925: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region Viii (Eastern Visayas)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2926: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region X (Northern Mindanao)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2927: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region Xi (Davao Region)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2928: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region Xii (Soccsksargen)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2929: {
        "gaul0_name": "Philippines",
        "gaul1_name": "Region Xiii (Caraga)",
        "iso2_code": "PH",
        "iso3_code": "PHL",
    },
    2930: {
        "gaul0_name": "Qatar",
        "gaul1_name": "Al Daayen Municipality",
        "iso2_code": "QA",
        "iso3_code": "QAT",
    },
    2931: {
        "gaul0_name": "Qatar",
        "gaul1_name": "Al Khor And Al Thakhira Municipality",
        "iso2_code": "QA",
        "iso3_code": "QAT",
    },
    2932: {
        "gaul0_name": "Qatar",
        "gaul1_name": "Al Rayyan Municipality",
        "iso2_code": "QA",
        "iso3_code": "QAT",
    },
    2933: {
        "gaul0_name": "Qatar",
        "gaul1_name": "Al Shamal Municipality",
        "iso2_code": "QA",
        "iso3_code": "QAT",
    },
    2934: {
        "gaul0_name": "Qatar",
        "gaul1_name": "Al Sheehaniya Municipality",
        "iso2_code": "QA",
        "iso3_code": "QAT",
    },
    2935: {
        "gaul0_name": "Qatar",
        "gaul1_name": "Al Wakra Municipality",
        "iso2_code": "QA",
        "iso3_code": "QAT",
    },
    2936: {
        "gaul0_name": "Qatar",
        "gaul1_name": "Doha Municipality",
        "iso2_code": "QA",
        "iso3_code": "QAT",
    },
    2937: {
        "gaul0_name": "Qatar",
        "gaul1_name": "Umm Slal Municipality",
        "iso2_code": "QA",
        "iso3_code": "QAT",
    },
    2938: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Cheju-Do",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2939: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Chollabuk-Do",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2940: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Chollanam-Do",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2941: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Chungchongbuk-Do",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2942: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Chungchongnam-Do",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2943: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Inchon",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2944: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Kang-Won-Do",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2945: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Kwangju",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2946: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Kyonggi-Do",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2947: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Kyongsangbuk-Do",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2948: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Kyongsangnam-Do",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2949: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Pusan",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2950: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Seoul",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2951: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Taegu",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2952: {
        "gaul0_name": "Republic Of Korea",
        "gaul1_name": "Taejon",
        "iso2_code": "KR",
        "iso3_code": "KOR",
    },
    2953: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Al Bahah",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2954: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Al Jawf",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2955: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Al Madinah Al Munawwarah",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2956: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Al Qaseem",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2957: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Ar Riyad",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2958: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Aseer",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2959: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Eastern Region",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2960: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Hail",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2961: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Jazan",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2962: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Makkah Al Mukarramah",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2963: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Najran",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2964: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Northern Borders",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2965: {
        "gaul0_name": "Saudi Arabia",
        "gaul1_name": "Tabuk",
        "iso2_code": "SA",
        "iso3_code": "SAU",
    },
    2966: {
        "gaul0_name": "Scarborough Reef",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xSR",
    },
    2967: {
        "gaul0_name": "Senkaku Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xSK",
    },
    2968: {
        "gaul0_name": "Singapore",
        "gaul1_name": "Ang Mo Kio-Cheng San",
        "iso2_code": "SG",
        "iso3_code": "SGP",
    },
    2969: {
        "gaul0_name": "Singapore",
        "gaul1_name": "Bukit Timah",
        "iso2_code": "SG",
        "iso3_code": "SGP",
    },
    2970: {
        "gaul0_name": "Singapore",
        "gaul1_name": "Central Singapore",
        "iso2_code": "SG",
        "iso3_code": "SGP",
    },
    2971: {
        "gaul0_name": "Singapore",
        "gaul1_name": "Hougang",
        "iso2_code": "SG",
        "iso3_code": "SGP",
    },
    2972: {
        "gaul0_name": "Singapore",
        "gaul1_name": "Marine Parade",
        "iso2_code": "SG",
        "iso3_code": "SGP",
    },
    2973: {
        "gaul0_name": "Singapore",
        "gaul1_name": "Northeast",
        "iso2_code": "SG",
        "iso3_code": "SGP",
    },
    2974: {
        "gaul0_name": "Singapore",
        "gaul1_name": "Potong Pasir",
        "iso2_code": "SG",
        "iso3_code": "SGP",
    },
    2975: {
        "gaul0_name": "Singapore",
        "gaul1_name": "Sembawang-Hong Kah",
        "iso2_code": "SG",
        "iso3_code": "SGP",
    },
    2976: {
        "gaul0_name": "Singapore",
        "gaul1_name": "Tanjong Pagar",
        "iso2_code": "SG",
        "iso3_code": "SGP",
    },
    2977: {
        "gaul0_name": "Spratly Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xSI",
    },
    2978: {
        "gaul0_name": "Sri Lanka",
        "gaul1_name": "Central",
        "iso2_code": "LK",
        "iso3_code": "LKA",
    },
    2979: {
        "gaul0_name": "Sri Lanka",
        "gaul1_name": "Eastern",
        "iso2_code": "LK",
        "iso3_code": "LKA",
    },
    2980: {
        "gaul0_name": "Sri Lanka",
        "gaul1_name": "North Central",
        "iso2_code": "LK",
        "iso3_code": "LKA",
    },
    2981: {
        "gaul0_name": "Sri Lanka",
        "gaul1_name": "North Western",
        "iso2_code": "LK",
        "iso3_code": "LKA",
    },
    2982: {
        "gaul0_name": "Sri Lanka",
        "gaul1_name": "Northern",
        "iso2_code": "LK",
        "iso3_code": "LKA",
    },
    2983: {
        "gaul0_name": "Sri Lanka",
        "gaul1_name": "Sabaragamuwa",
        "iso2_code": "LK",
        "iso3_code": "LKA",
    },
    2984: {
        "gaul0_name": "Sri Lanka",
        "gaul1_name": "Southern",
        "iso2_code": "LK",
        "iso3_code": "LKA",
    },
    2985: {
        "gaul0_name": "Sri Lanka",
        "gaul1_name": "Uva",
        "iso2_code": "LK",
        "iso3_code": "LKA",
    },
    2986: {
        "gaul0_name": "Sri Lanka",
        "gaul1_name": "Western",
        "iso2_code": "LK",
        "iso3_code": "LKA",
    },
    2987: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Al-Hasakeh",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2988: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Aleppo",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2989: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Ar-Raqqa",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2990: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "As-Sweida",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2991: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Damascus",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2992: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Dar'A",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2993: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Deir-Ez-Zor",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2994: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Hama",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2995: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Homs",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2996: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Idleb",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2997: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Lattakia",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2998: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Quneitra",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    2999: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Rural Damascus",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    3000: {
        "gaul0_name": "Syrian Arab Republic",
        "gaul1_name": "Tartous",
        "iso2_code": "SY",
        "iso3_code": "SYR",
    },
    3001: {
        "gaul0_name": "Taiwan Province of China",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "TW",
        "iso3_code": "TWN",
    },
    3002: {
        "gaul0_name": "Tajikistan",
        "gaul1_name": "Districts Of Republican Subordination",
        "iso2_code": "TJ",
        "iso3_code": "TJK",
    },
    3003: {
        "gaul0_name": "Tajikistan",
        "gaul1_name": "Dushanbe",
        "iso2_code": "TJ",
        "iso3_code": "TJK",
    },
    3004: {
        "gaul0_name": "Tajikistan",
        "gaul1_name": "Gorno-Badakhshan Autonomous Region",
        "iso2_code": "TJ",
        "iso3_code": "TJK",
    },
    3005: {
        "gaul0_name": "Tajikistan",
        "gaul1_name": "Khatlon Region",
        "iso2_code": "TJ",
        "iso3_code": "TJK",
    },
    3006: {
        "gaul0_name": "Tajikistan",
        "gaul1_name": "Sughd Region",
        "iso2_code": "TJ",
        "iso3_code": "TJK",
    },
    3007: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Amnat Charoen",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3008: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Ang Thong",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3009: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Bangkok",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3010: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Bueng Kan",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3011: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Buri Ram",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3012: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Chachoengsao",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3013: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Chai Nat",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3014: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Chaiyaphum",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3015: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Chanthaburi",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3016: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Chiang Mai",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3017: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Chiang Rai",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3018: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Chon Buri",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3019: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Chumphon",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3020: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Kalasin",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3021: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Kamphaeng Phet",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3022: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Kanchanaburi",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3023: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Khon Kaen",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3024: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Krabi",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3025: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Lampang",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3026: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Lamphun",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3027: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Loei",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3028: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Lop Buri",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3029: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Mae Hong Son",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3030: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Maha Sarakham",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3031: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Mukdahan",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3032: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Nakhon Nayok",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3033: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Nakhon Pathom",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3034: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Nakhon Phanom",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3035: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Nakhon Ratchasima",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3036: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Nakhon Sawan",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3037: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Nakhon Si Thammarat",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3038: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Nan",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3039: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Narathiwat",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3040: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Nong Bua Lam Phu",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3041: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Nong Khai",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3042: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Nonthaburi",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3043: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Pathum Thani",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3044: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Pattani",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3045: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Phangnga",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3046: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Phatthalung",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3047: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Phayao",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3048: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Phetchabun",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3049: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Phetchaburi",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3050: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Phichit",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3051: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Phitsanulok",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3052: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Phra Nakhon Si Ayutthaya",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3053: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Phrae",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3054: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Phuket",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3055: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Prachin Buri",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3056: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Prachuap Khiri Khan",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3057: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Ranong",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3058: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Ratchaburi",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3059: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Rayong",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3060: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Roi Et",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3061: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Sa Kaeo",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3062: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Sakon Nakhon",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3063: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Samut Prakan",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3064: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Samut Sakhon",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3065: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Samut Songkhram",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3066: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Saraburi",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3067: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Satun",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3068: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Si Sa Ket",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3069: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Sing Buri",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3070: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Songkhla",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3071: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Sukhothai",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3072: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Suphan Buri",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3073: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Surat Thani",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3074: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Surin",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3075: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Tak",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3076: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Trang",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3077: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Trat",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3078: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Ubon Ratchathani",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3079: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Udon Thani",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3080: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Uthai Thani",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3081: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Uttaradit",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3082: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Yala",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3083: {
        "gaul0_name": "Thailand",
        "gaul1_name": "Yasothon",
        "iso2_code": "TH",
        "iso3_code": "THA",
    },
    3084: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Aileu",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3085: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Ainaro",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3086: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Baucau",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3087: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Bobonaro",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3088: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Covalima",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3089: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Dili",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3090: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Ermera",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3091: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Lautém",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3092: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Liquiçá",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3093: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Manatuto",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3094: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Manufahi",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3095: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Oecussi",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3096: {
        "gaul0_name": "Timor-Leste",
        "gaul1_name": "Viqueque",
        "iso2_code": "TL",
        "iso3_code": "TLS",
    },
    3097: {
        "gaul0_name": "Turkmenistan",
        "gaul1_name": "Ahal",
        "iso2_code": "TM",
        "iso3_code": "TKM",
    },
    3098: {
        "gaul0_name": "Turkmenistan",
        "gaul1_name": "Balkan",
        "iso2_code": "TM",
        "iso3_code": "TKM",
    },
    3099: {
        "gaul0_name": "Turkmenistan",
        "gaul1_name": "Chardzhou",
        "iso2_code": "TM",
        "iso3_code": "TKM",
    },
    3100: {
        "gaul0_name": "Turkmenistan",
        "gaul1_name": "Mary",
        "iso2_code": "TM",
        "iso3_code": "TKM",
    },
    3101: {
        "gaul0_name": "Turkmenistan",
        "gaul1_name": "Tashauz",
        "iso2_code": "TM",
        "iso3_code": "TKM",
    },
    3102: {
        "gaul0_name": "Turkmenistan",
        "gaul1_name": "Waterbody",
        "iso2_code": "TM",
        "iso3_code": "TKM",
    },
    3103: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Adana",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3104: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Adiyaman",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3105: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Afyonkarahisar",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3106: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Agri",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3107: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Aksaray",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3108: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Amasya",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3109: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Ankara",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3110: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Antalya",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3111: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Ardahan",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3112: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Artvin",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3113: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Aydin",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3114: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Balikesir",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3115: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Bartin",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3116: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Batman",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3117: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Bayburt",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3118: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Bilecik",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3119: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Bingol",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3120: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Bitlis",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3121: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Bolu",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3122: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Burdur",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3123: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Bursa",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3124: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Canakkale",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3125: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Cankiri",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3126: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Corum",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3127: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Denizli",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3128: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Diyarbakir",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3129: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Duzce",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3130: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Edirne",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3131: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Elazig",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3132: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Erzincan",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3133: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Erzurum",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3134: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Eskisehir",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3135: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Gaziantep",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3136: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Giresun",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3137: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Gumushane",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3138: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Hakkari",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3139: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Hatay",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3140: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Igdir",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3141: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Isparta",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3142: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Istanbul",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3143: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Izmir",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3144: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Kahramanmaras",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3145: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Karabuk",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3146: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Karaman",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3147: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Kars",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3148: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Kastamonu",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3149: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Kayseri",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3150: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Kilis",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3151: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Kirikkale",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3152: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Kirklareli",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3153: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Kirsehir",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3154: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Kocaeli",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3155: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Konya",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3156: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Kutahya",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3157: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Malatya",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3158: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Manisa",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3159: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Mardin",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3160: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Mersin",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3161: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Mugla",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3162: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Mus",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3163: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Nevsehir",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3164: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Nigde",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3165: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Ordu",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3166: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Osmaniye",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3167: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Rize",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3168: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Sakarya",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3169: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Samsun",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3170: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Sanliurfa",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3171: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Siirt",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3172: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Sinop",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3173: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Sirnak",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3174: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Sivas",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3175: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Tekirdag",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3176: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Tokat",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3177: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Trabzon",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3178: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Tunceli",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3179: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Usak",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3180: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Van",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3181: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Yalova",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3182: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Yozgat",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3183: {
        "gaul0_name": "Türkiye",
        "gaul1_name": "Zonguldak",
        "iso2_code": "TR",
        "iso3_code": "TUR",
    },
    3184: {
        "gaul0_name": "United Arab Emirates",
        "gaul1_name": "Abu Dhabi",
        "iso2_code": "AE",
        "iso3_code": "ARE",
    },
    3185: {
        "gaul0_name": "United Arab Emirates",
        "gaul1_name": "Ajman",
        "iso2_code": "AE",
        "iso3_code": "ARE",
    },
    3186: {
        "gaul0_name": "United Arab Emirates",
        "gaul1_name": "Dubai",
        "iso2_code": "AE",
        "iso3_code": "ARE",
    },
    3187: {
        "gaul0_name": "United Arab Emirates",
        "gaul1_name": "Fujairah",
        "iso2_code": "AE",
        "iso3_code": "ARE",
    },
    3188: {
        "gaul0_name": "United Arab Emirates",
        "gaul1_name": "Ras Al Khaimah",
        "iso2_code": "AE",
        "iso3_code": "ARE",
    },
    3189: {
        "gaul0_name": "United Arab Emirates",
        "gaul1_name": "Sharjah",
        "iso2_code": "AE",
        "iso3_code": "ARE",
    },
    3190: {
        "gaul0_name": "United Arab Emirates",
        "gaul1_name": "Umm Al Quwain",
        "iso2_code": "AE",
        "iso3_code": "ARE",
    },
    3191: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Andizhan",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3192: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Bukhara",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3193: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Dzhizak",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3194: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Fergana",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3195: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Kashkadarya",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3196: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Khorezm",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3197: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Namangan",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3198: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Navoi",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3199: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Republic Of Karakalpakstan",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3200: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Samarkand",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3201: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Surkhandarya",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3202: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Syrdarya",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3203: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Tashkent",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3204: {
        "gaul0_name": "Uzbekistan",
        "gaul1_name": "Tashkent City",
        "iso2_code": "UZ",
        "iso3_code": "UZB",
    },
    3205: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "An Giang",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3206: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Ba Ria - Vung Tau",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3207: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Bac Giang",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3208: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Bac Kan",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3209: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Bac Lieu",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3210: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Bac Ninh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3211: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Ben Tre",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3212: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Binh Dinh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3213: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Binh Duong",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3214: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Binh Phuoc",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3215: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Binh Thuan",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3216: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Ca Mau",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3217: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Can Tho City",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3218: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Cao Bang",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3219: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Da Nang City",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3220: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Dak Lak",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3221: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Dak Nong",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3222: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Dien Bien",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3223: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Dong Nai",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3224: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Dong Thap",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3225: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Gia Lai",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3226: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Ha Giang",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3227: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Ha Nam",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3228: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Ha Noi",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3229: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Ha Tinh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3230: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Hai Duong",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3231: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Hai Phong City",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3232: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Hau Giang",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3233: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Ho Chi Minh City",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3234: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Hoa Binh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3235: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Hung Yen",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3236: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Khanh Hoa",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3237: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Kien Giang",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3238: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Kon Tum",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3239: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Lai Chau",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3240: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Lam Dong",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3241: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Lang Son",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3242: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Lao Cai",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3243: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Long An",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3244: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Nam Dinh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3245: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Nghe An",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3246: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Ninh Binh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3247: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Ninh Thuan",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3248: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Phu Tho",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3249: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Phu Yen",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3250: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Quang Binh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3251: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Quang Nam",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3252: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Quang Ngai",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3253: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Quang Ninh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3254: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Quang Tri",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3255: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Soc Trang",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3256: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Son La",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3257: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Tay Ninh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3258: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Thai Binh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3259: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Thai Nguyen",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3260: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Thanh Hoa",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3261: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Thua Thien Hue",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3262: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Tien Giang",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3263: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Tra Vinh",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3264: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Tuyen Quang",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3265: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Vinh Long",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3266: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Vinh Phuc",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3267: {
        "gaul0_name": "Viet Nam",
        "gaul1_name": "Yen Bai",
        "iso2_code": "VN",
        "iso3_code": "VNM",
    },
    3268: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Abyan",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3269: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Ad Dali'",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3270: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Aden",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3271: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Al Bayda",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3272: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Al Hodeidah",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3273: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Al Jawf",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3274: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Al Maharah",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3275: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Al Mahwit",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3276: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Amran",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3277: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Dhamar",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3278: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Hadramawt",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3279: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Hajjah",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3280: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Ibb",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3281: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Lahj",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3282: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Ma'Rib",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3283: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Raymah",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3284: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Sa'Dah",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3285: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Sana'A",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3286: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Sana'A City",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3287: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Shabwah",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3288: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Socotra",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3289: {
        "gaul0_name": "Yemen",
        "gaul1_name": "Ta'Iz",
        "iso2_code": "YE",
        "iso3_code": "YEM",
    },
    3290: {
        "gaul0_name": "Akrotiri",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xUK",
    },
    3291: {
        "gaul0_name": "Albania",
        "gaul1_name": "Berat",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3292: {
        "gaul0_name": "Albania",
        "gaul1_name": "Diber",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3293: {
        "gaul0_name": "Albania",
        "gaul1_name": "Durres",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3294: {
        "gaul0_name": "Albania",
        "gaul1_name": "Elbasan",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3295: {
        "gaul0_name": "Albania",
        "gaul1_name": "Fier",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3296: {
        "gaul0_name": "Albania",
        "gaul1_name": "Gjirokaster",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3297: {
        "gaul0_name": "Albania",
        "gaul1_name": "Korce",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3298: {
        "gaul0_name": "Albania",
        "gaul1_name": "Kukes",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3299: {
        "gaul0_name": "Albania",
        "gaul1_name": "Lezhe",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3300: {
        "gaul0_name": "Albania",
        "gaul1_name": "Shkoder",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3301: {
        "gaul0_name": "Albania",
        "gaul1_name": "Tirane",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3302: {
        "gaul0_name": "Albania",
        "gaul1_name": "Vlore",
        "iso2_code": "AL",
        "iso3_code": "ALB",
    },
    3303: {
        "gaul0_name": "Andorra",
        "gaul1_name": "Andorra La Vella",
        "iso2_code": "AD",
        "iso3_code": "AND",
    },
    3304: {
        "gaul0_name": "Andorra",
        "gaul1_name": "Canillo",
        "iso2_code": "AD",
        "iso3_code": "AND",
    },
    3305: {
        "gaul0_name": "Andorra",
        "gaul1_name": "Encamp",
        "iso2_code": "AD",
        "iso3_code": "AND",
    },
    3306: {
        "gaul0_name": "Andorra",
        "gaul1_name": "Escaldes-engordany",
        "iso2_code": "AD",
        "iso3_code": "AND",
    },
    3307: {
        "gaul0_name": "Andorra",
        "gaul1_name": "La Massana",
        "iso2_code": "AD",
        "iso3_code": "AND",
    },
    3308: {
        "gaul0_name": "Andorra",
        "gaul1_name": "Ordino",
        "iso2_code": "AD",
        "iso3_code": "AND",
    },
    3309: {
        "gaul0_name": "Andorra",
        "gaul1_name": "Sant Julia De Loria",
        "iso2_code": "AD",
        "iso3_code": "AND",
    },
    3310: {
        "gaul0_name": "Austria",
        "gaul1_name": "Burgenland",
        "iso2_code": "AT",
        "iso3_code": "AUT",
    },
    3311: {
        "gaul0_name": "Austria",
        "gaul1_name": "Kärnten",
        "iso2_code": "AT",
        "iso3_code": "AUT",
    },
    3312: {
        "gaul0_name": "Austria",
        "gaul1_name": "Niederöster",
        "iso2_code": "AT",
        "iso3_code": "AUT",
    },
    3313: {
        "gaul0_name": "Austria",
        "gaul1_name": "Oberösterre",
        "iso2_code": "AT",
        "iso3_code": "AUT",
    },
    3314: {
        "gaul0_name": "Austria",
        "gaul1_name": "Salzburg",
        "iso2_code": "AT",
        "iso3_code": "AUT",
    },
    3315: {
        "gaul0_name": "Austria",
        "gaul1_name": "Steiermark",
        "iso2_code": "AT",
        "iso3_code": "AUT",
    },
    3316: {
        "gaul0_name": "Austria",
        "gaul1_name": "Tirol",
        "iso2_code": "AT",
        "iso3_code": "AUT",
    },
    3317: {
        "gaul0_name": "Austria",
        "gaul1_name": "Vorarlberg",
        "iso2_code": "AT",
        "iso3_code": "AUT",
    },
    3318: {
        "gaul0_name": "Austria",
        "gaul1_name": "Wien",
        "iso2_code": "AT",
        "iso3_code": "AUT",
    },
    3319: {
        "gaul0_name": "Belarus",
        "gaul1_name": "Brest",
        "iso2_code": "BY",
        "iso3_code": "BLR",
    },
    3320: {
        "gaul0_name": "Belarus",
        "gaul1_name": "Gomel",
        "iso2_code": "BY",
        "iso3_code": "BLR",
    },
    3321: {
        "gaul0_name": "Belarus",
        "gaul1_name": "Grodno",
        "iso2_code": "BY",
        "iso3_code": "BLR",
    },
    3322: {
        "gaul0_name": "Belarus",
        "gaul1_name": "Minsk",
        "iso2_code": "BY",
        "iso3_code": "BLR",
    },
    3323: {
        "gaul0_name": "Belarus",
        "gaul1_name": "Minsk City",
        "iso2_code": "BY",
        "iso3_code": "BLR",
    },
    3324: {
        "gaul0_name": "Belarus",
        "gaul1_name": "Mogilev",
        "iso2_code": "BY",
        "iso3_code": "BLR",
    },
    3325: {
        "gaul0_name": "Belarus",
        "gaul1_name": "Vitebsk",
        "iso2_code": "BY",
        "iso3_code": "BLR",
    },
    3326: {
        "gaul0_name": "Belgium",
        "gaul1_name": "Brussels-Capital Region",
        "iso2_code": "BE",
        "iso3_code": "BEL",
    },
    3327: {
        "gaul0_name": "Belgium",
        "gaul1_name": "Flemish Region",
        "iso2_code": "BE",
        "iso3_code": "BEL",
    },
    3328: {
        "gaul0_name": "Belgium",
        "gaul1_name": "Walloon Region",
        "iso2_code": "BE",
        "iso3_code": "BEL",
    },
    3329: {
        "gaul0_name": "Bosnia and Herzegovina",
        "gaul1_name": "Brcko District",
        "iso2_code": "BA",
        "iso3_code": "BIH",
    },
    3330: {
        "gaul0_name": "Bosnia and Herzegovina",
        "gaul1_name": "Federation of Bosnia and Herzegovina",
        "iso2_code": "BA",
        "iso3_code": "BIH",
    },
    3331: {
        "gaul0_name": "Bosnia and Herzegovina",
        "gaul1_name": "Republika Srpska",
        "iso2_code": "BA",
        "iso3_code": "BIH",
    },
    3332: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Blagoevgrad",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3333: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Burgas",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3334: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Dobrich",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3335: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Gabrovo",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3336: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Haskovo",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3337: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Kardzhali",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3338: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Kyustendil",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3339: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Lovech",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3340: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Montana",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3341: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Pazardzhik",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3342: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Pernik",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3343: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Pleven",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3344: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Plovdiv",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3345: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Razgrad",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3346: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Ruse",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3347: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Shumen",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3348: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Silistra",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3349: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Sliven",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3350: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Smolyan",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3351: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Sofia",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3352: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Sofia-City",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3353: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Stara Zagora",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3354: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Targovishte",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3355: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Varna",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3356: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Veliko Tarnovo",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3357: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Vidin",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3358: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Vratsa",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3359: {
        "gaul0_name": "Bulgaria",
        "gaul1_name": "Yambol",
        "iso2_code": "BG",
        "iso3_code": "BGR",
    },
    3360: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Bjelovar",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3361: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Dubrovnik",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3362: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Gospić",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3363: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Grad Zagreb",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3364: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Karlovac",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3365: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Koprivnica",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3366: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Krapina",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3367: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Osijek",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3368: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Pazin",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3369: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Požega",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3370: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Rijeka",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3371: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Sisak",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3372: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Slavonski Brod",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3373: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Split",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3374: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Varaždin",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3375: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Virovitica",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3376: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Vukovar",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3377: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Zadar",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3378: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Zagreb",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3379: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Čakovec",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3380: {
        "gaul0_name": "Croatia",
        "gaul1_name": "Šibenik",
        "iso2_code": "HR",
        "iso3_code": "HRV",
    },
    3381: {
        "gaul0_name": "Curaçao",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "CW",
        "iso3_code": "CUW",
    },
    3382: {
        "gaul0_name": "Cyprus",
        "gaul1_name": "Ammochostos",
        "iso2_code": "CY",
        "iso3_code": "CYP",
    },
    3383: {
        "gaul0_name": "Cyprus",
        "gaul1_name": "Keryneia",
        "iso2_code": "CY",
        "iso3_code": "CYP",
    },
    3384: {
        "gaul0_name": "Cyprus",
        "gaul1_name": "Larnaka",
        "iso2_code": "CY",
        "iso3_code": "CYP",
    },
    3385: {
        "gaul0_name": "Cyprus",
        "gaul1_name": "Lefkosia",
        "iso2_code": "CY",
        "iso3_code": "CYP",
    },
    3386: {
        "gaul0_name": "Cyprus",
        "gaul1_name": "Lemesos",
        "iso2_code": "CY",
        "iso3_code": "CYP",
    },
    3387: {
        "gaul0_name": "Cyprus",
        "gaul1_name": "Pafos",
        "iso2_code": "CY",
        "iso3_code": "CYP",
    },
    3388: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Jihomoravský",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3389: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Jihočeský",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3390: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Karlovarský",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3391: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Královéhradecký",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3392: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Liberecký",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3393: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Moravskoslezský",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3394: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Olomoucký",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3395: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Pardubický",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3396: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Plzeňský",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3397: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Praha",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3398: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Středočeský",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3399: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Vysočina",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3400: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Zlínský",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3401: {
        "gaul0_name": "Czechia",
        "gaul1_name": "Ústecký",
        "iso2_code": "CZ",
        "iso3_code": "CZE",
    },
    3402: {
        "gaul0_name": "Dekelia",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "not found",
        "iso3_code": "xUK",
    },
    3403: {
        "gaul0_name": "Denmark",
        "gaul1_name": "Hovedstaden",
        "iso2_code": "DK",
        "iso3_code": "DNK",
    },
    3404: {
        "gaul0_name": "Denmark",
        "gaul1_name": "Midtjylland",
        "iso2_code": "DK",
        "iso3_code": "DNK",
    },
    3405: {
        "gaul0_name": "Denmark",
        "gaul1_name": "Nordjylland",
        "iso2_code": "DK",
        "iso3_code": "DNK",
    },
    3406: {
        "gaul0_name": "Denmark",
        "gaul1_name": "Sjælland",
        "iso2_code": "DK",
        "iso3_code": "DNK",
    },
    3407: {
        "gaul0_name": "Denmark",
        "gaul1_name": "Syddanmark",
        "iso2_code": "DK",
        "iso3_code": "DNK",
    },
    3408: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Harju maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3409: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Hiiu maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3410: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Ida-Viru maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3411: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Järva maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3412: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Jõgeva maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3413: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Lääne maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3414: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Lääne-Viru maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3415: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Pärnu maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3416: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Põlva maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3417: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Rapla maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3418: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Saare maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3419: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Tartu maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3420: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Valga maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3421: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Viljandi maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3422: {
        "gaul0_name": "Estonia",
        "gaul1_name": "Võru maakond",
        "iso2_code": "EE",
        "iso3_code": "EST",
    },
    3423: {
        "gaul0_name": "Faroe Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "FO",
        "iso3_code": "FRO",
    },
    3424: {
        "gaul0_name": "Finland",
        "gaul1_name": "Ahvenanmaan valtionvirasto",
        "iso2_code": "FI",
        "iso3_code": "FIN",
    },
    3425: {
        "gaul0_name": "Finland",
        "gaul1_name": "Etelä-Suomen aluehallintovirasto",
        "iso2_code": "FI",
        "iso3_code": "FIN",
    },
    3426: {
        "gaul0_name": "Finland",
        "gaul1_name": "Itä-Suomen aluehallintovirasto",
        "iso2_code": "FI",
        "iso3_code": "FIN",
    },
    3427: {
        "gaul0_name": "Finland",
        "gaul1_name": "Lapin aluehallintovirasto",
        "iso2_code": "FI",
        "iso3_code": "FIN",
    },
    3428: {
        "gaul0_name": "Finland",
        "gaul1_name": "Lounais-Suomen aluehallintovirasto",
        "iso2_code": "FI",
        "iso3_code": "FIN",
    },
    3429: {
        "gaul0_name": "Finland",
        "gaul1_name": "Länsi- ja Sisä-Suomen aluehallintovirasto",
        "iso2_code": "FI",
        "iso3_code": "FIN",
    },
    3430: {
        "gaul0_name": "Finland",
        "gaul1_name": "Pohjois-Suomen aluehallintovirasto",
        "iso2_code": "FI",
        "iso3_code": "FIN",
    },
    3431: {
        "gaul0_name": "France",
        "gaul1_name": "Auvergne-Rhône-Alpes",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3432: {
        "gaul0_name": "France",
        "gaul1_name": "Bourgogne-Franche-Comté",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3433: {
        "gaul0_name": "France",
        "gaul1_name": "Bretagne",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3434: {
        "gaul0_name": "France",
        "gaul1_name": "Centre-Val de Loire",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3435: {
        "gaul0_name": "France",
        "gaul1_name": "Corse",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3436: {
        "gaul0_name": "France",
        "gaul1_name": "Grand Est",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3437: {
        "gaul0_name": "France",
        "gaul1_name": "Hauts-de-France",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3438: {
        "gaul0_name": "France",
        "gaul1_name": "Normandie",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3439: {
        "gaul0_name": "France",
        "gaul1_name": "Nouvelle-Aquitaine",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3440: {
        "gaul0_name": "France",
        "gaul1_name": "Occitanie",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3441: {
        "gaul0_name": "France",
        "gaul1_name": "Pays de la Loire",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3442: {
        "gaul0_name": "France",
        "gaul1_name": "Provence-Alpes-Côte d'Azur",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3443: {
        "gaul0_name": "France",
        "gaul1_name": "Île-de-France",
        "iso2_code": "FR",
        "iso3_code": "FRA",
    },
    3444: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Abkhazeti (Abkhazia) Autonomous Republic",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3445: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Adjara (Adjaria) Autonomous Republic",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3446: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Guria",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3447: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Imereti",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3448: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Kakheti",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3449: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Kvemo Kartli",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3450: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Mtskheta-Mtianeti",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3451: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Racha-Lechkhumi-Kvemo Svaneti",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3452: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Samegrelo-Zemo Svaneti",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3453: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Samtskhe-Javakheti",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3454: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Shida Kartli",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3455: {
        "gaul0_name": "Georgia",
        "gaul1_name": "Tbilisi (City)",
        "iso2_code": "GE",
        "iso3_code": "GEO",
    },
    3456: {
        "gaul0_name": "Germany",
        "gaul1_name": "Baden-Württemberg",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3457: {
        "gaul0_name": "Germany",
        "gaul1_name": "Bayern",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3458: {
        "gaul0_name": "Germany",
        "gaul1_name": "Berlin",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3459: {
        "gaul0_name": "Germany",
        "gaul1_name": "Brandenburg",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3460: {
        "gaul0_name": "Germany",
        "gaul1_name": "Bremen",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3461: {
        "gaul0_name": "Germany",
        "gaul1_name": "Hamburg",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3462: {
        "gaul0_name": "Germany",
        "gaul1_name": "Hessen",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3463: {
        "gaul0_name": "Germany",
        "gaul1_name": "Mecklenburg-Vorpommern",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3464: {
        "gaul0_name": "Germany",
        "gaul1_name": "Niedersachsen",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3465: {
        "gaul0_name": "Germany",
        "gaul1_name": "Nordrhein-Westfalen",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3466: {
        "gaul0_name": "Germany",
        "gaul1_name": "Rheinland-Pfalz",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3467: {
        "gaul0_name": "Germany",
        "gaul1_name": "Saarland",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3468: {
        "gaul0_name": "Germany",
        "gaul1_name": "Sachsen",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3469: {
        "gaul0_name": "Germany",
        "gaul1_name": "Sachsen-Anhalt",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3470: {
        "gaul0_name": "Germany",
        "gaul1_name": "Schleswig-Holstein",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3471: {
        "gaul0_name": "Germany",
        "gaul1_name": "Thüringen",
        "iso2_code": "DE",
        "iso3_code": "DEU",
    },
    3472: {
        "gaul0_name": "Gibraltar",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "GI",
        "iso3_code": "GIB",
    },
    3473: {
        "gaul0_name": "Greece",
        "gaul1_name": "Aigaioy",
        "iso2_code": "GR",
        "iso3_code": "GRC",
    },
    3474: {
        "gaul0_name": "Greece",
        "gaul1_name": "Attikis",
        "iso2_code": "GR",
        "iso3_code": "GRC",
    },
    3475: {
        "gaul0_name": "Greece",
        "gaul1_name": "Ipeiroy Dytikis Makedonias",
        "iso2_code": "GR",
        "iso3_code": "GRC",
    },
    3476: {
        "gaul0_name": "Greece",
        "gaul1_name": "Kritis",
        "iso2_code": "GR",
        "iso3_code": "GRC",
    },
    3477: {
        "gaul0_name": "Greece",
        "gaul1_name": "Makedonias Thrakis",
        "iso2_code": "GR",
        "iso3_code": "GRC",
    },
    3478: {
        "gaul0_name": "Greece",
        "gaul1_name": "Peloponnisoy Dytikis Elladas kai Ionioy Nison",
        "iso2_code": "GR",
        "iso3_code": "GRC",
    },
    3479: {
        "gaul0_name": "Greece",
        "gaul1_name": "Thessalias Stereas Elladas",
        "iso2_code": "GR",
        "iso3_code": "GRC",
    },
    3480: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Alderney",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3481: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Castel",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3482: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Forest",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3483: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Herm",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3484: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Saint Andrew",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3485: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Saint Martin",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3486: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Saint Peter",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3487: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Saint Peter Port",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3488: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Saint Sampson",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3489: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Saint Saviour",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3490: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Sark",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3491: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Torteval",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3492: {
        "gaul0_name": "Guernsey",
        "gaul1_name": "Vale",
        "iso2_code": "GG",
        "iso3_code": "GGY",
    },
    3493: {
        "gaul0_name": "Holy See",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "VA",
        "iso3_code": "VAT",
    },
    3494: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Bacs Kiskun",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3495: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Baranya",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3496: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Bekes",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3497: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Borsod Abauj Zemplen",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3498: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Budapest",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3499: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Csongrad",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3500: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Fejer",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3501: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Gyor Moson Sopron",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3502: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Hajdu Bihar",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3503: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Heves",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3504: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Jasz Nagykun Szolnok",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3505: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Komarom Esztergom",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3506: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Nograd",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3507: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Pest megye",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3508: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Somogy",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3509: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Szabolcs Szatmar Bereg",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3510: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Tolna",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3511: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Vas",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3512: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Veszprem",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3513: {
        "gaul0_name": "Hungary",
        "gaul1_name": "Zala",
        "iso2_code": "HU",
        "iso3_code": "HUN",
    },
    3514: {
        "gaul0_name": "Iceland",
        "gaul1_name": "Austurland",
        "iso2_code": "IS",
        "iso3_code": "ISL",
    },
    3515: {
        "gaul0_name": "Iceland",
        "gaul1_name": "Höfuðborgarsvæðið",
        "iso2_code": "IS",
        "iso3_code": "ISL",
    },
    3516: {
        "gaul0_name": "Iceland",
        "gaul1_name": "Norðurland eystra",
        "iso2_code": "IS",
        "iso3_code": "ISL",
    },
    3517: {
        "gaul0_name": "Iceland",
        "gaul1_name": "Norðurland vestra",
        "iso2_code": "IS",
        "iso3_code": "ISL",
    },
    3518: {
        "gaul0_name": "Iceland",
        "gaul1_name": "Suðurland",
        "iso2_code": "IS",
        "iso3_code": "ISL",
    },
    3519: {
        "gaul0_name": "Iceland",
        "gaul1_name": "Suðurnes",
        "iso2_code": "IS",
        "iso3_code": "ISL",
    },
    3520: {
        "gaul0_name": "Iceland",
        "gaul1_name": "Vestfirðir",
        "iso2_code": "IS",
        "iso3_code": "ISL",
    },
    3521: {
        "gaul0_name": "Iceland",
        "gaul1_name": "Vesturland",
        "iso2_code": "IS",
        "iso3_code": "ISL",
    },
    3522: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Carlow",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3523: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Cavan",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3524: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Clare",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3525: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Cork",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3526: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Donegal",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3527: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Dublin",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3528: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Galway",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3529: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Kerry",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3530: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Kildare",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3531: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Kilkenny",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3532: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Laois",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3533: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Leitrim",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3534: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Limerick",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3535: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Longford",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3536: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Louth",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3537: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Mayo",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3538: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Meath",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3539: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Monaghan",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3540: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Offaly",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3541: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Roscommon",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3542: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Sligo",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3543: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Tipperary",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3544: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Waterford",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3545: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Westmeath",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3546: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Wexford",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3547: {
        "gaul0_name": "Ireland",
        "gaul1_name": "Wicklow",
        "iso2_code": "IE",
        "iso3_code": "IRL",
    },
    3548: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Andreas",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3549: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Arbory",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3550: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Ballaugh",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3551: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Braddan",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3552: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Bride",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3553: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Castletown",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3554: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Douglas",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3555: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "German",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3556: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Jurby",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3557: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Laxey",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3558: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Lezayre",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3559: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Lonan",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3560: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Malew",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3561: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Marown",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3562: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Maughold",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3563: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Michael",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3564: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Onchan",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3565: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Patrick",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3566: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Peel",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3567: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Port Erin",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3568: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Port St.Mary",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3569: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Ramsey",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3570: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Rushen",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3571: {
        "gaul0_name": "Isle of Man",
        "gaul1_name": "Santan",
        "iso2_code": "IM",
        "iso3_code": "IMN",
    },
    3572: {
        "gaul0_name": "Italy",
        "gaul1_name": "Abruzzo",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3573: {
        "gaul0_name": "Italy",
        "gaul1_name": "Basilicata",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3574: {
        "gaul0_name": "Italy",
        "gaul1_name": "Calabria",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3575: {
        "gaul0_name": "Italy",
        "gaul1_name": "Campania",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3576: {
        "gaul0_name": "Italy",
        "gaul1_name": "Emilia-Romagna",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3577: {
        "gaul0_name": "Italy",
        "gaul1_name": "Friuli-Venezia Giulia",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3578: {
        "gaul0_name": "Italy",
        "gaul1_name": "Lazio",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3579: {
        "gaul0_name": "Italy",
        "gaul1_name": "Liguria",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3580: {
        "gaul0_name": "Italy",
        "gaul1_name": "Lombardia",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3581: {
        "gaul0_name": "Italy",
        "gaul1_name": "Marche",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3582: {
        "gaul0_name": "Italy",
        "gaul1_name": "Molise",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3583: {
        "gaul0_name": "Italy",
        "gaul1_name": "Piemonte",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3584: {
        "gaul0_name": "Italy",
        "gaul1_name": "Puglia",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3585: {
        "gaul0_name": "Italy",
        "gaul1_name": "Sardegna",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3586: {
        "gaul0_name": "Italy",
        "gaul1_name": "Sicilia",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3587: {
        "gaul0_name": "Italy",
        "gaul1_name": "Toscana",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3588: {
        "gaul0_name": "Italy",
        "gaul1_name": "Trentino-Alto Adige",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3589: {
        "gaul0_name": "Italy",
        "gaul1_name": "Umbria",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3590: {
        "gaul0_name": "Italy",
        "gaul1_name": "Valle d'Aosta",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3591: {
        "gaul0_name": "Italy",
        "gaul1_name": "Veneto",
        "iso2_code": "IT",
        "iso3_code": "ITA",
    },
    3592: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Brelade",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3593: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Grouville",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3594: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Saint Clement",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3595: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Saint Helier",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3596: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Saint John",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3597: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Saint Lawrence",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3598: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Saint Martin",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3599: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Saint Mary",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3600: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Saint Ouen",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3601: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Saint Peter",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3602: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Saint Saviour",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3603: {
        "gaul0_name": "Jersey",
        "gaul1_name": "Trinity",
        "iso2_code": "JE",
        "iso3_code": "JEY",
    },
    3604: {
        "gaul0_name": "Latvia",
        "gaul1_name": "Kurzeme",
        "iso2_code": "LV",
        "iso3_code": "LVA",
    },
    3605: {
        "gaul0_name": "Latvia",
        "gaul1_name": "Latgale",
        "iso2_code": "LV",
        "iso3_code": "LVA",
    },
    3606: {
        "gaul0_name": "Latvia",
        "gaul1_name": "Rīga",
        "iso2_code": "LV",
        "iso3_code": "LVA",
    },
    3607: {
        "gaul0_name": "Latvia",
        "gaul1_name": "Vidzeme",
        "iso2_code": "LV",
        "iso3_code": "LVA",
    },
    3608: {
        "gaul0_name": "Latvia",
        "gaul1_name": "Zemgale",
        "iso2_code": "LV",
        "iso3_code": "LVA",
    },
    3609: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Balzers",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3610: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Eschen",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3611: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Gamprin",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3612: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Mauren",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3613: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Planken",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3614: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Ruggell",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3615: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Schaan",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3616: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Schellenberg",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3617: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Triesen",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3618: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Triesenberg",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3619: {
        "gaul0_name": "Liechtenstein",
        "gaul1_name": "Vaduz",
        "iso2_code": "LI",
        "iso3_code": "LIE",
    },
    3620: {
        "gaul0_name": "Lithuania",
        "gaul1_name": "Alytaus apskr.",
        "iso2_code": "LT",
        "iso3_code": "LTU",
    },
    3621: {
        "gaul0_name": "Lithuania",
        "gaul1_name": "Kauno apskr.",
        "iso2_code": "LT",
        "iso3_code": "LTU",
    },
    3622: {
        "gaul0_name": "Lithuania",
        "gaul1_name": "Klaipėdos apskr.",
        "iso2_code": "LT",
        "iso3_code": "LTU",
    },
    3623: {
        "gaul0_name": "Lithuania",
        "gaul1_name": "Marijampolės apskr.",
        "iso2_code": "LT",
        "iso3_code": "LTU",
    },
    3624: {
        "gaul0_name": "Lithuania",
        "gaul1_name": "Panevėžio apskr.",
        "iso2_code": "LT",
        "iso3_code": "LTU",
    },
    3625: {
        "gaul0_name": "Lithuania",
        "gaul1_name": "Tauragės apskr.",
        "iso2_code": "LT",
        "iso3_code": "LTU",
    },
    3626: {
        "gaul0_name": "Lithuania",
        "gaul1_name": "Telšių apskr.",
        "iso2_code": "LT",
        "iso3_code": "LTU",
    },
    3627: {
        "gaul0_name": "Lithuania",
        "gaul1_name": "Utenos apskr.",
        "iso2_code": "LT",
        "iso3_code": "LTU",
    },
    3628: {
        "gaul0_name": "Lithuania",
        "gaul1_name": "Vilniaus apskr.",
        "iso2_code": "LT",
        "iso3_code": "LTU",
    },
    3629: {
        "gaul0_name": "Lithuania",
        "gaul1_name": "Šiaulių apskr.",
        "iso2_code": "LT",
        "iso3_code": "LTU",
    },
    3630: {
        "gaul0_name": "Luxembourg",
        "gaul1_name": "Diekirch",
        "iso2_code": "LU",
        "iso3_code": "LUX",
    },
    3631: {
        "gaul0_name": "Luxembourg",
        "gaul1_name": "Grevenmacher",
        "iso2_code": "LU",
        "iso3_code": "LUX",
    },
    3632: {
        "gaul0_name": "Luxembourg",
        "gaul1_name": "Luxembourg",
        "iso2_code": "LU",
        "iso3_code": "LUX",
    },
    3633: {
        "gaul0_name": "Malta",
        "gaul1_name": "Gozo Region",
        "iso2_code": "MT",
        "iso3_code": "MLT",
    },
    3634: {
        "gaul0_name": "Malta",
        "gaul1_name": "Northern Harbour Reg",
        "iso2_code": "MT",
        "iso3_code": "MLT",
    },
    3635: {
        "gaul0_name": "Malta",
        "gaul1_name": "Northern Region",
        "iso2_code": "MT",
        "iso3_code": "MLT",
    },
    3636: {
        "gaul0_name": "Malta",
        "gaul1_name": "South Eastern Region",
        "iso2_code": "MT",
        "iso3_code": "MLT",
    },
    3637: {
        "gaul0_name": "Malta",
        "gaul1_name": "Southern Harbour Reg",
        "iso2_code": "MT",
        "iso3_code": "MLT",
    },
    3638: {
        "gaul0_name": "Malta",
        "gaul1_name": "Western Region",
        "iso2_code": "MT",
        "iso3_code": "MLT",
    },
    3639: {
        "gaul0_name": "Monaco",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "MC",
        "iso3_code": "MCO",
    },
    3640: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Andrijevica",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3641: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Bar",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3642: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Berane",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3643: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Bijelo Polje",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3644: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Budva",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3645: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Cetinje",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3646: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Danilovgrad",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3647: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Herceg Novi",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3648: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Kolasin",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3649: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Kotor",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3650: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Mojkovac",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3651: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Niksic",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3652: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Plav",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3653: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Pljevlja",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3654: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Pluzine",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3655: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Podgorica",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3656: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Rozaj",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3657: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Savnik",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3658: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Tivat",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3659: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Ulcinj",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3660: {
        "gaul0_name": "Montenegro",
        "gaul1_name": "Zabljak",
        "iso2_code": "ME",
        "iso3_code": "MNE",
    },
    3661: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Drenthe",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3662: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Flevoland",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3663: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Fryslân",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3664: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Gelderland",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3665: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Groningen",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3666: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Limburg",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3667: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Noord-Brabant",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3668: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Noord-Holland",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3669: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Overijssel",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3670: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Utrecht",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3671: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Zeeland",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3672: {
        "gaul0_name": "Netherlands (Kingdom of the)",
        "gaul1_name": "Zuid-Holland",
        "iso2_code": "NL",
        "iso3_code": "NLD",
    },
    3673: {
        "gaul0_name": "North Macedonia",
        "gaul1_name": "East",
        "iso2_code": "MK",
        "iso3_code": "MKD",
    },
    3674: {
        "gaul0_name": "North Macedonia",
        "gaul1_name": "Northeast",
        "iso2_code": "MK",
        "iso3_code": "MKD",
    },
    3675: {
        "gaul0_name": "North Macedonia",
        "gaul1_name": "Pelagonia",
        "iso2_code": "MK",
        "iso3_code": "MKD",
    },
    3676: {
        "gaul0_name": "North Macedonia",
        "gaul1_name": "Polog",
        "iso2_code": "MK",
        "iso3_code": "MKD",
    },
    3677: {
        "gaul0_name": "North Macedonia",
        "gaul1_name": "Skopje",
        "iso2_code": "MK",
        "iso3_code": "MKD",
    },
    3678: {
        "gaul0_name": "North Macedonia",
        "gaul1_name": "Southeast",
        "iso2_code": "MK",
        "iso3_code": "MKD",
    },
    3679: {
        "gaul0_name": "North Macedonia",
        "gaul1_name": "Southwest",
        "iso2_code": "MK",
        "iso3_code": "MKD",
    },
    3680: {
        "gaul0_name": "North Macedonia",
        "gaul1_name": "Vardar",
        "iso2_code": "MK",
        "iso3_code": "MKD",
    },
    3681: {
        "gaul0_name": "Norway",
        "gaul1_name": "Agder",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3682: {
        "gaul0_name": "Norway",
        "gaul1_name": "Akershus",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3683: {
        "gaul0_name": "Norway",
        "gaul1_name": "Buskerud",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3684: {
        "gaul0_name": "Norway",
        "gaul1_name": "Finnmark - Finnmárku - Finmarkku",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3685: {
        "gaul0_name": "Norway",
        "gaul1_name": "Innlandet",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3686: {
        "gaul0_name": "Norway",
        "gaul1_name": "Møre og Romsdal",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3687: {
        "gaul0_name": "Norway",
        "gaul1_name": "Nordland - Nordlánnda",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3688: {
        "gaul0_name": "Norway",
        "gaul1_name": "Oslo",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3689: {
        "gaul0_name": "Norway",
        "gaul1_name": "Rogaland",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3690: {
        "gaul0_name": "Norway",
        "gaul1_name": "Telemark",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3691: {
        "gaul0_name": "Norway",
        "gaul1_name": "Troms - Romsa - Tromssa",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3692: {
        "gaul0_name": "Norway",
        "gaul1_name": "Trøndelag - Trööndelage",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3693: {
        "gaul0_name": "Norway",
        "gaul1_name": "Vestfold",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3694: {
        "gaul0_name": "Norway",
        "gaul1_name": "Vestland",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3695: {
        "gaul0_name": "Norway",
        "gaul1_name": "Østfold",
        "iso2_code": "NO",
        "iso3_code": "NOR",
    },
    3696: {
        "gaul0_name": "Poland",
        "gaul1_name": "Dolnośląskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3697: {
        "gaul0_name": "Poland",
        "gaul1_name": "Kujawsko-Pomorskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3698: {
        "gaul0_name": "Poland",
        "gaul1_name": "Lubelskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3699: {
        "gaul0_name": "Poland",
        "gaul1_name": "Lubuskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3700: {
        "gaul0_name": "Poland",
        "gaul1_name": "Mazowieckie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3701: {
        "gaul0_name": "Poland",
        "gaul1_name": "Małopolskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3702: {
        "gaul0_name": "Poland",
        "gaul1_name": "Opolskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3703: {
        "gaul0_name": "Poland",
        "gaul1_name": "Podkarpackie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3704: {
        "gaul0_name": "Poland",
        "gaul1_name": "Podlaskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3705: {
        "gaul0_name": "Poland",
        "gaul1_name": "Pomorskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3706: {
        "gaul0_name": "Poland",
        "gaul1_name": "Warmińsko-Mazurskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3707: {
        "gaul0_name": "Poland",
        "gaul1_name": "Wielkopolskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3708: {
        "gaul0_name": "Poland",
        "gaul1_name": "Zachodniopomorskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3709: {
        "gaul0_name": "Poland",
        "gaul1_name": "Łódzkie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3710: {
        "gaul0_name": "Poland",
        "gaul1_name": "Śląskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3711: {
        "gaul0_name": "Poland",
        "gaul1_name": "Świętokrzyskie",
        "iso2_code": "PL",
        "iso3_code": "POL",
    },
    3712: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Aveiro",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3713: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Beja",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3714: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Braga",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3715: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Bragança",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3716: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Castelo Branco",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3717: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Coimbra",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3718: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Faro",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3719: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Guarda",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3720: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Ilha Terceira",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3721: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Ilha da Graciosa",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3722: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Ilha da Madeira (Madeira)",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3723: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Ilha de Porto Santo (Madeira)",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3724: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Ilha de São Jorge",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3725: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Ilha do Faial",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3726: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Ilha do Pico",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3727: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Leiria",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3728: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Lisboa",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3729: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Portalegre",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3730: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Porto",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3731: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Santarém",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3732: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Setúbal",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3733: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Viana do Castelo",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3734: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Vila Real",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3735: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Viseu",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3736: {
        "gaul0_name": "Portugal",
        "gaul1_name": "Évora",
        "iso2_code": "PT",
        "iso3_code": "PRT",
    },
    3737: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Anenii Noi",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3738: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Balti",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3739: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Basarabeasca",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3740: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Bender",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3741: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Briceni",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3742: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Cahul",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3743: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Calarasi",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3744: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Cantemir",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3745: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Causeni",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3746: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Chisinau",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3747: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Cimislia",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3748: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Criuleni",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3749: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Donduseni",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3750: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Drochia",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3751: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Dubasari",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3752: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Edinet",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3753: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Falesti",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3754: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Floresti",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3755: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Glodeni",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3756: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Hincesti",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3757: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Ialoveni",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3758: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Leova",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3759: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Nisporeni",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3760: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Ocnita",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3761: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Orhei",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3762: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Rezina",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3763: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Riscani",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3764: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Singerei",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3765: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Soldanesti",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3766: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Soroca",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3767: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Stefan Voda",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3768: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Straseni",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3769: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Taraclia",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3770: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Telenesti",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3771: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Transnistria",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3772: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Ungheni",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3773: {
        "gaul0_name": "Republic of Moldova",
        "gaul1_name": "Unitate Teritoriala Autonoma Gagauzia",
        "iso2_code": "MD",
        "iso3_code": "MDA",
    },
    3774: {
        "gaul0_name": "Romania",
        "gaul1_name": "Alba",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3775: {
        "gaul0_name": "Romania",
        "gaul1_name": "Arad",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3776: {
        "gaul0_name": "Romania",
        "gaul1_name": "Arges",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3777: {
        "gaul0_name": "Romania",
        "gaul1_name": "Bacau",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3778: {
        "gaul0_name": "Romania",
        "gaul1_name": "Bihor",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3779: {
        "gaul0_name": "Romania",
        "gaul1_name": "Bistrita-Nasaud",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3780: {
        "gaul0_name": "Romania",
        "gaul1_name": "Botosani",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3781: {
        "gaul0_name": "Romania",
        "gaul1_name": "Braila",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3782: {
        "gaul0_name": "Romania",
        "gaul1_name": "Brasov",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3783: {
        "gaul0_name": "Romania",
        "gaul1_name": "Bucuresti",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3784: {
        "gaul0_name": "Romania",
        "gaul1_name": "Buzau",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3785: {
        "gaul0_name": "Romania",
        "gaul1_name": "Calarasi",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3786: {
        "gaul0_name": "Romania",
        "gaul1_name": "Caras-Severin",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3787: {
        "gaul0_name": "Romania",
        "gaul1_name": "Cluj",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3788: {
        "gaul0_name": "Romania",
        "gaul1_name": "Constanta",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3789: {
        "gaul0_name": "Romania",
        "gaul1_name": "Covasna",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3790: {
        "gaul0_name": "Romania",
        "gaul1_name": "Dambovita",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3791: {
        "gaul0_name": "Romania",
        "gaul1_name": "Dolj",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3792: {
        "gaul0_name": "Romania",
        "gaul1_name": "Galati",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3793: {
        "gaul0_name": "Romania",
        "gaul1_name": "Giurgiu",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3794: {
        "gaul0_name": "Romania",
        "gaul1_name": "Gorj",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3795: {
        "gaul0_name": "Romania",
        "gaul1_name": "Harghita",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3796: {
        "gaul0_name": "Romania",
        "gaul1_name": "Hunedoara",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3797: {
        "gaul0_name": "Romania",
        "gaul1_name": "Ialomita",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3798: {
        "gaul0_name": "Romania",
        "gaul1_name": "Iasi",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3799: {
        "gaul0_name": "Romania",
        "gaul1_name": "Ilfov",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3800: {
        "gaul0_name": "Romania",
        "gaul1_name": "Maramures",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3801: {
        "gaul0_name": "Romania",
        "gaul1_name": "Mehedinti",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3802: {
        "gaul0_name": "Romania",
        "gaul1_name": "Mures",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3803: {
        "gaul0_name": "Romania",
        "gaul1_name": "Neamt",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3804: {
        "gaul0_name": "Romania",
        "gaul1_name": "Olt",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3805: {
        "gaul0_name": "Romania",
        "gaul1_name": "Prahova",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3806: {
        "gaul0_name": "Romania",
        "gaul1_name": "Salaj",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3807: {
        "gaul0_name": "Romania",
        "gaul1_name": "Satu Mare",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3808: {
        "gaul0_name": "Romania",
        "gaul1_name": "Sibiu",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3809: {
        "gaul0_name": "Romania",
        "gaul1_name": "Suceava",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3810: {
        "gaul0_name": "Romania",
        "gaul1_name": "Teleorman",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3811: {
        "gaul0_name": "Romania",
        "gaul1_name": "Timis",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3812: {
        "gaul0_name": "Romania",
        "gaul1_name": "Tulcea",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3813: {
        "gaul0_name": "Romania",
        "gaul1_name": "Valcea",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3814: {
        "gaul0_name": "Romania",
        "gaul1_name": "Vaslui",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3815: {
        "gaul0_name": "Romania",
        "gaul1_name": "Vrancea",
        "iso2_code": "RO",
        "iso3_code": "ROU",
    },
    3816: {
        "gaul0_name": "Russian Federation",
        "gaul1_name": "Central Federal District",
        "iso2_code": "RU",
        "iso3_code": "RUS",
    },
    3817: {
        "gaul0_name": "Russian Federation",
        "gaul1_name": "Far Eastern Federal District",
        "iso2_code": "RU",
        "iso3_code": "RUS",
    },
    3818: {
        "gaul0_name": "Russian Federation",
        "gaul1_name": "North Caucasus federal district",
        "iso2_code": "RU",
        "iso3_code": "RUS",
    },
    3819: {
        "gaul0_name": "Russian Federation",
        "gaul1_name": "Northwestern Federal District",
        "iso2_code": "RU",
        "iso3_code": "RUS",
    },
    3820: {
        "gaul0_name": "Russian Federation",
        "gaul1_name": "Siberian Federal District",
        "iso2_code": "RU",
        "iso3_code": "RUS",
    },
    3821: {
        "gaul0_name": "Russian Federation",
        "gaul1_name": "South federal district",
        "iso2_code": "RU",
        "iso3_code": "RUS",
    },
    3822: {
        "gaul0_name": "Russian Federation",
        "gaul1_name": "Ural Federal District",
        "iso2_code": "RU",
        "iso3_code": "RUS",
    },
    3823: {
        "gaul0_name": "Russian Federation",
        "gaul1_name": "Volga Federal District",
        "iso2_code": "RU",
        "iso3_code": "RUS",
    },
    3824: {
        "gaul0_name": "Russian Federation",
        "gaul1_name": "Waterbody",
        "iso2_code": "RU",
        "iso3_code": "RUS",
    },
    3825: {
        "gaul0_name": "San Marino",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "SM",
        "iso3_code": "SMR",
    },
    3826: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Borski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3827: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Braničevski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3828: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Grad Beograd",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3829: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Jablanički",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3830: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Južno-Banatski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3831: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Južno-Bački",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3832: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Kolubarski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3833: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Kosovo",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3834: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Mačvanski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3835: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Moravički",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3836: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Nišavski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3837: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Pirotski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3838: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Podunavski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3839: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Pomoravski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3840: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Pčinjski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3841: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Rasinski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3842: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Raški",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3843: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Severno-Banatski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3844: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Severno-Bački",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3845: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Srednje-Banatski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3846: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Sremski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3847: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Toplički",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3848: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Zaječarski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3849: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Zapadno-Bački",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3850: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Zlatiborski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3851: {
        "gaul0_name": "Serbia",
        "gaul1_name": "Šumadijski",
        "iso2_code": "RS",
        "iso3_code": "SRB",
    },
    3852: {
        "gaul0_name": "Sint Maarten (Dutch part)",
        "gaul1_name": "Cole Bay",
        "iso2_code": "SX",
        "iso3_code": "SXM",
    },
    3853: {
        "gaul0_name": "Sint Maarten (Dutch part)",
        "gaul1_name": "Cul-de-Sac",
        "iso2_code": "SX",
        "iso3_code": "SXM",
    },
    3854: {
        "gaul0_name": "Sint Maarten (Dutch part)",
        "gaul1_name": "Little Bay Area",
        "iso2_code": "SX",
        "iso3_code": "SXM",
    },
    3855: {
        "gaul0_name": "Sint Maarten (Dutch part)",
        "gaul1_name": "Low Lands",
        "iso2_code": "SX",
        "iso3_code": "SXM",
    },
    3856: {
        "gaul0_name": "Sint Maarten (Dutch part)",
        "gaul1_name": "Lower Prince's Quarter",
        "iso2_code": "SX",
        "iso3_code": "SXM",
    },
    3857: {
        "gaul0_name": "Sint Maarten (Dutch part)",
        "gaul1_name": "Philipsburg",
        "iso2_code": "SX",
        "iso3_code": "SXM",
    },
    3858: {
        "gaul0_name": "Sint Maarten (Dutch part)",
        "gaul1_name": "Simpson Bay",
        "iso2_code": "SX",
        "iso3_code": "SXM",
    },
    3859: {
        "gaul0_name": "Sint Maarten (Dutch part)",
        "gaul1_name": "Upper Prince's Quarter",
        "iso2_code": "SX",
        "iso3_code": "SXM",
    },
    3860: {
        "gaul0_name": "Slovakia",
        "gaul1_name": "Banskobystrický kraj",
        "iso2_code": "SK",
        "iso3_code": "SVK",
    },
    3861: {
        "gaul0_name": "Slovakia",
        "gaul1_name": "Bratislavský kraj",
        "iso2_code": "SK",
        "iso3_code": "SVK",
    },
    3862: {
        "gaul0_name": "Slovakia",
        "gaul1_name": "Košický kraj",
        "iso2_code": "SK",
        "iso3_code": "SVK",
    },
    3863: {
        "gaul0_name": "Slovakia",
        "gaul1_name": "Nitriansky kraj",
        "iso2_code": "SK",
        "iso3_code": "SVK",
    },
    3864: {
        "gaul0_name": "Slovakia",
        "gaul1_name": "Prešovský kraj",
        "iso2_code": "SK",
        "iso3_code": "SVK",
    },
    3865: {
        "gaul0_name": "Slovakia",
        "gaul1_name": "Trenčiansky kraj",
        "iso2_code": "SK",
        "iso3_code": "SVK",
    },
    3866: {
        "gaul0_name": "Slovakia",
        "gaul1_name": "Trnavský kraj",
        "iso2_code": "SK",
        "iso3_code": "SVK",
    },
    3867: {
        "gaul0_name": "Slovakia",
        "gaul1_name": "Žilinský kraj",
        "iso2_code": "SK",
        "iso3_code": "SVK",
    },
    3868: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Gorenjska",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3869: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Goriška",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3870: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Jugovzhodna Slovenija",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3871: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Koroška",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3872: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Obalno-kraška",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3873: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Osrednjeslovenska",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3874: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Podravska",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3875: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Pomurska",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3876: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Posavska",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3877: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Primorsko-notranjska",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3878: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Savinjska",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3879: {
        "gaul0_name": "Slovenia",
        "gaul1_name": "Zasavska",
        "iso2_code": "SI",
        "iso3_code": "SVN",
    },
    3880: {
        "gaul0_name": "Spain",
        "gaul1_name": "Andalucía",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3881: {
        "gaul0_name": "Spain",
        "gaul1_name": "Aragón",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3882: {
        "gaul0_name": "Spain",
        "gaul1_name": "Canarias",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3883: {
        "gaul0_name": "Spain",
        "gaul1_name": "Cantabria",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3884: {
        "gaul0_name": "Spain",
        "gaul1_name": "Castilla y León",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3885: {
        "gaul0_name": "Spain",
        "gaul1_name": "Castilla-La Mancha",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3886: {
        "gaul0_name": "Spain",
        "gaul1_name": "Cataluña",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3887: {
        "gaul0_name": "Spain",
        "gaul1_name": "Ceuta",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3888: {
        "gaul0_name": "Spain",
        "gaul1_name": "Comunidad Foral de Navarra",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3889: {
        "gaul0_name": "Spain",
        "gaul1_name": "Comunidad de Madrid",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3890: {
        "gaul0_name": "Spain",
        "gaul1_name": "Comunitat Valenciana",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3891: {
        "gaul0_name": "Spain",
        "gaul1_name": "Extremadura",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3892: {
        "gaul0_name": "Spain",
        "gaul1_name": "Galicia",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3893: {
        "gaul0_name": "Spain",
        "gaul1_name": "Illes Balears",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3894: {
        "gaul0_name": "Spain",
        "gaul1_name": "La Rioja",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3895: {
        "gaul0_name": "Spain",
        "gaul1_name": "Melilla",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3896: {
        "gaul0_name": "Spain",
        "gaul1_name": "País Vasco",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3897: {
        "gaul0_name": "Spain",
        "gaul1_name": "Principado de Asturias",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3898: {
        "gaul0_name": "Spain",
        "gaul1_name": "Región de Murcia",
        "iso2_code": "ES",
        "iso3_code": "ESP",
    },
    3899: {
        "gaul0_name": "Svalbard and Jan Mayen Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "SJ",
        "iso3_code": "SJM",
    },
    3900: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Blekinge",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3901: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Dalarna",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3902: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Gotland",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3903: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Gävleborg",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3904: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Halland",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3905: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Jämtland",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3906: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Jönköping",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3907: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Kalmar",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3908: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Kronoberg",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3909: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Norrbotten",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3910: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Skåne",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3911: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Stockholm",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3912: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Södermanland",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3913: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Uppsala",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3914: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Värmland",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3915: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Västerbotten",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3916: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Västernorrland",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3917: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Västmanland",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3918: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Västra Götaland",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3919: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Örebro",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3920: {
        "gaul0_name": "Sweden",
        "gaul1_name": "Östergötland",
        "iso2_code": "SE",
        "iso3_code": "SWE",
    },
    3921: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Aargau",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3922: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Appenzell Ausserrhoden",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3923: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Appenzell Innerrhoden",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3924: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Basel-Landschaft",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3925: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Basel-Stadt",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3926: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Bern",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3927: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Fribourg",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3928: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Genève",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3929: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Glarus",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3930: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Graubünden",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3931: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Jura",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3932: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Luzern",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3933: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Neuchâtel",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3934: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Nidwalden",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3935: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Obwalden",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3936: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Schaffhausen",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3937: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Schwyz",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3938: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Solothurn",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3939: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "St. Gallen",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3940: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Thurgau",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3941: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Ticino",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3942: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Uri",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3943: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Valais",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3944: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Vaud",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3945: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Zug",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3946: {
        "gaul0_name": "Switzerland",
        "gaul1_name": "Zürich",
        "iso2_code": "CH",
        "iso3_code": "CHE",
    },
    3947: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Avtonomna Respublika Krym",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3948: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Cherkaska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3949: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Chernihivska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3950: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Chernivetska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3951: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Dnipropetrovska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3952: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Donetska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3953: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Ivano-Frankivska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3954: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Kharkivska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3955: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Khersonska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3956: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Khmelnytska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3957: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Kirovohradska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3958: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Kyivska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3959: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Luhanska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3960: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Lvivska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3961: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Mykolaivska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3962: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Odeska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3963: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Poltavska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3964: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Rivnenska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3965: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Sevastopilska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3966: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Sumska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3967: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Ternopilska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3968: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Vinnytska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3969: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Volynska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3970: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Zakarpatska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3971: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Zaporizka",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3972: {
        "gaul0_name": "Ukraine",
        "gaul1_name": "Zhytomyrska",
        "iso2_code": "UA",
        "iso3_code": "UKR",
    },
    3973: {
        "gaul0_name": "United Kingdom of Great Britain and Northern Ireland",
        "gaul1_name": "England",
        "iso2_code": "GB",
        "iso3_code": "GBR",
    },
    3974: {
        "gaul0_name": "United Kingdom of Great Britain and Northern Ireland",
        "gaul1_name": "Northern Ireland",
        "iso2_code": "GB",
        "iso3_code": "GBR",
    },
    3975: {
        "gaul0_name": "United Kingdom of Great Britain and Northern Ireland",
        "gaul1_name": "Scotland",
        "iso2_code": "GB",
        "iso3_code": "GBR",
    },
    3976: {
        "gaul0_name": "United Kingdom of Great Britain and Northern Ireland",
        "gaul1_name": "Wales",
        "iso2_code": "GB",
        "iso3_code": "GBR",
    },
    3977: {
        "gaul0_name": "Åland Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "AX",
        "iso3_code": "ALA",
    },
    3978: {
        "gaul0_name": "American Samoa",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "AS",
        "iso3_code": "ASM",
    },
    3979: {
        "gaul0_name": "Ashmore and Cartier Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "AU",
        "iso3_code": "AUS",
    },
    3980: {
        "gaul0_name": "Australia",
        "gaul1_name": "Australian Capital Territory",
        "iso2_code": "AU",
        "iso3_code": "AUS",
    },
    3981: {
        "gaul0_name": "Australia",
        "gaul1_name": "New South Wales",
        "iso2_code": "AU",
        "iso3_code": "AUS",
    },
    3982: {
        "gaul0_name": "Australia",
        "gaul1_name": "Northern Territory",
        "iso2_code": "AU",
        "iso3_code": "AUS",
    },
    3983: {
        "gaul0_name": "Australia",
        "gaul1_name": "Other Territories",
        "iso2_code": "AU",
        "iso3_code": "AUS",
    },
    3984: {
        "gaul0_name": "Australia",
        "gaul1_name": "Queensland",
        "iso2_code": "AU",
        "iso3_code": "AUS",
    },
    3985: {
        "gaul0_name": "Australia",
        "gaul1_name": "South Australia",
        "iso2_code": "AU",
        "iso3_code": "AUS",
    },
    3986: {
        "gaul0_name": "Australia",
        "gaul1_name": "Tasmania",
        "iso2_code": "AU",
        "iso3_code": "AUS",
    },
    3987: {
        "gaul0_name": "Australia",
        "gaul1_name": "Victoria",
        "iso2_code": "AU",
        "iso3_code": "AUS",
    },
    3988: {
        "gaul0_name": "Australia",
        "gaul1_name": "Western Australia",
        "iso2_code": "AU",
        "iso3_code": "AUS",
    },
    3989: {
        "gaul0_name": "Christmas Island",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "CX",
        "iso3_code": "CXR",
    },
    3990: {
        "gaul0_name": "Clipperton Island",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "PF",
        "iso3_code": "PYF",
    },
    3991: {
        "gaul0_name": "Cocos (Keeling) Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "CC",
        "iso3_code": "CCK",
    },
    3992: {
        "gaul0_name": "Cook Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "CK",
        "iso3_code": "COK",
    },
    3993: {
        "gaul0_name": "Fiji",
        "gaul1_name": "Central Division",
        "iso2_code": "FJ",
        "iso3_code": "FJI",
    },
    3994: {
        "gaul0_name": "Fiji",
        "gaul1_name": "Eastern Division",
        "iso2_code": "FJ",
        "iso3_code": "FJI",
    },
    3995: {
        "gaul0_name": "Fiji",
        "gaul1_name": "Northern  Division",
        "iso2_code": "FJ",
        "iso3_code": "FJI",
    },
    3996: {
        "gaul0_name": "Fiji",
        "gaul1_name": "Western Division",
        "iso2_code": "FJ",
        "iso3_code": "FJI",
    },
    3997: {
        "gaul0_name": "French Polynesia",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "PF",
        "iso3_code": "PYF",
    },
    3998: {
        "gaul0_name": "Guam",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "GU",
        "iso3_code": "GUM",
    },
    3999: {
        "gaul0_name": "Kiribati",
        "gaul1_name": "Central",
        "iso2_code": "KI",
        "iso3_code": "KIR",
    },
    4000: {
        "gaul0_name": "Kiribati",
        "gaul1_name": "Line Is. & Phoenix",
        "iso2_code": "KI",
        "iso3_code": "KIR",
    },
    4001: {
        "gaul0_name": "Kiribati",
        "gaul1_name": "Northern",
        "iso2_code": "KI",
        "iso3_code": "KIR",
    },
    4002: {
        "gaul0_name": "Kiribati",
        "gaul1_name": "South Tarawa",
        "iso2_code": "KI",
        "iso3_code": "KIR",
    },
    4003: {
        "gaul0_name": "Kiribati",
        "gaul1_name": "Southern",
        "iso2_code": "KI",
        "iso3_code": "KIR",
    },
    4004: {
        "gaul0_name": "Marshall Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "MH",
        "iso3_code": "MHL",
    },
    4005: {
        "gaul0_name": "Micronesia (Federated States of)",
        "gaul1_name": "Chuuk",
        "iso2_code": "FM",
        "iso3_code": "FSM",
    },
    4006: {
        "gaul0_name": "Micronesia (Federated States of)",
        "gaul1_name": "Kosrae",
        "iso2_code": "FM",
        "iso3_code": "FSM",
    },
    4007: {
        "gaul0_name": "Micronesia (Federated States of)",
        "gaul1_name": "Pohnpei",
        "iso2_code": "FM",
        "iso3_code": "FSM",
    },
    4008: {
        "gaul0_name": "Micronesia (Federated States of)",
        "gaul1_name": "Yap",
        "iso2_code": "FM",
        "iso3_code": "FSM",
    },
    4009: {
        "gaul0_name": "New Caledonia",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "NC",
        "iso3_code": "NCL",
    },
    4010: {
        "gaul0_name": "Nauru",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "NR",
        "iso3_code": "NRU",
    },
    4011: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Area Outside Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4012: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Auckland Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4013: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Bay of Plenty Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4014: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Canterbury Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4015: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Gisborne Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4016: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Hawke's Bay Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4017: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Manawatū-Whanganui Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4018: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Marlborough Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4019: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Nelson Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4020: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Northland Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4021: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Otago Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4022: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Southland Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4023: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Taranaki Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4024: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Tasman Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4025: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Waikato Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4026: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "Wellington Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4027: {
        "gaul0_name": "New Zealand",
        "gaul1_name": "West Coast Region",
        "iso2_code": "NZ",
        "iso3_code": "NZL",
    },
    4028: {
        "gaul0_name": "Niue",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "NU",
        "iso3_code": "NIU",
    },
    4029: {
        "gaul0_name": "Norfolk Island",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "NF",
        "iso3_code": "NFK",
    },
    4030: {
        "gaul0_name": "Northern Mariana Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "MP",
        "iso3_code": "MNP",
    },
    4031: {
        "gaul0_name": "Palau",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "PW",
        "iso3_code": "PLW",
    },
    4032: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Autonomous Region of Bougainville",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4033: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Central Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4034: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Chimbu (Simbu) Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4035: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "East New Britain Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4036: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "East Sepik Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4037: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Eastern Highlands Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4038: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Enga Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4039: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Gulf Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4040: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Hela Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4041: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Jiwaka Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4042: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Madang Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4043: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Manus Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4044: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Milne Bay Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4045: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Morobe Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4046: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "National Capital District",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4047: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "New Ireland Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4048: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Northern (Oro) Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4049: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Southern Highlands Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4050: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "West New Britain Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4051: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "West Sepik (Sandaun) Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4052: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Western Highlands Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4053: {
        "gaul0_name": "Papua New Guinea",
        "gaul1_name": "Western Province",
        "iso2_code": "PG",
        "iso3_code": "PNG",
    },
    4054: {
        "gaul0_name": "Pitcairn",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "PN",
        "iso3_code": "PCN",
    },
    4055: {
        "gaul0_name": "Samoa",
        "gaul1_name": "A'ana",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4056: {
        "gaul0_name": "Samoa",
        "gaul1_name": "Aiga-i-le-Tai",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4057: {
        "gaul0_name": "Samoa",
        "gaul1_name": "Atua",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4058: {
        "gaul0_name": "Samoa",
        "gaul1_name": "Fa'asaleleage",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4059: {
        "gaul0_name": "Samoa",
        "gaul1_name": "Gaga'emauga",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4060: {
        "gaul0_name": "Samoa",
        "gaul1_name": "Gaga'ifomauga",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4061: {
        "gaul0_name": "Samoa",
        "gaul1_name": "Palauli",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4062: {
        "gaul0_name": "Samoa",
        "gaul1_name": "Satupa'itea",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4063: {
        "gaul0_name": "Samoa",
        "gaul1_name": "Tuamasaga",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4064: {
        "gaul0_name": "Samoa",
        "gaul1_name": "Va'a-o-Fonoti",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4065: {
        "gaul0_name": "Samoa",
        "gaul1_name": "Vaisigano",
        "iso2_code": "WS",
        "iso3_code": "WSM",
    },
    4066: {
        "gaul0_name": "Solomon Islands",
        "gaul1_name": "Central",
        "iso2_code": "SB",
        "iso3_code": "SLB",
    },
    4067: {
        "gaul0_name": "Solomon Islands",
        "gaul1_name": "Choiseul",
        "iso2_code": "SB",
        "iso3_code": "SLB",
    },
    4068: {
        "gaul0_name": "Solomon Islands",
        "gaul1_name": "Guadalcanal",
        "iso2_code": "SB",
        "iso3_code": "SLB",
    },
    4069: {
        "gaul0_name": "Solomon Islands",
        "gaul1_name": "Honiara",
        "iso2_code": "SB",
        "iso3_code": "SLB",
    },
    4070: {
        "gaul0_name": "Solomon Islands",
        "gaul1_name": "Isabel",
        "iso2_code": "SB",
        "iso3_code": "SLB",
    },
    4071: {
        "gaul0_name": "Solomon Islands",
        "gaul1_name": "Makira-Ulawa",
        "iso2_code": "SB",
        "iso3_code": "SLB",
    },
    4072: {
        "gaul0_name": "Solomon Islands",
        "gaul1_name": "Malaita",
        "iso2_code": "SB",
        "iso3_code": "SLB",
    },
    4073: {
        "gaul0_name": "Solomon Islands",
        "gaul1_name": "Rennell-Bell",
        "iso2_code": "SB",
        "iso3_code": "SLB",
    },
    4074: {
        "gaul0_name": "Solomon Islands",
        "gaul1_name": "Temotu",
        "iso2_code": "SB",
        "iso3_code": "SLB",
    },
    4075: {
        "gaul0_name": "Solomon Islands",
        "gaul1_name": "Western",
        "iso2_code": "SB",
        "iso3_code": "SLB",
    },
    4076: {
        "gaul0_name": "Tokelau",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "TK",
        "iso3_code": "TKL",
    },
    4077: {
        "gaul0_name": "Tonga",
        "gaul1_name": "'Eua",
        "iso2_code": "TO",
        "iso3_code": "TON",
    },
    4078: {
        "gaul0_name": "Tonga",
        "gaul1_name": "Ha'apai",
        "iso2_code": "TO",
        "iso3_code": "TON",
    },
    4079: {
        "gaul0_name": "Tonga",
        "gaul1_name": "Niuas",
        "iso2_code": "TO",
        "iso3_code": "TON",
    },
    4080: {
        "gaul0_name": "Tonga",
        "gaul1_name": "Tongatapu",
        "iso2_code": "TO",
        "iso3_code": "TON",
    },
    4081: {
        "gaul0_name": "Tonga",
        "gaul1_name": "Vava'u",
        "iso2_code": "TO",
        "iso3_code": "TON",
    },
    4082: {
        "gaul0_name": "Tuvalu",
        "gaul1_name": "Funafuti",
        "iso2_code": "TV",
        "iso3_code": "TUV",
    },
    4083: {
        "gaul0_name": "Tuvalu",
        "gaul1_name": "Nanumaga",
        "iso2_code": "TV",
        "iso3_code": "TUV",
    },
    4084: {
        "gaul0_name": "Tuvalu",
        "gaul1_name": "Nanumea",
        "iso2_code": "TV",
        "iso3_code": "TUV",
    },
    4085: {
        "gaul0_name": "Tuvalu",
        "gaul1_name": "Niulakita",
        "iso2_code": "TV",
        "iso3_code": "TUV",
    },
    4086: {
        "gaul0_name": "Tuvalu",
        "gaul1_name": "Niutao",
        "iso2_code": "TV",
        "iso3_code": "TUV",
    },
    4087: {
        "gaul0_name": "Tuvalu",
        "gaul1_name": "Nui",
        "iso2_code": "TV",
        "iso3_code": "TUV",
    },
    4088: {
        "gaul0_name": "Tuvalu",
        "gaul1_name": "Nukufetau",
        "iso2_code": "TV",
        "iso3_code": "TUV",
    },
    4089: {
        "gaul0_name": "Tuvalu",
        "gaul1_name": "Nukulaelae",
        "iso2_code": "TV",
        "iso3_code": "TUV",
    },
    4090: {
        "gaul0_name": "Tuvalu",
        "gaul1_name": "Vaitupu",
        "iso2_code": "TV",
        "iso3_code": "TUV",
    },
    4091: {
        "gaul0_name": "United States Minor Outlying Islands",
        "gaul1_name": "Baker Island",
        "iso2_code": "UM",
        "iso3_code": "UMI",
    },
    4092: {
        "gaul0_name": "United States Minor Outlying Islands",
        "gaul1_name": "Howland Island",
        "iso2_code": "UM",
        "iso3_code": "UMI",
    },
    4093: {
        "gaul0_name": "United States Minor Outlying Islands",
        "gaul1_name": "Jarvis Island",
        "iso2_code": "UM",
        "iso3_code": "UMI",
    },
    4094: {
        "gaul0_name": "United States Minor Outlying Islands",
        "gaul1_name": "Johnston Atoll",
        "iso2_code": "UM",
        "iso3_code": "UMI",
    },
    4095: {
        "gaul0_name": "United States Minor Outlying Islands",
        "gaul1_name": "Kingman Reef",
        "iso2_code": "UM",
        "iso3_code": "UMI",
    },
    4096: {
        "gaul0_name": "United States Minor Outlying Islands",
        "gaul1_name": "Midway Islands",
        "iso2_code": "UM",
        "iso3_code": "UMI",
    },
    4097: {
        "gaul0_name": "United States Minor Outlying Islands",
        "gaul1_name": "Palmyra Atoll",
        "iso2_code": "UM",
        "iso3_code": "UMI",
    },
    4098: {
        "gaul0_name": "United States Minor Outlying Islands",
        "gaul1_name": "Wake Island",
        "iso2_code": "UM",
        "iso3_code": "UMI",
    },
    4099: {
        "gaul0_name": "Vanuatu",
        "gaul1_name": "Malampa - Rural",
        "iso2_code": "VU",
        "iso3_code": "VUT",
    },
    4100: {
        "gaul0_name": "Vanuatu",
        "gaul1_name": "Penama - Rural",
        "iso2_code": "VU",
        "iso3_code": "VUT",
    },
    4101: {
        "gaul0_name": "Vanuatu",
        "gaul1_name": "Sanma - Rural",
        "iso2_code": "VU",
        "iso3_code": "VUT",
    },
    4102: {
        "gaul0_name": "Vanuatu",
        "gaul1_name": "Sanma - Urban",
        "iso2_code": "VU",
        "iso3_code": "VUT",
    },
    4103: {
        "gaul0_name": "Vanuatu",
        "gaul1_name": "Shefa - Rural",
        "iso2_code": "VU",
        "iso3_code": "VUT",
    },
    4104: {
        "gaul0_name": "Vanuatu",
        "gaul1_name": "Shefa - Urban",
        "iso2_code": "VU",
        "iso3_code": "VUT",
    },
    4105: {
        "gaul0_name": "Vanuatu",
        "gaul1_name": "Tafea - Rural",
        "iso2_code": "VU",
        "iso3_code": "VUT",
    },
    4106: {
        "gaul0_name": "Vanuatu",
        "gaul1_name": "Torba - Rural",
        "iso2_code": "VU",
        "iso3_code": "VUT",
    },
    4107: {
        "gaul0_name": "Wallis and Futuna Islands",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "WF",
        "iso3_code": "WLF",
    },
    4108: {
        "gaul0_name": "Antarctica",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "AQ",
        "iso3_code": "ATA",
    },
    4109: {
        "gaul0_name": "Greenland",
        "gaul1_name": "Administrative Unit Not Available",
        "iso2_code": "GL",
        "iso3_code": "GRL",
    },
}
